function [U,V,X,C,S] = gsvd(A,B)
%GSVD   Generalized Singular Value Decompostion
%   [U,V,X,C,S] = GSVD(A,B) returns unitary matrices U and V,
%   a (usually) square matrix X, and nonnegative diagonal matrices
%   C and S so that
%
%       A = U*C*X'
%       B = V*S*X'
%       C'*C + S'*S = I 
%
%   A and B must have the same number of columns, but may have
%   different numbers of rows.  If A is m-by-p and B is n-by-p, then
%   U is m-by-m, V is n-by-n and X is p-by-q where q = min(m+n,p).
%
%   SIGMA = GSVD(A,B) returns the vector of generalized singular
%   values, sqrt(diag(C'*C)./diag(S'*S)).
%
%   In this formulation of the GSVD, no assumptions are made about
%   the rank of A or B.  The matrix X has full rank if and only if
%   the matrix [A; B] has full rank.  Other formulations, eg. G. Golub
%   and C. Van Loan, "Matrix Computations", require that null(A) and
%   null(B) do not overlap and replace X by inv(X) or inv(X').

%   P. C. Hansen and C. B. Moler, 11/20/97.

[m,p]  = size(A);
[n,p1] = size(B);
if p1 ~= p
   error('A and B must have same number of columns')
end
[Q,R] = qr([A;B],0);
[U,V,Z,C,S] = csd(Q(1:m,:),Q(m+1:m+n,:));
if nargout > 1
   X = R'*Z;
else
   warn = warning;
   warning('off');   % Avoid divide by zero warning.
   U = sqrt(diag(C'*C)./diag(S'*S));
   warning(warn)
end

% ------------------------

function [U,V,Z,C,S] = csd(Q1,Q2,eco)
% CSD  Cosine-Sine Decomposition
% [U,V,Z,C,S] = csd(Q1,Q2)
%
% Given Q1 and Q2 such that Q1'*Q1 + Q2'*Q2 = I, the
% C-S Decomposition is a joint factorization of the form
%    Q1 = U*C*Z' and Q2=V*S*Z'
% where U, V, and Z are orthogonal matrices and C and S
% are diagonal matrices (not necessarily square) satisfying
%    C'*C + S'*S = I
% A third argument leads to computation of the economy size version.

% PCH+CBM 11/20/97.

[m,p] = size(Q1);
[n,p] = size(Q2);
eco = (nargin==3);
if eco
   [U,C,Z] = svd(Q1,0);
else
   [U,C,Z] = svd(Q1);
end
q = min(m,p);
C(1:q,1:q) = C(q:-1:1,q:-1:1);
U(:,1:q) = U(:,q:-1:1);
Z(:,1:q) = Z(:,q:-1:1);
Q2 = Q2*Z;
if m <= p & n <= p
   Q2(:,m+n+1-p:m) = settozero(Q2(:,m+n+1-p:m));
   k = n;
elseif min(size(C)) == 1
   k = 0;
else
   k = max([0; find(diag(C) <= 1/sqrt(2))]);
end
[V,R] = qr(Q2(:,1:k));
S = V'*Q2;
r = min(k,m);
S(:,1:r) = diagonal(S(:,1:r));
if k < min(n,p)
   r = min(n,p);
   [UT,ST,VT] = svd(S(k+1:n,k+1:r));
   S(1:k,k+1:r) = settozero(S(1:k,k+1:r));
   S(k+1:n,k+1:r) = ST;
   C(:,k+1:r) = C(:,k+1:r)*VT;
   i = k+1:m;
   j = k+1:r;
   C(i,j) = diagonal(C(i,j));
   V(:,k+1:n) = V(:,k+1:n)*UT;
   Z(:,k+1:r) = Z(:,k+1:r)*VT;
   if eco
      V = V(:,1:r);
      S = S(1:r,1:p);
   end
   [Q,R] = qr(C(k+1:q,k+1:r));
   i = k+1:q;
   j = k+1:r;
   C(i,j) = diagonal(R);
   U(:,k+1:q) = U(:,k+1:q)*Q;
end

if n < p
   % Diagonalize final block of S
   % and shuffle blocks of C and S.
   q = m+n-p;
   i = q+1:n;
   j = m+1:p;
   V(:,i) = V(:,i)*S(i,j);
   S(i,j) = diagonal(S(i,j)'*S(i,j));
   if ~isempty(j)
      S(1:q,j) = settozero(S(1:q,j));
   end
   i = [q+1:n, 1:min(n,q)];
   j = [m+1:p, 1:min(m,p)];
   C = C(:,j);
   S = S(i,j);
   Z = Z(:,j);
   V = V(:,i);
end

% Make sure C and S are positive.
k = find(sum(C,2) < 0);
U(:,k) = -U(:,k);
C(k,:) = -C(k,:);

k = find(sum(S,2) < 0);
V(:,k) = -V(:,k);
S(k,:) = -S(k,:);


% ------------------------

function D = diagonal(X)
%DIAGONAL  Make diagonal.
% X = DIAGONAL(X) zeros all the elements off the main diagonal of X.
D = triu(tril(X));
delta = norm(D-X,1);
if delta > 100*eps*norm(X)
   warning(['diagonal delta = ' num2str(delta)])
   D = X;
end


% ------------------------

function Z = settozero(X)
%SETTOZERO  Set to zero, with check
% X = SETTOZERO(X).
% Like X = 0, with a check that X is negligible.
Z = zeros(size(X));;
delta = norm(X,1);
if delta > 100*eps
   warning(['settozero delta = ' num2str(delta)])
   Z = X;
end

