function [mi,vari]=rmeanvar(i,x,mold,varold)
%RMEANVAR Pointwise recursive mean and variance estimation for vectors or arrays
%
%	[mi,vari]=RMEANVAR(i,x,mold,varold)
%	recursively calculates estimates of the mean and variance of random
%	variables. Each iterative step yields the arithmetic mean
%	and the corrected empirical variance of the already available samples.
%	2*sqrt(vari/i) gives the approximate 95% bounds of the mean values mi.
%	RMEANVAR is useful when a function is estimated several times (e.g.
%	the autocorrelation function is calculated from different time records),
%	and the uncertainty of the average is sought, but the storage of every
%	estimated function would consume too much memory.
%
%	Input arguments:
%	i = actual cycle number
%	xi = new data in cycle i (scalar, vector or array, may be complex)
%	mold = previous mean value estimate (in cycle i-1)
%	varold = previous variance estimate (in cycle i-1)
%
%	mi = new mean estimate
%	vari = new variance estimate
%
%	Reference:
%	Nelson, L. S., "Further Remark on Stably Updating Mean and Standard
%	Deviation Estimates," Comm. ACM, Vol. 22, No. 8, p. 483, Aug. 1979.

%	Written by I. Kollar
%	Last modified: 12-Jun-1995

if i<1, error('i is less than 1'), end
if rem(i,1)~=0, error('i is not integer'), end
if i==1
  mi=x;
  if nargout>1, vari=zeros(size(x)); end
else %i>1
  if (nargout>0)&(nargin<3)
    error('Previous mean value is not given')
  end
  mi=(i-1)/i*mold+(1/i)*x;
  if (nargout>1)&(nargin<4)
    error('Previous variance value is not given')
  end
  if nargout>1
    vari=(i-2)/(i-1)*varold + i/(i-1)^2*abs(x-mi).^2;
  end
end
%
%End of rmeanvar
