function [mant,exp] = bitround(sign,mant,exp,limit,round,thalf)
%BITROUND          Performs binary rounding with vector-form mantissa.
%  [MANT,EXP] = BITROUND(SIGN,MANT,EXP,LIMIT,ROUND,THALF)
%  Arguments:
%     MANT  - vector-form mantissa vith values of one and zero e.g. [1 0 1 1 0]
%     EXP   - exponent as decimal number
%     SIGN  - sign of mantissa: (+1) - positive, (-1) - negative
%             Needed in case of sign-related roundings like 'floor' or 'ceil'
%     LIMIT - after which bit of matissa is the rounding point (dec.num., may be negative)
%             E.g. LIMIT==1 then the r.p. is _after_ the first bit.
%     ROUND - string what shows the type of rounding.
%             May be: 'fix' 'floor' 'ceil' 'toinf' 'round'.
%     THALF - treating of exactly-half values in case of 'round' rounding. Otherwise
%             (other ROUND as 'round') is neglected.
%             May be: 'fix' 'floor' 'ceil' 'toinf' 'toeven' 'toodd'.
%
%  BITROUND is used by the string-to-segment function 'str2seg'

% Copyright (C) 2003,2004 by JUHASZ Laszlo

error(nargchk(6,6,nargin))

% check input
if isempty(sign) | isempty(mant) | isempty(exp) | isempty(limit) | isempty(round) | ...
      ~isa(sign,'double') | ~isa(mant,'double') | ~isa(exp,'double') | ...
      ~isa(limit,'double') | ~isa(round,'char') | ~isa(thalf,'char') | ...
      prod(size(sign))~=1 | prod(size(mant))~=length(mant) | prod(size(exp))~=1 | ...
      prod(size(limit))~=1 | size(round,1)~=1 | (~isempty(thalf) & size(thalf,1)~=1) | ...
      ~isfinite(sign) | ~all(isfinite(mant)) | ~isfinite(exp) | ~isfinite(limit) | ...
      rem(sign,1) | any(rem(mant,1)) | rem(exp,1) | rem(limit,1) | ...
      ~(sign<=1 & sign>=-1) | ~all(mant==(~~mant))
   error('Internal error. Bad arguments.')
end

% filtering out full zero mantissa -> no chance of bitup so always truncate
if ~any(mant)
   mant = 0;
   exp = 0;
   return
end

if limit >= length(mant)-1                     % make mant enough long
   %                                             limit is e.g. in case of [0,0,0] mant
   %                                             which length is 3 - 2.
   %                                             We need the second bit after r.p.
   mant(limit+2) = 0;
end

if limit < 1
   % if r.p. is before the whole mant
   switch round
   case 'fix'
      % simply truncate
      mant = 0;
      exp = 0;
   case 'toinf'
      % simply overflow (bitup) to the digit before r.p.
      mant = 1;
      exp = exp-limit+1;
   case 'floor'
      if sign==1
         % positive -> truncate
         mant = 0;
         exp = 0;
      else
         % negative -> bitup
         mant = 1;
         exp = exp-limit+1;
      end
   case 'ceil'
      if sign==-1
         % negative -> truncate
         mant = 0;
         exp = 0;
      else
         % positive -> bitup
         mant = 1;
         exp = exp-limit+1;
      end
   case 'round'
      if limit<0 | ~mant(1)
         % by e.g. -1 limit the whole mant after r.p. is surely less than half
         % so truncating is quaranteed.
         % The same case is when limit is zero and the first bit of mant is also zero.
         mant = 0;
         exp = 0;
         % ----
         % in what follows LIMIT is 0
      elseif mant(1)
         % there is chance of bitup if the first bit of mant is one
         if sum(mant)>=2
            % minimum two ones in mant so mant after r.p. is surely greater than half
            % so bitup is needed.
            mant = 1;
            exp = exp+1;
         else
            % mant has one bit set in the first position directly after r.p.
            % This is exactly half: 0.[1 0 0 0...]
            switch thalf
            case {'fix', 'toeven'}
               % toeven is included here because the bit before r.p. is zero (even)
               % so truncating is needed.
               mant = 0;
               exp = 0;
            case {'toinf', 'toodd'}
               % toodd is included here because the bit before r.p. is zero (even)
               % so bitup is needed.
               mant = 1;
               exp = exp+1;
            case 'floor'
               if sign==1
                  % positive -> truncate
                  mant = 0;
                  exp = 0;
               else
                  % negative -> bitup
                  mant = 1;
                  exp = exp+1;
               end
            case 'ceil'
               if sign==-1
                  % negative -> truncate
                  mant = 0;
                  exp = 0;
               else
                  % positive -> bitup
                  mant = 1;
                  exp = exp+1;
               end
            otherwise
               error(['Unsupported THALF ',thalf])
            end
         end      % of 'if sum(mant)>=2'
      else
         % all alternatives has been covered
         error('Internal error 1')
      end
   otherwise
      error(['Unsupported ROUND ',round])
   end
else
   % rounding point follows a real bit and isn't before the whole mant
   switch round
   case 'fix'
      % simply truncate
      mant(limit+1:end) = 0;
   case 'toinf'
      % bitup all
      [mant,exp] = bitup(mant,exp,limit);
   case 'floor'
      if sign==1
         % positive -> truncate
         mant(limit+1:end) = 0;
      else
         % negative -> bitup
         [mant,exp] = bitup(mant,exp,limit);
      end
   case 'ceil'
      if sign==-1
         % negative -> truncate
         mant(limit+1:end) = 0;
      else
         % positive -> bitup
         [mant,exp] = bitup(mant,exp,limit);
      end
   case 'round'
      if mant(limit+1) & ~any(mant(limit+2:end))
         % after r.p. is exactly half [.1 0 0...]
         switch thalf
         case 'fix'
            % simply truncate
            mant(limit+1:end) = 0;
         case 'toinf'
            % bitup all
            [mant,exp] = bitup(mant,exp,limit);
         case 'floor'
            if sign==1
               % positive -> truncate
               mant(limit+1:end) = 0;
            else
               % negative -> bitup
               [mant,exp] = bitup(mant,exp,limit);
            end
         case 'ceil'
            if sign==-1
               % negative -> truncate
               mant(limit+1:end) = 0;
            else
               % positive -> bitup
               [mant,exp] = bitup(mant,exp,limit);
            end
         case 'toeven'
            if ~mant(limit)
               % bit before r.p. is 0 (even) -> truncate
               mant(limit+1:end) = 0;
            else
               % bit before r.p. is 1 (odd)  -> bitup
               [mant,exp] = bitup(mant,exp,limit);
            end
         case 'toodd'
            if mant(limit)
               % bit before r.p. is 1 (odd)  -> truncate
               mant(limit+1:end) = 0;
            else
               % bit before r.p. is 0 (even) -> bitup
               [mant,exp] = bitup(mant,exp,limit);
            end
         otherwise
            error(['Unsupported THALF ',thalf])
         end
      elseif mant(limit+1)
         % after rounding point is 1 and there are another bits set after that
         % rounding upwards -> bitup
         [mant,exp] = bitup(mant,exp,limit);
      else
         % after rounding point is 0 so round downwards -> truncate
         mant(limit+1:end)=0;
      end
   otherwise
      error(['Unsupported ROUND ',round])
   end
end

% simplify
[mant,exp] = rmnull(mant,exp);
