function [exp2,err] = procexp(exp)
%PROCEXP           Process exponent.
%  Usage: [EXP,ERR] = PROCEXP(EXP)
%  Input argument: string containing exponent.
%  Output arguments:
%     EXP - scalar decimal exponent
%     ERR - error message

% Copyright (C) 2004 by JUHASZ Laszlo

error(nargchk(1,1,nargin))

exp2 = []; err = '';

if ~ischar(exp) | (~isempty(exp) & size(exp,1)~=1)
   error('Bad arguments.')
end

if isempty(exp)
   err = 'Empty exponent.';
   return
end

% filter out zeros
if all(exp=='0')
   exp2 = 0;
   return
end

format = exp(end);
exp(end) = '';

switch format
case 'd'                            % decimal
   % check
   if isempty(exp)
      err = 'Missing exponent before decimal format indicator.';
      return
   end
   err = strchk(exp,'+-0123456789','decimal exponent');
   if ~isempty(err), return, end
case 'b'                            % binary
   % check
   if isempty(exp)
      err = 'Missing exponent before binary format indicator.';
      return
   end
   err = strchk(exp,'+-01','binary exponent');
   if ~isempty(err), return, end
case 'x'                            % hexadecimal
   % check
   if isempty(exp)
      err = 'Missing exponent before hexadecimal format indicator.';
      return
   end
   err = strchk(exp,'+-0123456789abcdef','hexadecimal mantissa');
   if ~isempty(err), return, end
   % replace
   exp = strrep(exp, '0', '0000');
   exp = strrep(exp, '1', '0001');
   exp = strrep(exp, '2', '0010');
   exp = strrep(exp, '3', '0011');
   exp = strrep(exp, '4', '0100');
   exp = strrep(exp, '5', '0101');
   exp = strrep(exp, '6', '0110');
   exp = strrep(exp, '7', '0111');
   exp = strrep(exp, '8', '1000');
   exp = strrep(exp, '9', '1001');
   exp = strrep(exp, 'a', '1010');
   exp = strrep(exp, 'b', '1011');
   exp = strrep(exp, 'c', '1100');
   exp = strrep(exp, 'd', '1101');
   exp = strrep(exp, 'e', '1110');
   exp = strrep(exp, 'f', '1111');
case 'o'                            % octal
   % check
   if isempty(exp)
      err = 'Missing exponent before octal format indicator.';
      return
   end
   err = strchk(exp,'+-01234567','octal mantissa');
   if ~isempty(err), return, end
   % replace
   exp = strrep(exp, '0', '000');
   exp = strrep(exp, '1', '001');
   exp = strrep(exp, '2', '010');
   exp = strrep(exp, '3', '011');
   exp = strrep(exp, '4', '100');
   exp = strrep(exp, '5', '101');
   exp = strrep(exp, '6', '110');
   exp = strrep(exp, '7', '111');
otherwise
   err = ['Incorrect format indicator ''',format,''' at the end of exponent.'];
   return
end

% analizing sign
signpos = find(exp=='+' | exp=='-');
if isempty(signpos)                          % no sign character
   sign = 1;                                 % sign is plus
elseif length(signpos)>1
   err = ['More than one ''',exp(signpos),''' sign characters in exponent.'];
   return
elseif signpos~=1
   err = ['Sign isn''t at start of exponent.'];
   return
elseif signpos==length(exp)
   err = 'Exponent is only a sign character.';
   return
elseif all(exp(signpos+1:end)=='.')
   err = 'Exponent contains only comma(s) after sign.';
   return
else
   if exp(1) == '+'
      sign = 1;
   elseif exp(1) == '-'
      sign = -1;
   else
      error('Internal error 1')
   end
   exp(1) = '';                    % cutting sign from start
end

if format=='d'
   % decimal exponent - try to read in with 'sscanf'
   [exp,ssn,sserr] = sscanf(exp,'%d',1);
   if ~isempty(sserr)
      err = ['Error by decimal exponent scan: ',sserr];
      return
   elseif ssn~=1
      err = 'Number of decimal exponent isn''t 1.';
      return
   end
   exp = sign*exp;
else
   % binary exponent
   % convert to numeric
   exp = double(exp-'0');
   % cut leading zeros
   if any(exp)
      exp = exp(min(find(exp)):end);
   else
      exp = 0;
   end
   if length(exp)>53
      err = 'Too long exponent.';
      return
   end
   exp = sign*bindec(exp);
end

exp2 = exp;
