function [x,of,uf,nanv,warn,err] = seground(x,code)
%SEGROUND  Performs numerical rounding on segment-form numbers.
%
% [X,OF,UF,NANV,WARN,ERR] = SEGROUND(X,CODE)
%
% input arguments
%  X: two-dimensional array in the form that segments are in the rows and
%     numbers are in the columns. The support for multi-dimensional arrays must be
%     rather a workaround in the user method file such as qdata/plus.m and not in this
%     function.
%     Example:
%           num(1)seg1 num(2)seg1 num(3)seg1 ...
%           num(1)seg2 num(2)seg2 ...
%           num(1)seg3 ...
%           ...
%  CODE: codes in string-form
%
% output arguments
%  X:    rounded and corrected segments with the same dimension as the initial X
%  OF:   overflow index vector (if needed to store by CODE)
%  UF:   underflow index vector (if needed to store by CODE)
%  NANV: unallowed NaN index vector (if needed to store by CODE)
%  WARN: warning message
%  ERR:  error message

% Written by JUHASZ Laszlo, 2004
% Modifyied by Istvn Kollr, 12-May-2006

% __        __                   _
% \ \      / /__ _  _ __  _ __  (_) _ __    __ _
%  \ \ /\ / // _` || '__|| '_ \ | || '_ \  / _` |
%   \ V  V /| (_| || |   | | | || || | | || (_| |
%    \_/\_/  \__,_||_|   |_| |_||_||_| |_| \__, |
%                                          |___/
% Detecting and converting back to 'normal form' of physically denormalized numbers
% is complete but the proper writing back is missing.

error(nargchk(2,2,nargin))
% checking the dimension of 'x'. It's here because assuming 'x' as a
% two-dimensional array is an important part of the algorithm.
if ndims(x)~=2, error('Internal error. Input must be 2-dimensional.'), end
% initialize warning and error messages
warn = ''; err = '';
eval(code)

xsiz=size(x);
if (prod(xsiz(2:end))>1)%&strcmp(TYPE,'fixed-point')
  %temporary solution: x vector
  of=zeros(size(x(1,:,:,:))); uf=of; nanv=of; warn=''; err='';
  for ii=1:prod(xsiz(2:end))
    [xii,ofii,ufii,nanvii,warnii,errii] = seground(x(:,ii),code);
    x(1:length(xii),ii)=xii; of(ii)=ofii; uf(ii)=ufii; nanv(ii)=nanvii;
    if isempty(warn), warn=warnii; end
    if isempty(err), err=errii; end
  end %for ii
  return
end

% Make 'x' enough long for rounding points greater than the length of number
% and make 'x' minimum 3 words wide.
% number of segments is in the 1st dimension
x(max(size(x,1),SEGNUM)+2,:) = 0;

switch TYPE
  case {'floating-point' 'fixed-point'}
    % calculate rounding point
    if strcmp(TYPE,'floating-point')
      rp = FRACTION+1;
    else %fixed-point
      %if size(x,2)>1, error('Not yet ready for vector inputs'), end
      [f,e]=log2(x(1,:));
      rp = e+FRACTIONBITS; EMAX=PRECISION-FRACTIONBITS-1; EMIN=-FRACTIONBITS;
    end
    % num of segment to round
    numofseg = max(ceil(rp/22),1);
    % rounding point within segment
    rpinseg = rem(rp-1,22)+1; 
    % generating 'free' vector to index the actually not yet handled colums.
    % The initial state of 'free' contains all the columns yet.
    free = ones(1,size(x,2));
    % initializing counters
    of = ~free;             % full zeros
    uf = of; in = of;
    % detect NaNs - number treated as NaN when the first segment
    % (currently the first row) is NaN
    nanv = isnan(x(1,:));
    % clear the columns containing NaNs
    x(:,nanv) = 0;
    % delete 'NaN-columns' from the available free columns
    free = free & ~nanv;
    % detect zeros - The 'any' operates on columns so it's zero at columns
    % where all rows is zero
    zerosarr = ~any(x);
    % delete 'zero-columns' from the available free columns
    free = free & ~zerosarr;
    % detect sign - It detects the sign in case of physically denormalized
    % numbers as well where the leading first segment(s) can be zero.
    % 's' is a row-vector
    s = sign(sum(sign(x))).*free;
    % in what follows we operate on positive segments
    x = abs(x);
    % detect Infs - number treated as Inf when the first segment is Inf
    infs = isinf(x(1,:)) & free;
    % clear Infs
    x(:,infs) = 0;
    % delete 'Inf-columns' from the available free columns
    free = free & ~infs;
    % extract mantissas and exponents - the mantissa goes back to 'x'
    [x,e] = log2(x);
    if any(free)
      % detect physically normalized numbers
      % Number is normalized when first segment is nonzero and its exponent isn't
      % less than -1053 so in the actual notation where the maximum mantissa is
      % 0.5 - when the first segment's exponent isn't less than -1052.
      %    xxx1=(e(1,:)>=-1052)
      %    xxx2=x(1,:)
      %    xxx3=free
      %    xxx1 & xxx2 & xxx3
      normld = (e(1,:)>=-1052) & x(1,:) & free;
      
      % Only one chance has remained: number is denormalized
      denorm = free & ~normld;
      if any(denorm)
        % finding first notnull segment for every number with a
        % sneaky algorithm :) - 'fs' is a row-vector
        fs = max(cumsum(~cumsum(x)))+1;
        % extracting the exponent at the first notnull segment - To do this we
        % have to perform a bit of index-manipulation.
        te = e(:).';     % converting 'exp' to column-vector in 'te'
        % converting 'fs' to absolute address instead of relative address
        % within column
        as = fs+((1:size(e,2))-1)*size(e,1);
        % We have the correct indices so we can calculate now exponents in first
        % notnull segment:
        te = te(as);
        % extend to the size of 'e' - multiply rows
        re = te(ones(size(e,1),1),denorm);
        % exponent corrector values - These are zeros in the place of 1st notnull segment
        % and increments with 22 towards the lesser significant segments.
        % At first we create the first row of 'ecorr' then append the
        % properly-sized array full of value 22, at last we let 'cumsum' to do the
        % cumulative addition along columns.
        ecorr = cumsum([-22.*(fs(denorm)-1);22.*ones(size(e,1)-1,sum(denorm))]);
        % correct mantissas - the 1st notnull segment wouldn't be corrected
        x(:,denorm) = pow2(x(:,denorm),re-e(:,denorm)+ecorr);
        % generating main exponent - It's smaller than the exponent of first
        % notnull segment with 22*(fs-1)
        e(1,denorm) = te(denorm)-22.*(fs(denorm)-1);
        % 'bring-up' denormalized segments to the first place where the first notnull
        % segment is second or greater
        for k = find(denorm & (fs>=2))
          % copy the segments beginning with the first notnull segment to the
          % first place
          x(1:end-fs(k)+1,k) = x(fs(k):end,k);
          % delete the segments from the end of number
          x(end-fs(k)+2:end,k) = 0;
        end
        % how many digits are in the first notnull segment? - The central
        % expression generates the parallel lines plotted against the main
        % exponent and the 1st notnull segment:
        %                           /   /   /
        %                          /   /   /
        %                         /   /   /
        %                        /   /   /
        % The min-max expression limits the output of function to the interval
        % [0,22]
        d = min(max(e(1,denorm)+1074+22.*(fs(denorm)-1),0),22);
        % re-arrange denormalized mantissas in ascending order.
        %             ________    _____
        %            V        \  V     \
        %  |_xxx_|_empty_|   |_yyy_|_zzzzz_|      <-- segments
        %
        %  \__ _/ \__ __/    \__ _/ \__ __/
        %     V      V          V      V
        %     d    22-d         d    22-d         <-- size of slices
        %  \_____ ______/    \_____ ______/
        %        V                 V
        %        22                22             <-- size of segments
        %
        % pick out the first row from loop because the second operation in loop
        % mustn't be performed on it
        x(1,denorm) = x(1,denorm)+pow2(fix(pow2(x(2,denorm),22-d)),-22);
        for k = 2:size(x,1)-1
          % Add the slice from the next segment to the current segment
          % (for the first time the current segment is the 2nd)
          % The size of slice is '22-d'.
          x(k,denorm) = x(k,denorm)+pow2(fix(pow2(x(k+1,denorm),22-d)),-22);
          % slide up the remained slice - 'rem' filters out the already
          % moved slice with '22-d' length, so the length of result is 'd'
          % (for the first time the remained slice is the 3rd)
          x(k+1,denorm) = rem(pow2(x(k+1,denorm),22-d),1);
        end
      end
      
      % variable for reference exponents - the reference exponent is the first
      % segment's exponent so the first row's exponent
      refe = e(1,:);
      % extend to 'x's size - multiple rows
      refe = refe(ones(size(x,1),1),:);
      % sliding mantissas to reference exponent position - this operation
      % transforms back 'x' to its initial state with possible leading zeros
      % (avoiding the built-in normalization of Matlab)
      x(:,normld) = pow2(x(:,normld),e(:,normld)-refe(:,normld));
      % saving 'x' and 'refe' for use at denormalization for re-rounding because the following
      % operations operate on 'x' directly
      densavex = x;
      densaverefe = refe;
      % The next segment is included after the last bit of row 'numofseg'
      % in order to transfer less argument to the rounding function.
      % Subsequent segments are included after the last bit of column 'numofseg'
      % and after the aready copied next segment
      % in order to transfer less argument to the rounding function.
      % The exponent in second 'pow2' operation is 45 instead of 44 because
      % the result of 'any' is  always 1 so it's the double amount of 0.5 what
      % is the greatest available power-of-two in segments.
      x(numofseg,:) = x(numofseg,:)+pow2(x(numofseg+1,:),-22)+pow2(0+any(x(numofseg+2:end,:)),-45);
      % clear subsequent segments - All information is already encoded into row
      % 'numofseg'.
      x(numofseg+1:end,:) = 0;
      % call specific internal rounding function (see end of this file)
      % The type of rounding is coded into the name of function.
      [f1,e1]=log2(x(1,:));
      if ~all(x(2:end)==0)
        if f1-fix(f1*2^22)/2^22>0
          error('Too many bits in first segment')
        end
      end
      [fii,eii]=log2(x(numofseg,:));
      x(numofseg,:) = feval([ROUNDINGDIR,TREATHALF],s,x(numofseg,:),rpinseg);
      % for case if segment overflow occured correct segment-rows in reversed
      % order. It's meaningless to investigate rows after 'numofseg'.
      % ('numofseg' is common to all numbers)
      for k = numofseg:-1:1
        % add carry to the more significant segment-column
        x(k,:) = x(k,:)+(x(k+1,:)>=1)/2^22;
        % substract carry from the less significant segment-column
        x(k+1,:) = rem(x(k+1,:),1);
      end
      % if overflow of 1st segment occurred, modify exps and mants
      % in which columns occurred?
      if strcmp(CODING,'twos-comp')&isequal(s,-1), segovf = x(1,:)>1;
      else segovf = x(1,:)>=1;
      end
      % increase exponents
      refe(1,segovf) = refe(1,segovf)+1;
      % decrease mants
      x(1,segovf) = x(1,segovf)/2;
      % branches for overflow, underflow, illegal NaN and denormalization
      if strcmp(TYPE,'floating-point')
        switch CODING
          case 'sign-mag'
            % ***** SIGN-MAGNITUDE coded numbers *****
            % check OVERFLOW
            % simply where the exponent is greater than EMAX so with 0.5 maximum
            % mantissa where the exponent is greater than EMAX+1...
            of = (refe(1,:)>(EMAX+1)) & free;
            % ...and where are unallowed Infs
            if ~strcmp(OVERFLOW,'inf')
              of = of | infs;
              % prevention from filling Inf-columns with Inf
              infs = infs.*0;
            end
            % handling of OVERFLOW
            if any(of)
              switch OVERFLOW
                case 'inf'
                  % filling the overflowed cols with correctly signed infs
                  x(:,of) = Inf.*s(ones(size(x,1),1),of);
                case 'clip'
                  % filling the overflowed rows with correctly signed clipped segments
                  % generating one row to further extension
                  clip = clipgen(FRACTION+1,EMAX,size(x,1));
                  % extension and multiplying with proper sign in one step
                  x(:,of) = clip(:,ones(sum(of),1)).*s(:,of);
                otherwise
                  error(['Unsupported OVERFLOW ',OVERFLOW])
              end
            end
            % check UNDERFLOW
            switch UNDERFLOW
              case 'flush'
                % denormalized numbers are underflowed as well
                % simply where the exponent is smaller than EMIN so with 0.5
                % maximum mantissa where the exponent is smaller than EMIN+1
                uf = (refe(1,:)<(EMIN+1)) & free;
                den = [];
              case 'gradual'
                % underflow occurrs only below EMIN-FRACTION
                uf = (refe(1,:)<(EMIN-FRACTION+1)) & free;
                % denormalization occurrs between EMIN and EMIN-FRACTION (EMIN
                % not included)
                den = (refe(1,:)<(EMIN+1)) & (refe(1,:)>=(EMIN-FRACTION+1)) & free;
              otherwise
                error(['Unsupported UNDERFLOW ',UNDERFLOW])
            end
            if any(uf)
              % fill the underflowed columns with zeros
              x(:,uf) = 0;
            end
            if any(den)
              % Reinstall original configuration in selected columns from the
              % saved variables 'densavex' and 'densaverefe'.
              x(:,den) = densavex(:,den);
              refe(:,den) = densaverefe(:,den);
              % recalculate rounding point for every denormalized number in one
              % step - The rounding points can be different according to their
              % exponents. (Note: we compute 'rp' for all numbers but only the
              % denormalized ones will be used)
              rp = refe(1,:)+FRACTION-EMIN;
              % rounding - The parameters 'numofseg' and 'rpinseg' must be
              % computed for every number beacuse the 'rp' rounding point
              % parameter can be different for every number as well.
              for k = find(den)
                % number of segment to round
                numofseg = ceil(rp(k)/22);
                % The next segment is included after the last bit of row 'numofseg'
                % in order to transfer less argument to the rounding function.
                % Subsequent segments are included after the last bit of column 'numofseg'
                % and after the aready copied next segment
                % in order to transfer less argument to the rounding function
                % The exponent in second 'pow2' operation is 45 instead of 44 because
                % the result of 'any' is  always 1 so it's the double amount of 0.5 what
                % is the greatest available power-of-two in segments.
                x(numofseg,k) = x(numofseg,k)+pow2(x(numofseg+1,k),-22)+...
                  pow2(0+any(x(numofseg+2:end,k)),-45);
                % clear subsequent segments - All information is already encoded into row
                % 'numofseg'.
                x(numofseg+1:end,k) = 0;
                % rounding point within segment
                rpinseg = rem(rp(k)-1,22)+1;
                % call specific rounding function
                x(numofseg,k) = feval([ROUNDINGDIR,TREATHALF],s(k),x(numofseg,k),rpinseg);
              end
              % for case if segment overflow occured correct segment-rows in
              % reversed order. 'numofseg' isn't common so we have to
              % investigate all rows
              for k = size(x,1)-1:-1:1
                % add carry to the more significant next segment-column
                x(k,den) = x(k,den)+(x(k+1,den)>=1)/2^22;
                % substract carry from the less significant segment-column
                x(k+1,den) = rem(x(k+1,den),1);
              end
              % if overflow of 1st segment occurred, modify exps and mants
              % in which columns occurred?
              segovf = (x(1,:)>=1) & den;
              % increase exps
              refe(1,segovf) = refe(1,segovf)+1;
              % decrease mants
              x(1,segovf) = x(1,segovf)/2;
            end
          case 'twos-comp'
            % ***** TWOS-COMPLEMENT coded numbers *****
            % check OVERFLOW
            % on positive numbers simply where the exponent is greater than EMAX
            % so with 0.5 maximum mantissa where the exponent is greater than
            % EMAX+1...
            pof = (refe(1,:)>(EMAX+1)) & (s>=0) & free;
            % ...on negative numbers where the length of mantissa is greater than
            % one -> simply where the exponent is greater than EMAX...
            nof = (rem(2.*x(1,:),1) | any(x(2:end,:))) & (refe(1,:)>(EMAX+1)) & (s<0) & free;
            % ...on negative numbers where the length of mantissa is one ->
            % where the exponent is greater than EMAX+1...
            nof = nof | (~rem(2.*x(1,:),1) & ~any(x(2:end,:)) & (refe(1,:)>(EMAX+2)) & (s<0) & free);
            % generate summarized overflow index
            of = pof | nof;
            % ...and where are unallowed Infs
            if ~strcmp(OVERFLOW,'inf')
              of = of | infs;
              % prevention from filling Inf-columns with Inf
              infs = infs.*0;
            end
            % handling of OVERFLOW
            if any(of)
              switch OVERFLOW
                case 'inf'
                  % filling the overflowed cols with correctly signed infs
                  x(:,of) = Inf.*s(ones(size(x,1),1),of);
                case 'clip'
                  if any(pof)
                    % in case of positive numbers -> filling the overflowed
                    % columns with positive clipped segments
                    % generating one row to further extension
                    clip = clipgen(FRACTION+1,EMAX,size(x,1));
                    % extension
                    x(:,pof) = clip(:,ones(sum(pof),1));
                  end
                  if any(nof)
                    % in case of negative numbers -> fill the overflowed
                    % columns in the first row with 1*2^(EMAX+1) that is
                    % 0.5*2^(EMAX+2)
                    x(1,nof) = 0.5;
                    refe(1,nof) = EMAX+2;
                    % clear the subsequent segments
                    x(2:end,nof) = 0;
                  end
                otherwise
                  error(['Unsupported OVERFLOW ',OVERFLOW])
              end
            end
            % check UNDERFLOW
            switch UNDERFLOW
              case 'flush'
                % denormalized numbers are underflowed as well
                den = [];
                % in case of positive numbers simply where the exponent is smaller
                % than EMIN so with 0.5 maximum mantissa where the exponent is
                % smaller than EMIN+1...
                pof = (refe(1,:)<(EMIN+1)) & (s>=0) & free;
                % ...in case of negative numbers and 'flush' underflow where the
                % length of mantissa is greater than one -> simply where the
                % exponent is smaller than EMIN...
                nuf = (rem(2.*x(1,:),1) | any(x(2:end,:))) & (refe(1,:)<(EMIN+1)) & (s<0) & free;
                % ...in case of negative numbers and 'flush' underflow where the
                % length of mantissa is exactly one -> where the
                % exponent is smaller than EMIN+1 because in twos-complement
                % notation -1*2^EMIN is already denormalized -> so underflowed
                % in 'flush' case.
                nuf = nuf | (~rem(2.*x(1,:),1) & ~any(x(2:end,:)) & (refe(1,:)<(EMIN+2)) & (s<0) & free);
                % generate summarized underflow index
                uf = puf | nuf;
              case 'gradual'
                % underflow occurrs only below EMIN-FRACTION - This limit is
                % equal in case of positive and negative twos-comp numbers (I hope).
                uf = (refe(1,:)<(EMIN-FRACTION+1)) & free;
                % Denormalization should be separated to positive and negative
                % situations and for two further possible cases by negative numbers
                % where the length of mantissa is only one or more than one.
                % In case of negative twos-comp numbers -1*2^EMIN is already
                % denormalized but this special case affects only displaying
                % so re-rounding must not be performed. That's why the
                % negative case isn't separated.
                %
                % In case of positive numbers denormalization occurrs between
                % EMIN and EMIN-FRACTIION (EMIN not included)...
                pden = (refe(1,:)<(EMIN+1)) & (refe(1,:)>=(EMIN-FRACTION+1)) & (s>=0) & free;
                % ...in case of negative numbers where the
                % length of mantissa is greater than one -> simply where the
                % exponent is between EMIN and EMIN-FRACTION (EMIN not
                % included).
                nden = (rem(2.*x(1,:),1) | any(x(2:end,:))) & ...
                  (refe(1,:)<(EMIN+1)) & (refe(1,:)>=(EMIN-FRACTION+1)) & (s<0) & free;
                % generate summarized denormalization index
                den = pden | nden;
              case ''
                den=[];
              otherwise
                error(['Unsupported UNDERFLOW ',UNDERFLOW])
            end
            if any(uf)
              % fill the underflowed columns with zeros
              x(:,uf) = 0;
            end
            if any(den)
              % re-rounding is necessary in case of logically denormalized numbers
              % Reinstall original configuration in selected columns from the
              % saved variable 'densave'.
              % original mantissas and exponents in selected columns
              [x(:,den),e(:,den)] = log2(densave(:,den));
              % temporary variable for reference exponents in selected columns for
              % further extension
              t = e(1,den);
              % extend reference exponents
              refe(:,den) = t(ones(size(x,1),1),:);
              % sliding mantissas to reference exponent position
              mant(:,den) = pow2(x(:,den),e(:,den)-refe(:,den));
              % recalculate rounding point for every denormalized number in one
              % step - The rounding points can be different according to their
              % exponents. (Note: we compute 'rp' for all numbers but only the
              % denormalized ones will be used)
              % __        __                   _
              % \ \      / /__ _  _ __  _ __  (_) _ __    __ _
              %  \ \ /\ / // _` || '__|| '_ \ | || '_ \  / _` |
              %   \ V  V /| (_| || |   | | | || || | | || (_| |
              %    \_/\_/  \__,_||_|   |_| |_||_||_| |_| \__, |
              %                                          |___/
              % I don't know whether is this algorithm correct for twos-comp
              % denormalized numbers as well
              rp = refe(1,:)+FRACTION-EMIN;
              % rounding - The parameters 'numofseg' and 'rpinseg' must be
              % computed for every number beacuse the 'rp' rounding point
              % parameter can be different for every number as well.
              for k = find(den)
                % number of segment to truncate
                numofseg = ceil(rp(k)/22);
                % The next segment is included after the last bit of row 'numofseg'
                % in order to transfer less argument to the rounding function.
                % Subsequent segments are included after the last bit of column 'numofseg'
                % and after the aready copied next segment
                % in order to transfer less argument to the rounding function
                % The exponent in second 'pow2' operation is 45 instead of 44 because
                % the result of 'any' is  always 1 so it's the double amount of 0.5 what
                % is the greatest available power-of-two in segments.
                x(numofseg,k) = x(numofseg,k)+pow2(x(numofseg+1,k),-22)+...
                  pow2(0+any(x(numofseg+2:end,k)),-45);
                % clear subsequent segments - All information is already encoded into row
                % 'numofseg'.
                x(numofseg+1:end,k) = 0;
                % rounding point within segment
                rpinseg = rem(rp(k)-1,22)+1;
                % call specific rounding function
                x(numofseg,k) = feval([ROUNDINGDIR,TREATHALF],s(k),x(numofseg,k),rpinseg);
              end
              % for case if segment overflow occured correct segment-rows in
              % reversed order. 'numofseg' isn't common so we have to
              % investigate all rows
              for k = size(x,1)-1:-1:1
                % add carry to the more significant next segment-column
                x(k,den) = x(k,den)+(x(k+1,den)>=1)/2^22;
                % substract carry from the less significant segment-column
                x(k+1,den) = rem(x(k+1,den),1);
              end
              % if overflow of 1st segment occurred, modify exps and mants
              % in which columns occurred?
              segovf = (x(1,:)>=1) & den;
              % increase exps
              refe(1,segovf) = refe(1,segovf)+1;
              % decrease mants
              x(1,segovf) = x(1,segovf)/2;
            end
          otherwise
            error(['Unsupported CODING ',CODING])
        end
      elseif strcmp(TYPE,'fixed-point')
        x=pow2(s*x,refe);
        of = compvect(x,MAXIMUM,'>') & free;
        if ~strcmp(OVERFLOW,'inf')
          of = of | infs;
          % prevention from filling Inf-columns with Inf
          infs = infs.*0;
        end
        % handling of OVERFLOW
        if any(of)
          switch OVERFLOW
            case 'inf'
              % filling the overflowed cols with correctly signed infs
              x(:,of) = Inf.*s(ones(size(x,1),1),of);
            case 'clip' %*%*%*%
              % filling the overflowed rows with correctly signed clipped segments
              % generating one row to further extension
              maxv=MAXIMUM; while size(x,1)>size(maxv,1), maxv=[maxv;0]; end
              x(:,of) = maxv(:,of);
              % extension and multiplying with proper sign in one step
              %x(:,of) = clip(:,ones(sum(of),1)).*s(:,of);
              %case 'modular'
              %maxv=MAXIMUM;
              %  x=x-
            otherwise
              error(['Unsupported OVERFLOW ',OVERFLOW])
          end
        end
        % check UNDERFLOW
        uf = compvect(x,MINIMUM,'<') & free;
        if any(uf)
          switch OVERFLOW
            case 'inf'
              % filling the overflowed cols with correctly signed infs
              x(:,of) = Inf.*s(ones(size(x,1),1),of);
            case 'clip'
              % filling the overflowed rows with correctly signed clipped segments
              % generating one row to further extension
              minv=MINIMUM; while size(x,1)>size(minv,1), minv=[minv;0]; end
              x(:,uf) = minv(:,uf);
              % extension and multiplying with proper sign in one step
              %x(:,of) = clip(:,ones(sum(of),1)).*s(:,of);
              %case 'modular'
              %maxv=MAXIMUM;
              %  x=x-
            otherwise
              error(['Unsupported OVERFLOW ',OVERFLOW])
          end
        end
      end %floating-point or fixed-point
    end % end of 'any free'
    % COMMON PART of Sign-Mag & Twos-Comp
    % put back Infs with proper sign - extend sign to SEGNUM then multiply
    % with Inf
    if any(infs)
      x(:,infs) = Inf.*s(ones(size(x,1),1),infs);
    end
  otherwise
    error(['Unsupported TYPE ',TYPE])
end               % end of switch TYPE
% COMMON PART of floating-point & fixed-point
% generate error and warning messages, reset statistical indexes when storing is
% not necessary
% handling of NaNs
if any(nanv)
  % check claim to error message
  if any(findstr(NANSIGNAL,'error'))
    err = 'Target format doesn''t include NaN.';
    return
  end
  % filling the NaN-columns with NaN
  x(:,nanv) = nan;
  % check claim to warning message
  if any(findstr(NANSIGNAL,'warning'))
    warn = 'NaN occurred during operation.';
  end
  % check claim to non-stored stat info -> not goes to quantizers
  if ~any(findstr(NANSIGNAL,'count'))
    nanv = nanv.*0;
  end
end
% handling of overflow messages
if any(of)
  % check claim to error message
  if any(findstr(OFSIGNAL,'error'))
    err = 'Overflow occurred during operation.';
    return
  end
  % check claim to warning message
  if any(findstr(OFSIGNAL,'warning'))
    warn = 'Overflow occurred during operation.';
  end
  % check claim to non-stored stat info -> not goes to quantizers
  if ~any(findstr(OFSIGNAL,'count'))
    of = of.*0;
  end
end
% handling of underflow messages
if any(uf)
  % check claim to error message
  if any(findstr(UFSIGNAL,'error'))
    err = 'Underflow occurred during operation.';
    return
  end
  % check claim to warning message
  if any(findstr(UFSIGNAL,'warning'))
    warn = 'Underflow occurred during operation.';
  end
  % check claim to non-stored stat info -> not goes to quantizers
  if ~any(findstr(UFSIGNAL,'count'))
    uf = uf.*0;
  end
end

if strcmp(TYPE,'floating-point')
  % preventing operation on already handled special cases
  free = free & ~(of|uf|nanv);
  if any(free)
    % detect physically normalized numbers
    % Number is normalized when first segment is nonzero and its exponent isn't
    % less than -1053 so in the actual notation where the maximum mantissa is
    % 0.5 - when the first segment's exponent isn't less than -1052.
    normld = (e(1,:)>=-1052) & x(1,:) & free;
    % Only one chance has remained: number is denormalized
    denorm = free & ~normld;
    % write back 'mant' raised to 'refexp' and multiply with sign on phy. normalized numbers
    x(:,normld) = s(ones(size(x,1),1),normld).*pow2(x(:,normld),refe(:,normld));
    % write back phy. denormalized numbers according to format
    if any(denorm)
      %  _____           _
      % |_   _|___    __| |  ___    _
      %   | | / _ \  / _` | / _ \  (_)
      %   | || (_) || (_| || (_) |  _
      %   |_| \___/  \__,_| \___/  (_)
      %
      % write back phy. denormalized numbers in correct form
      warning(sprintf(['Writing back physically denormalized numbers isn''t supported yet.\n',...
        'They have converted to zero.']))
      x(:,denorm) = 0;
    end
  end
  % make SEGNUM number of segments
end
x = x(1:SEGNUM,:);

% ----------------------------------------------------------------------------------------

% ROUNDING FUNCTIONS
function x = fixfix(s,x,rpinseg)
% simply truncate
%disp('fixfix')
x = pow2(fix(pow2(x,rpinseg)),-rpinseg);

function x = trunctrunc(s,x,rpinseg)
% simply truncate non-negative numbers
%disp('trunctrunc')
if s==1, x = pow2(fix(pow2(x,rpinseg)),-rpinseg);
elseif s==-1, x = pow2(ceil(pow2(x,rpinseg)),-rpinseg);
else error('Something wrong with sign')
end

function x = floorfloor(s,x,rpinseg)
% exploit the properties of built-in 'floor' function
%disp('floorfloor')
x = pow2(abs(floor(s.*pow2(x,rpinseg))),-rpinseg);

function x = ceilceil(s,x,rpinseg)
% exploit the properties of built-in 'ceil' function
%disp('ceilceil')
x = pow2(abs(ceil(s.*pow2(x,rpinseg))),-rpinseg);

function x = toinftoinf(s,x,rpinseg)
% exploit the properties of built-in 'ceil' function on always positive numbers
%disp('toinftoinf')
x = pow2(ceil(pow2(x,rpinseg)),-rpinseg);


function x = roundfix(s,x,rpinseg)
%disp('roundfix')
% eh is the indicator of 'exactly half' values
eh = ~rem(pow2(x,rpinseg+1),1);
% on 'eh' values use 'fix' function
x(eh) = pow2(fix(pow2(x(eh),rpinseg)),-rpinseg);
% not 'eh' values use 'round' function
x(~eh) = pow2(round(pow2(x(~eh),rpinseg)),-rpinseg);

function x = roundfloor(s,x,rpinseg)
%disp('roundfloor')
% eh is the indicator of 'exactly half' values
eh = ~rem(pow2(x,rpinseg+1),1);
% on 'eh' values use 'floor' function
x(eh) = pow2(abs(floor(s(eh).*pow2(x(eh),rpinseg))),-rpinseg);
% not 'eh' values use 'round' function
x(~eh) = pow2(round(pow2(x(~eh),rpinseg)),-rpinseg);

function x = roundceil(s,x,rpinseg)
%disp('roundceil')
% eh is the indicator of 'exactly half' values
eh = ~rem(pow2(x,rpinseg+1),1);
% on 'eh' values use 'ceil' function
x(eh) = pow2(abs(ceil(s(eh).*pow2(x(eh),rpinseg))),-rpinseg);
% not 'eh' values use 'round' function
x(~eh) = pow2(round(pow2(x(~eh),rpinseg)),-rpinseg);

function x = roundtoinf(s,x,rpinseg)
% exploit the properties of built-in 'round' function
% with exact 'roundtoinf' behaviour
%disp('roundtoinf')
x = pow2(round(pow2(x,rpinseg)),-rpinseg);

function x = roundtoeven(s,x,rpinseg)
% exploit the rounding properties of built-in co-processor
%disp('roundtoeven')
[f,e]=log2(x);
x=(x+pow2(52-rpinseg))-pow2(52-rpinseg);
%x = pow2(pow2(x,rpinseg-1074),1074-rpinseg); %Juhasz Laci

function x = roundtoodd(s,x,rpinseg)
%disp('roundtoodd')
% eh is the indicator of 'exactly half' values
eh = ~rem(pow2(x,rpinseg+1),1);
% not 'eh' values use 'round' function
x(~eh) = pow2(round(pow2(x(~eh),rpinseg)),-rpinseg);
% indicator of odd values before rounding point
odd = logical(fix(rem(pow2(x,rpinseg-1),1)*2));
% odd and exactly half values use 'fix'
x(odd & eh) = pow2(fix(pow2(x(odd & eh),rpinseg)),-rpinseg);
% even and exactly half values use 'ceil'
x(~odd & eh) = pow2(ceil(pow2(x(~odd & eh),rpinseg)),-rpinseg);
