function [y,of,uf,inv,warn,err] = str2seg(x,code)
%STR2SEG           Generates segments from strings containing numbers.
%   [Y,OF,UF,INV,WARN,ERR] = STR2SEG(X,CODE) reads a number from the string X
%      and returns the 22 bit wide segments of X in the two-row matrix Y.
%      Real part is in the 1st, imaginary part is in the 2nd row of Y.
%      The number of segments and other informations are taken by evaluating
%      the expression that is stored in the string CODE and coming from
%      the quantizer-manager core function 'quantizers'. It is in the form like
%      'TYPE='fixed-point';PRECISION=[22];<etc.>'.
%      STR2SEG performs rounding, overflow-, underflow-checking and other
%      operations before generating segments. The indication of denormalization and
%      re-rounding also will be performed.
%      If an error has occured during processing then ERR contains the error message
%      else ERR is empty. The same apply to WARN as a warning message.
%      The OF, UF and INV indicators will be set when the core function (quantizers)
%      must calculate and store statistic informations, otherwise they wouldn't be set
%      even if the corresponding event has occured.
%
%   STR2SEG is used by the qdata constructor function '@qdata/qdata.m'.
%
%   For format of numbers type 'help qformat'.

% Copyright (C) 2003,2004 by JUHASZ Laszlo

error(nargchk(2,2,nargin))

eval(code)

% initializing
y = zeros(2,SEGNUM);           % allocating zeros in output
of = 0;
uf = 0;
inv = 0;
warn='';
err='';
%.........................................................................................
%****************
% GENERAL CHECK *
%****************
% checking x
if isempty(x) | ~ischar(x) | prod(size(x))~=length(x) | size(x,1)~=1
   err = 'Empty or not character row vector input.';
   return
end

% make sure case insensitiveness
x = lower(x);

% checking x for valid characters
err = strchk(x,'+-.@ 0123456789abcdefxoijn/\()','number');
if ~isempty(err), return, end

% deblank x
x(x==' ') = '';
if isempty(x), err = 'Number contains only space(s).'; return, end

%.........................................................................................
%***************************************
%* SEPARATING REAL AND IMAGINARY PARTS *
%***************************************
num = {'0b','0b'};                     % for storing re & im parts
% temporary replace inf's with asterisk
x = strrep(x,'inf','***');
jpos = sort([findstr(x,'i'),findstr(x,'j')]);
% replace back inf's
x = strrep(x,'***','inf');
if isempty(jpos)
   num{1} = x;                         % pure real number, imag. part remains zero
elseif length(jpos)>1
   err = ['More than one imaginary separator ''',x(jpos),''' characters.'];
   return
elseif length(x)==1
   err = ['Number is only an ''',x,''' imaginary separator.'];
   return
elseif jpos==1 | jpos==length(x)
   num{2} = x(x~='i' & x~='j');        % pure imag. number, real part remains zero
else
   num{1} = x(1:jpos-1);    % complex number: string before 'i'/'j' will be real part
   num{2} = x(jpos+1:end);  % and string after 'i'/'j' will be imag. part
end

%disp(['after separating imag. from real REAL: ',num{1}])
%disp(['after separating imag. from real IMAG: ',num{2}])

%.........................................................................................
isanan = 0;                 % internal indicator of detected NaN
%*************************************
%* CHECK FOR NaN in REAL & IMAG PART *
%*************************************
if strcmp(num{1},'nan') | strcmp(num{1},'+nan') | strcmp(num{1},'-nan') | ...
      strcmp(num{2},'nan') | strcmp(num{2},'+nan') | strcmp(num{2},'-nan')
   isanan = 1;
   % check claim to error message
   if any(findstr(INVSIGNAL,'error'))
      err = 'Target format doesn''t include NaN.';
      return
   end
   % fill up output with NaN
   n = nan;
   y = n(1,ones(SEGNUM,1));
   % check claim to warning message
   if any(findstr(INVSIGNAL,'warning'))
      warn = 'NaN occurred during assignment.';
   end
   % check claim to stored stat info -> goes to quantizers
   if any(findstr(INVSIGNAL,'count'))
      inv = 1;
   end
   return
end

% check valid nan
if any(findstr(x,'nan')) & ~isanan
   err = 'Incorrectly given NaN.';
   return
end

%.........................................................................................
isainf = 0;                 % internal indicator of detected Inf
%******************************
%* CHECK FOR Inf IN REAL PART *
%******************************
t = num{1};                      % temp variable
if any(findstr(t,'inf'))
   % detect sign
   if strcmp(t,'-inf')
      sign = -1;
   elseif strcmp(t,'inf') | strcmp(t,'+inf')
      sign = 1;
   else
      % only one chance remains: the sign is after 'inf'
      err = 'Incorrectly given Inf.';
      return
   end
   isainf = 1;
   % check the fact of overflow - if Inf is allowed then no overflow has occured
   % in case of typing 'inf' in the command line
   if ~strcmp(OVERFLOW,'inf')
      % check claim to error message
      if any(findstr(OFSIGNAL,'error'))
         err = 'Inf occured during assignment.';
         return
      end
      % check claim to warning message
      if any(findstr(OFSIGNAL,'warning'))
         warn = 'Inf occured during assignment.';
      end
      % check claim to stored stat info -> goes to quantizers
      if any(findstr(OFSIGNAL,'count'))
         of = 1;
      end
   end
   % generating output - it can be Inf and it can be clipped too
   [mant,exp] = bitovfgen(sign,code);
   y(1,:) = str2seg_core(sign,mant,exp,SEGNUM);
   % no returning, the real part is Inf but imag. part can be 'normal' number
end

%.........................................................................................
%******************************
%* CHECK FOR Inf IN IMAG PART *
%******************************
t = num{2};                      % temp variable
if any(findstr(t,'inf'))
   % detect sign
   if strcmp(t,'-inf')
      sign = -1;
   elseif strcmp(t,'inf') | strcmp(t,'+inf')
      sign = 1;
   else
      % only one chance remainds: the sign is after 'inf'
      err = 'Incorrectly given Inf.';
      return
   end
   isainf = 1;
   % check the fact of overflow - if Inf is allowed then no overflow has occured
   % in case of typing 'inf' in the command line
   if ~strcmp(OVERFLOW,'inf')
      % check claim to error message
      if any(findstr(OFSIGNAL,'error'))
         err = 'Inf occured during assignment.';
         return
      end
      % check claim to warning message
      if any(findstr(OFSIGNAL,'warning'))
         warn = 'Inf occured during assignment.';
      end
      % check claim to stored stat info -> goes to quantizers
      if any(findstr(OFSIGNAL,'count'))
         of = 1;
      end
      % generating output - it can be only clipped
      [mant,exp] = bitovfgen(sign,code);
      y(2,:) = str2seg_core(sign,mant,exp,SEGNUM);
      % no returning, the imag. part is Inf -> clipped but real part 
      % can be a 'normal' number
   else
      % generating Inf output...
      % ***** WARNING *****
      % Inf in imaginary part produces NaN according to Matlab's behaviour,
      % so re-check for NaN's properties is needed
      % check claim to error message
      if any(findstr(INVSIGNAL,'error'))
         err = 'Target format doesn''t include NaN.';
         return
      end
      % fill up output with NaN
      n = nan;
      y = n(1,ones(SEGNUM,1));
      % check claim to warning message
      if any(findstr(INVSIGNAL,'warning'))
         warn = 'NaN occured during assignment.';
      end
      % check claim to stored stat info -> goes to quantizers
      if any(findstr(INVSIGNAL,'count'))
         inv = 1;
      end
      return
   end
end

% check valid Inf
if any(findstr(x,'inf')) & ~isainf
   err = 'Incorrectly given Inf';
   return
end

% selecting 'zero yet' rows of 'y'
for k = find(~any(y'))
   x = num{k};
   % in what follows 'x' is either real or imaginary part
   %......................................................................................
   %************************************
   %* SEPARATING MANTISSA AND EXPONENT *
   %************************************
   epos = findstr(x,'@');
   if isempty(epos)                       % no '@' char
      mant = x; exp = '0b';
   elseif length(epos)>1
      err = 'More than one exponent separator ''@'' characters.';
      return
   elseif length(x)==1
      err = 'Missing mantissa and exponent.';
      return
   elseif epos==1
      err = 'Missing mantissa.';
      return
   elseif epos==length(x)
      err = 'Missing exponent.';
      return
   else
      mant = x(1:epos-1);                 % mantissa is the chars before '@'
      exp = x(epos+1:end);                % exponent is the chars after '@'
   end
   
   %*******************************************************
   %* CHECKING MANTISSA AND EXPONENT FOR VALID CHARACTERS *
   %*******************************************************
   err = strchk(mant,'+-.0123456789abcdefxo/\()','mantissa');
   if ~isempty(err), return, end
   err = strchk(exp,'+-0123456789abcdefxo/\','exponent');
   if ~isempty(err), return, end
   
   %disp(['after separating mant:',mant])
   %disp(['after separating  exp:',exp])
   
   % process mantissa
   mdig=sum((mant=='0')|(mant=='1'));
   if any(findstr(CODING,'s-comp'))&(mdig==PRECISION+1)&strcmp(mant(1),'1')
     %if strcmp(mant(end),'b'), mant(end)=''; end
     %mant(1)='0'; 
     ind0=find(mant=='0'); ind1=find(mant=='1'); 
     mant(ind0)=mant(ind0)+1; mant(ind1)=mant(ind1)-1;
     mant(end-1)=mant(end-1)+1;
     for ii=length(mant)-1:-1:1
       if strcmp(mant(ii),'2')
         mant(ii)='0'; 
         if (ii>1)&~strcmp(mant(ii-1),'.'), mant(ii-1)=mant(ii-1)+1; 
         elseif (ii>2), mant(ii-2)=mant(ii-2)+1; 
         end, 
       elseif ~strcmp(mant(ii),'.')
         break
       end
     end
     [sign,mant,expcorr,err] = procmant(mant);
     sign=-1;
   else
     [sign,mant,expcorr,err] = procmant(mant);
   end
   if ~isempty(err), return, end
   % process exponent
   [exp,err] = procexp(exp);
   if ~isempty(err), return, end
   % calculate correction
   exp = exp+expcorr;
   
   %disp(['after processing mant:',char(mant+'0')])
   %disp(['after processing  exp:',num2str(exp)])
   
   %**************************
   %* FILTER OUT ZERO VALUES *
   %**************************
   if ~any(mant)
      mant = 0;
      exp = 0;
   else
      %***********************************
      %* COLLECT STATISTICAL INFORMATION *
      %***********************************
      locof = 0;                             % local indicator of: overflow
      locuf = 0;                                                 % underflow
      locden = 0;                                                % denormalization
      
      switch TYPE
      case 'floating-point'
         % ****** FLOATING-POINT NUMBERS ******
         % ROUNDINGDIR
         [mant,exp] = bitround(sign,mant,exp,FRACTION+1,ROUNDINGDIR,TREATHALF);
         % check for OVERFLOW
         % filtering out negative 'twos-comp' numbers
         if (sign==-1) & strcmp(CODING,'twos-comp')
            if length(mant)>1             % more than 1 bit wide mantissa
               if exp>EMAX                % usual condition
                  locof = 1;
               end
            elseif length(mant)==1        % only 1 bit wide mantissa
               if exp>EMAX+1              % by 1 bit mantissa 1@emax+1 is allowed
                  locof = 1;
               end
            else
               error('Internal error.')
            end
         else
            % not negative 'twos-comp' number
            if exp>EMAX
               locof = 1;
            end
         end
         % handling of OVERFLOW
         if locof                            % overflow occured
            % generating owerfloved bit pattern with 'bitovfgen'
            [mant,exp] = bitovfgen(sign,code);
            % check claim to error message
            if any(findstr(OFSIGNAL,'error'))
               err = 'Overflow occured during assignment.';
               return
            end
            % check claim to warning message
            if any(findstr(OFSIGNAL,'warning'))
               warn = 'Overflow occured during assignment.';
            end
            % check claim to stored stat info -> goes to quantizers
            if any(findstr(OFSIGNAL,'count'))
               of = 1;
            end
         end
         % check for DENORMALIZED
         % filtering out negative 'twos-comp' numbers
         if (sign==-1) & strcmp(CODING,'twos-comp')
            if exp==EMIN & length(mant)==1 & all(mant)
               % -1@emin is denormalized yet by twos'comp
               locden = 1;
            else
               if exp<EMIN
                  locden = 1;
               end
            end
         else
            if exp<EMIN
               locden = 1;
            end
         end
         % check for UNDERFLOW
         if exp<EMIN-FRACTION
            locuf = 1;
         end
         % make sure that 'flush' numbers would underflow
         if locden & strcmp(UNDERFLOW,'flush')
            locuf = 1;
         end
         % handling UNDERFLOW
         if locuf
            mant = 0;
            exp = 0;
            % check claim to error message
            if any(findstr(UFSIGNAL,'error'))
               err = 'Underflow occured during assignment.';
               return
            end
            % check claim to warning message
            if any(findstr(UFSIGNAL,'warning'))
               warn = 'Underflow occured during assignment.';
            end
            % check claim to stored stat info -> goes to quantizers
            if any(findstr(UFSIGNAL,'count'))
               uf = 1;
            end
         end
         % handling of DENORMALIZED
         if locden & ~locuf
            % re-rounding
            if (sign==-1) & strcmp(CODING,'twos-comp')
               % filtering out negative 'twos-comp'
               if exp==EMIN & length(mant)==1 & all(mant)
                  [mant,exp] = bitround(sign,mant,exp,FRACTION,ROUNDINGDIR,TREATHALF);
               else
                  [mant,exp] = bitround(sign,mant,exp,...
                     exp-EMIN+FRACTION+1,ROUNDINGDIR,TREATHALF);
               end
            else
               % re-rounding
               [mant,exp] = bitround(sign,mant,exp,...
                  exp-EMIN+FRACTION+1,ROUNDINGDIR,TREATHALF);
            end
         end
      case 'fixed-point'
         % ****** FIXED-POINT NUMBERS ******
         % ROUNDINGDIR
         [mant,exp] = bitround(sign,mant,exp,exp+FRACTION+1,ROUNDINGDIR,TREATHALF);
         switch CODING
         case 'unsign'                                % unsigned numbers
            % generating EMAX and EMIN in order that we can handle fixed-point
            % numbers in a similar way to floating-point numbers
            EMAX = 0; %PRECISION-FRACTION-1;
            EMIN = 0; %-FRACTION;
            % check & handle of OVERFLOW
            if exp>EMAX
               locof = 1;
               switch OVERFLOW
               case 'clip'
                  mant = repmat(1,1,PRECISION);
                  exp = EMAX;
               otherwise
                  error(['Unsupported OVERFLOW ',OVERFLOW])
               end
               % check & handle of UNDERFLOW
            elseif ~any(mant) | mantsign
               % negative numbers will also underflow in this representation
               locuf = 1;
               switch UNDERFLOW
               case 'flush'
                  mant = 0;
                  exp = 0;
               otherwise
                  error(['Unsupported UNDERFLOW ',UNDERFLOW])
               end
            end
         case 'ones-comp'                                % ones-complement numbers
            % generating EMAX and EMIN in order that we can handle fixed-point
            % numbers in a similar way to floating-point numbers
            EMAX = 0; %PRECISION-FRACTION-2;
            EMIN = 0; %-FRACTION;
            % check & handle of OVERFLOW
            if exp>EMAX
               locof = 1;
               switch OVERFLOW
               case 'clip'
                  mant = repmat(1,1,PRECISION-1);
                  exp = EMAX;
               otherwise
                  error(['Unsupported OVERFLOW ',OVERFLOW])
               end
               % check & handle of UNDERFLOW
            elseif ~any(mant)
               locuf = 1;
               switch UNDERFLOW
               case 'flush'
                  mant = 0;
                  exp = 0;
               otherwise
                  error(['Unsupported UNDERFLOW ',UNDERFLOW])
               end
            end
         case 'twos-comp'                                % twos-complement numbers
            % generating EMAX and EMIN in order that we can handle fixed-point
            % numbers in a similar way to floating-point numbers
            EMAX = PRECISION-FRACTION;
            EMIN = -FRACTION;
            % check & handle of OVERFLOW
            % filtering out negative numbers
            if exp>EMAX & ~(sign==-1 & exp==EMAX+1 & sum(mant)==1 & mant(1))
               locof = 1;
               switch OVERFLOW
               case 'clip'
                  if sign==1                             % positive numbers: 0111...
                     mant = repmat(1,1,PRECISION-1);
                     exp = EMAX;
                  else                                   % negative numbers: 1000...
                     mant = 1;
                     exp = EMAX+1;
                  end
               otherwise
                  error(['Unsupported OVERFLOW ',OVERFLOW])
               end
               % check & handle of UNDERFLOW
            elseif ~any(mant)
               locuf = 1;
               switch UNDERFLOW
               case 'flush'
                  mant = 0;
                  exp = 0;
               otherwise error(['Unsupported UNDERFLOW ',UNDERFLOW])
               end
            end
         otherwise error(['Unsupported CODING ',CODING])
         end
         % handle overflow messages
         if locof                            % overflow occured
            % check claim to error message
            if any(findstr(OFSIGNAL,'error'))
               err = 'Overflow occured during assignment.';
               return
            end
            % check claim to warning message
            if any(findstr(OFSIGNAL,'warning'))
               warn = 'Overflow occured during assignment.';
            end
            % check claim to stored stat info -> goes to quantizers
            if any(findstr(OFSIGNAL,'count'))
               of = 1;
            end
         end
         % handle underflow messages
         if locuf
            mant = 0;
            exp = 0;
            % check claim to error message
            if any(findstr(UFSIGNAL,'error'))
               err = 'Underflow occured during assignment.';
               return
            end
            % check claim to warning message
            if any(findstr(UFSIGNAL,'warning'))
               warn = 'Underflow occured during assignment.';
            end
            % check claim to stored stat info -> goes to quantizers
            if any(findstr(UFSIGNAL,'count'))
               uf = 1;
            end
         end
      otherwise error(['Unsupported TYPE ',TYPE])
      end
   end         % end of 'if ~any(mant)'
   
   %******************
   %* WRITE SEGMENTS *
   %******************
   y(k,:) = str2seg_core(sign,mant,exp,SEGNUM);
end         % end of 'for k = 1:2'

% finally, compress two rows of 'y' into one row
y = y(1,:)+j*y(2,:);
