function y = str2seg_core(sign,mant,exp,segnum)
%STR2SEG_CORE      Core function for str2seg.
%  Usage: Y = STR2SEG_CORE(SIGN,MANT,EXP,SEGNUM)
%  Input arguments:
%     SIGN   - (+1) - positive, (-1) - negative
%     MANT   - mantissa row vector with ones and zeros. Must be normalized (1st bit is 1)
%     EXP    - exponent of the first bit of mantissa or infinity
%     SEGNUM - number of 22bit segments
%              Rule: MANT can't be greater than SEGNUM*22
%  Output argument:
%     Y - output row vector with segments. Length of Y is SEGNUM.
%
%  STR2SEG can't produce NaN.

% Copyright (C) 2004 by JUHASZ Laszlo

error(nargchk(4,4,nargin))

% check inputs
if isempty(sign) | isempty(mant) | isempty(exp) | isempty(segnum) | ...
      ~isa(sign,'double') | ~isa(exp,'double') | ~isa(mant,'double') | ...
      ~isa(segnum,'double') | prod(size(sign))~=1 | size(mant,1)~=1 | ...
      prod(size(exp))~=1 | prod(size(segnum))~=1 | ~isfinite(sign) | ...
      ~all(isfinite(mant)) | isnan(exp) | ~isfinite(segnum) | ...
      rem(sign,1) | ~(sign>=-1 & sign<=1) | any(mant~=(~~mant)) | ...
      (~isinf(exp) & rem(exp,1)) | rem(segnum,1)
   error('Internal error. Bad arguments.')
end

if any(mant) & ~mant(1)                   % check mant
   error('Internal error. Zero at start of mantissa.')
elseif isinf(exp) & any(mant)             % check of Inf
   % generating +Inf exponent and +/-Inf segments with sign according to 'sign'
   y = repmat(sign*Inf,1,segnum);
   return
elseif isinf(exp)
   error('Internal error. Inf exponent with zero mantissa.')
end

% check length of mant
if length(mant)>(22*segnum)
   error('Internal error. Mant is too long.')
end

% check of zero
if ~any(mant)
   % the representation of zero is 'segnum' piece of zero segments
   y = repmat(0,1,segnum);
   return
end

% enlarge mant minimum to segnum
mant(22*segnum+1) = 0;

if exp>=-1053                                % magic limit of internal denormalization
   % fill up output - actual algorithm
   for k = 1:segnum
      % actual piece of mant
      m = mant((k-1)*22+1 : k*22);
      % transform and store
      y(k) = sign*pow2(bindec(m),exp-21);
   end
else
   % internally denormalized numbers
   for k = 1:segnum
      % how many bits is in the actual segment
      bits = bitinseg(exp,k);
      % actual piece of mant
      m = mant(1:bits);
      if isempty(m)
         % the segment is zero in actual condition
         y(k) = 0;
      else
         % clear actual piece of mant from the start of mant
         mant(1:bits) = [];
         % transform and store
         y(k) = sign*pow2(bindec(m),-bits+1+exp+22*(k-1));
      end
   end
end
