function [y,warn,err] = str2seg_kp(x,code) %of,uf,inv,
%STR2SEG           Generates segments from strings containing numbers.
%   [Y,OF,UF,INV,WARN,ERR] = STR2SEG(X,CODE) reads a number from the string X
%      and returns the 22 bit wide segments of X in the two-row matrix Y.
%      Real part is in the 1st, imaginary part is in the 2nd row of Y.
%      The number of segments and other informations are taken by evaluating
%      the expression that is stored in the string CODE and coming from
%      the quantizer-manager core function 'quantizers'. It is in the form like
%      'TYPE='fixed-point';PRECISION=[22];<etc.>'.
%      STR2SEG performs rounding, overflow-, underflow-checking and other
%      operations before generating segments. The indication of denormalization and
%      re-rounding also will be performed.
%      If an error has occured during processing then ERR contains the error message
%      else ERR is empty. The same apply to WARN as a warning message.
%      The OF, UF and INV indicators will be set when the core function (quantizers)
%      must calculate and store statistic informations, otherwise they wouldn't be set
%      even if the corresponding event has occured.
%
%   STR2SEG_KP is used by the qdata constructor function '@qdata/qdata.m'.
%
%   For format of numbers type 'help qformat'.

% Copyright (C) 2003,2004 by JUHASZ Laszlo

error(nargchk(2,2,nargin))

eval(code)

% initializing
y = zeros(2,SEGNUM);           % allocating zeros in output
of = 0;
uf = 0;
inv = 0;
warn='';
err='';
%.........................................................................................
%****************
% GENERAL CHECK *
%****************
% checking x
if isempty(x) | ~ischar(x) | prod(size(x))~=length(x) | size(x,1)~=1
   err = 'Empty or not character row vector input.';
   return
end

% make sure case insensitiveness
x = lower(x);

% checking x for valid characters

err = strchk_kp(x,'+-. 0123456789()[d]*^','number');
if ~isempty(err), return, end

% deblank x
x(x==' ') = '';
if isempty(x), err = 'Number contains only space(s).'; return, end


%.........................................................................................
%***************************************
%* SEPARATING REAL AND IMAGINARY PARTS *
%***************************************
num = {'0b','0b'};                     % for storing re & im parts
% temprary replace inf's with asterisk
x = strrep(x,'inf','***');
jpos = sort([findstr(x,'i'),findstr(x,'j')]);
% replace back inf's
x = strrep(x,'***','inf');
if isempty(jpos)
   num{1} = x;                         % pure real number, imag. part remains zero
elseif length(jpos)>1
   err = ['More than one imaginary separator ''',x(jpos),''' characters.'];
   return
elseif length(x)==1
   err = ['Number is only an ''',x,''' imaginary separator.'];
   return
elseif jpos==1 | jpos==length(x)
   num{2} = x(x~='i' & x~='j');        % pure imag. number, real part remains zero
else
   num{1} = x(1:jpos-1);    % complex number: string before 'i'/'j' will be real part
   num{2} = x(jpos+1:end);  % and string after 'i'/'j' will be imag. part
end

%disp(['after separating imag. from real REAL: ',num{1}])
%disp(['after separating imag. from real IMAG: ',num{2}])

%.........................................................................................
isanan = 0;                 % internal indicator of detected NaN


%.........................................................................................
isainf = 0;                 % internal indicator of detected Inf


   % in what follows 'x' is either real or imaginary part
   %......................................................................................
   %************************************
   %* SEPARATING MANTISSA AND EXPONENT *
   %************************************
    
   epos = findstr(x,'*2^');
   
   
   if isempty(epos)                            % no '*2^' char
      mant = x; exp = '0b';
   elseif length(epos)>1
      err = 'More than one exponent separator ''*2^'' characters.';
      return
   elseif length(x)==1
      err = 'Missing mantissa and exponent.';
      return
   elseif epos==1
      err = 'Missing mantissa.';
      return
   elseif epos==length(x)
      err = 'Missing exponent.';
      return
   else
      mant = x(1:epos-1);                 % mantissa is the chars before '*2^'
      exp = x(epos+1:end);                % exponent is the chars after '*2^'
   end
   
   %*******************************************************
   %* CHECKING MANTISSA AND EXPONENT FOR VALID CHARACTERS *
   %*******************************************************
   err = strchk(mant,'+-.01/\()','mantissa');
   if ~isempty(err), return, end
   err = strchk(exp,'0123456789d/\[]^()','exponent');
   if ~isempty(err), return, end
   
   
   % process mantissa
   [egesz,tort,err] = procmant_kp(mant);
   if ~isempty(err), return, end
   % process exponent
   [exp,err] = procexp_kp(exp);
   if ~isempty(err), return, end
   % calculate correction
   %egesz
   %tort
   %exp
   
   
   %**************************
   %* FILTER OUT ZERO VALUES *
   %**************************
   if ~any(mant)
      mant = 0;
      exp = 0;
   else
      %***********************************
      %* COLLECT STATISTICAL INFORMATION *
      %***********************************
      locof = 0;                             % local indicator of: overflow
      locuf = 0;                                                 % underflow
      locden = 0;                                                % denormalization
      
      switch TYPE
      case 'floating-point'
      
        switch CODING
         case 'sign-mag'                                
            if strcmp(egesz(1),'+')
                sign=1;
            elseif strcmp(egesz(1),'-')
                sign=-1;
            else err = 'Wrong sign in mantissa.'; return;
            end;    
             
            egesz2=0;
            
            for i=2:length(egesz)
                if (strcmp(egesz(i),'1')) egesz2 = egesz2+1*2^(length(egesz)-i);
                end
            end
            
            %egesz=egesz2
            
            %disp('                              ')
            %disp('Az egszrsz szm formtumban:')
            %disp('                              ')
            
            egesz_number=egesz2%
            
            tort2=0;
            
            for i=1:length(tort)
                if (strcmp(tort(i),'1')) tort2=tort2+1*2^(-1*i);
                end
            end
            
            %tort=tort2
            %disp('                              ')
            %disp('Az trt rsz szm formtumban:')
            %disp('                              ')
            
            tort_number=tort2%
            
            
            %disp('                              ')
            %disp('Az exponens szm formtumban:')
            %disp('                              ')
            
            %exp=str2num(exp(1:end-1))
            exp_number=str2num(exp(1:end-1));
            
        end
      end
   end
  
   %sprintf('%d d @ %d',egesz+tort,exp)
   %end

   
   % A cell formtum mantissza:
   mant_cell=num2str(sign*(egesz_number+tort_number));
   
   % A cell formtum exponens:
   exp_cell=num2str(exp_number);
   
   %disp('                   ');
   %disp('A string formtum mantissza:')
   %disp('                   ');
   
   mant_str=cellstr(mant_cell);
   
   %disp('                   ');
   %disp('A string formtum exponens:')
   %disp('                   ');
   exp_str=cellstr(exp_cell);
   
   y=[mant_cell,' d @ ',exp_cell,' d'];
    
end
%