function [y,ovfind] = qdata(varargin)
%QDATA  Constructor function of the @qdata quantized object.
%
%   For information about the format of numbers that the constructor can handle, type
%     'help qformats'.
%   For the help on QDSP, type 'help qdsp' or 'qdspstart',
%   for a help on the methods, type 'help(qdata)'.
%
%   Usage (QSTR is a string, the name of quantizer):
%   1) Calling with double scalar or array
%      Y = QDATA(D,QSTR) where D is a double scalar or array. The size of resulting object
%      will be the same as D. Multidimensional double arrays are supported as well.
%   2) Calling with a string
%      Y = QDATA(S,QSTR) where S is a string such as '2x'. The resulting object will be a scalar.
%      See 'help qformats'. If the string is missing, double precision will be assumed. Empty
%      string causes to extract a quantizer 'FROMSTRING' from the string
%   3) calling with cell array of strings
%      Y = QDATA(C,QSTR) where C is a cell array of strings
%      such as {'1x','2x';'3x','4x'}.
%      The size of resulting object will be the same as C.
%      Multidimensional cell arrays are supported as well.
%   4) calling with a qdata object
%      Y = QDATA(QOBJ,QSTR) where QOBJ is a qdata object
%      QSTR is the new quantizer
%   5) calling with a fi object
%      Y = QDATA(F,QSTR) where F is a fi object
%      QSTR is the new quantizer
%      Y = QDATA(F) uses the data of the fi numerics
%
%   Default (always existing) quantizers: double, single, BigFP/BFP, Q16bit, Q32bit, Q64bit
%
%   See also QDATA/DISPLAY, QDATA/PLUS, QDATA/TIMES.

%   Written by Lszl Juhsz, 2003-2004.
%   Modified by Istvn Kollr, 12-Apr-2007

%   6) Calling with internal format: data
%      Y = QDATA(qd,QSTR,'data') where qd is data of a qdsp object
%      QSTR is the new quantizer

warn='';
err='';
if nargin==0
  % called without argument
  y = emptyclass;
  y = class(y,'qdata');
  return
end

if nargin>=1
  if isa(varargin{1},'double')
    if length(varargin)==1, varargin{1,2}='double';, end
    if any(imag(varargin{1}(:)))
      y=qdata(real(varargin{1}),varargin{2})+j*qdata(imag(varargin{1}),varargin{2}); return
    end
  elseif isa(varargin{1},'qdata')
    if length(varargin)>1
      if any(double(imag(varargin{1})))
        y=qdata(real(varargin{1}),varargin{2})+j*qdata(imag(varargin{1}),varargin{2}); return
      end
    end
  elseif isa(varargin{1},'embedded.fi')
    x=varargin{1};
    B=x.wordlength; fb=x.FractionLength;
    roundmode=x.roundmode;
    if strcmp(roundmode,'convergent'), roundingdir='round'; treathalf='toeven';
    elseif strcmp(roundmode,'round'), roundingdir='round'; treathalf='toinf';
    elseif strcmp(roundmode,'fix'), roundingdir='fix'; treathalf='fix';
    elseif strcmp(roundmode,'floor'), roundingdir='floor'; treathalf='floor';
    elseif strcmp(roundmode,'ceil'), roundingdir='ceil'; treathalf='ceil';
    elseif strcmp(roundmode,'nearest'), roundingdir='round'; treathalf='ceil';
    end
    if ~isquant('Qfi')
      Qfi=qmake('name','Qfi','type','fixed-point','Bits',B,'fractionbits',fb,...
        'roundingdir',roundingdir,'treathalf',treathalf);
    else
      Qfitmp=qmake('name','Qfitmp','type','fixed-point','Bits',B,'fractionbits',fb,...
        'roundingdir',roundingdir,'treathalf',treathalf);
      if isqequal('Qfi','Qfitmp')
        Qfi='Qfi';
      else
        Qname='Qfi'; Qi=2;
        while isquant([Qname,num2str(Qi)])
          if isqequal('Qfitmp',[Qname,num2str(Qi)]), Qfi=[Qname,num2str(Qi)]; break
          else Qi=Qi+1;
          end
        end
        if ~isquant([Qname,num2str(Qi)])
          Qfi=qmake('name',[Qname,num2str(Qi)],'type','fixed-point','Bits',B,'fractionbits',fb,...
            'roundingdir',roundingdir,'treathalf',treathalf);
        end
      end
      qclear('Qfitmp')
    end
    b=x.bin;
    %two's complement
    if strcmp(b(1),'1')
      ind=find(b=='0');
      b=b-1; if ~isempty(ind), b(ind)=b(ind)+2; end
      b(end)=b(end)+1;
      for bi=length(b):-1:2
        if b(bi)=='2'+0, b(bi)='0'; b(bi-1)=b(bi-1)+1; else break, end
        %if b(1)=='1'+0, b=['10',b(2:end)]; end
      end
      b=['-',setstr(b)];
    end
    FRACTION=x.FractionLength;
    if FRACTION>0
      if FRACTION>length(b), b=[setstr('0'*ones(1,FRACTION-length(b))),b]; end
      b=[b(1:end-FRACTION),'.',b(end-FRACTION+1:end)];
    end
    b=[b,'b'];
    y=qdata(b,Qfi);
    if (nargin>1)&isquant(varargin{2})
      y=qdata(y,varargin{2}); %varargin(2)=[];
    end
    return
  end %embedded-fi
end %nargin>1

if nargin==1
  name=varargin{1};
  if isstruct(name)
    %name-ban az adatok amit mg kvantlni kell
    [code,err] = qdspdemo('quantizers','get',name.name); error(err);
    [code,err] = qdspdemo('quantizers',name.name); error(err);
    a = name.data;
    orsiza = size(a);
    SEGNUM = orsiza(1);
    orsiza = orsiza(2:end);
    a = reshape(a,[SEGNUM,prod(orsiza)]);
    eval(code);
    [re_name,of,uf,in,warn,err] = seground(real(a),code);
    [im_name,of,uf,in,warn,err] = seground(imag(a),code);
    a = re_name + j*im_name;
    if(isdscalar(orsiza))
      orsiza = [orsiza 1];
    end
    a = reshape(a,[max(size(a,1),SEGNUM),orsiza]); %*** javitas...
    if strcmp(OPERATION,'quantize')
      name.data = a;
    else %PQN
      if ~isequal(a,name.data)
        q=lsb(name,qdata);
        noise=qdata(q*rand(1,1),'BigFP');
        y=class(name,'qdata');
        xq=struct(noise+y); xq.name=name.name;
        name=xq;
        %error('Not yet ready...')
      end
    end
    y=class(name,'qdata');
    return
  end
end %nargin==1

if (nargin==3)&isstr(varargin{end})&strcmp(varargin{end},'data'), isdata=1; varargin(end)=[];
else isdata=0;
end
%Make name
if (isquant(varargin{1})|isempty(varargin{1}))&~isquant(varargin{end})
  %old call
  name=varargin{1}; varargin(1)=[];
  warning('Old-form call is detected: qdata(qname,var). Change to qdata(var,qname).')
elseif isquant(varargin{end})
  name=varargin{end}; varargin(end)=[];
elseif nargin==1
  [code,err] = qdspdemo('quantizers','get','-all'); error(err);
  if isa(varargin{1},'qdata'), name=varargin{1}.name;
  else name='double';
  end
elseif (nargin==2)&isstr(varargin{1})&isstr(varargin{2})&isempty(varargin{2})
  qmake('FROMSTRING',varargin{1});
  name='FROMSTRING'; varargin(2)=[];
%elseif (nargin==2)&isstr(varargin{2})&strcmp(varargin{2},'BigFP')&~isquant('BigFP')
	%  qmake('name','BigFP','precision',110);
%  name='BigFP';
%elseif (nargin==2)&isstr(varargin{2})&strcmp(varargin{2},'BFP')&~isquant('BFP')
	%  qmake('name','BFP','precision',110);
%  name='BFP';
elseif nargin==2
  name=varargin{2}; varargin(2)=[];
end

% obtaining quantizer properties
% the 'get' function of 'quantizers' has error correction so double call required
[code,err] = qdspdemo('quantizers','get',name); error(err)
% name is OK, we can call the unchecked fast version of function
[code,err] = qdspdemo('quantizers',name); error(err)
eval(code);

if (nargin==1)&isquant(varargin{1})
  % called with one argument - the quantizer name without data
  y = class(emptyclass,'qdata');
  return
end

if (nargin>2)&~isdata
  % only chance to use multiple arguments is calling with multiple strings
  for k = 1:length(varargin)
    if ~ischar(varargin{k})
      error('Multiple arguments must be all strings.')
    elseif prod(size(varargin{k}))~=length(varargin{k})
      % filtering out strings with e.g. line-break
      error('Multiple strings must be all simple.')
    end
  end
  for k = 1:length(varargin)
    if isempty(varargin{k})
      warning(['Size of input has been probably modified ',...
        'due to deleted empty element.'])
      % delete empty strings
      varargin(k) = [];
    end
  end
  % the whole 'varargin' goes to 'v'
  v = varargin;
  % save size
  siz = size(v);
else
  % chances are: only one string, cell array of strings, double scalar or array, 'data'
  if isempty(varargin)
    % if the only argument is an empty cell or something
    y = class(emptyclass,'qdata');
    return
  elseif prod(size(varargin))~=1
    error('Internal error 1')
  end
  % the "inside" of 'varargin' goes to 'v'
  v = varargin{1};
  switch class(v)
    case 'char'
      if prod(size(v))~=length(v)
        % filtering out string with e.g. line-break
        error('String must be simple.')
      end
      siz = [1 1];
      % convert to cell
      v = {v};
    case 'double'
      siz = size(v);
    case 'cell'
      for k = 1:prod(size(v))
        if ~ischar(v{k})
          error('Cell matrices can contain only strings.')
        elseif prod(size(v{k}))~=length(v{k})
          % filtering out strings with e.g. line-break
          error('Strings in cell must be all simple.')
        end
      end
      for k = 1:prod(size(v))
        if isempty(v{k})
          if prod(size(v))~=length(v)
            % if the cell array isn't a vector then we couldn't delete the
            % empty element without upsetting the layout of the array
            error('Non-vector cell array contains empty element.')
          else
            warning(['Size of input has been probably modified ',...
              'due to deleted empty element.'])
          end
          % delete empty strings
          v(k) = [];
        end
      end
      siz = size(v);
    case 'qdata' %reassign quantizer
      if nargin==1, y=qdata(get(v,'data'),get(v,'name'),'data'); return
      elseif (nargin==2)&isstr(name)&isquant(name)
        y=qdata(get(v,'data'),name,'data');
        return
      else
        error(['Cannot reassign new quantizer ''',varargin{2},''' to qdata'])
      end
    otherwise
      error(['Unsupported input type ''',class(v),'''.'])
  end
end

% if 'v' has contained only empty chars/doubles and these have been erased...
if isempty(v)
  y = class(emptyclass,'qdata');
  return
end

if isdata
  %'data' given
  siz=siz(2:end); if length(siz)==1, siz=[siz,1]; end
  vsiz=size(v); vlen=prod(vsiz(2:end));
  % initialize statistic variables
  ofs = zeros(siz);         % overflow index vector
  ufs = zeros(siz);         % underflow index vector
  ins = zeros(siz);         % 'unallowed NaN' index vector
  for k = 1:vlen
    t=v(:,k);
    % rounding and collecting stat info - real part
    s=sign(real(t(1))); ta=abs(real(t));
    if length(ta)>1
      [f1,e1]=log2(ta(1));
      t1=pow2(fix(f1*2^22)/2^22,e1);
      ta(2)=ta(2)+ta(1)-t1;
      ta(1)=t1;
      for di=2:length(ta)-1
        [fi,ei]=log2(ta(di));
        ti=pow2(fix(fi*2^(22-e1+ei))/2^(22-e1+ei),ei);
        ta(di+1)=ta(di+1)+ta(di)-ti;
        ta(di)=ti;
      end
      t=s*ta+j*imag(t);
    end
    [re,of,uf,in,warn,err] = seground(real(t),code);
    if ~isempty(err), error(['Error for number ',num2str(v(1,k)),': ', err]), end
    if ~isempty(warn),warning(['For number ',num2str(v(1,k)),': ',warn]), end
    % saving of, uf and inv values into 'ofs', 'ufs' and 'ins' index vectors
    ofs(k) = ofs(k)|of;
    ufs(k) = ufs(k)|uf;
    ins(k) = ins(k)|in;
    % rounding and collecting stat info - imag part
    s=sign(imag(t(1))); ta=abs(imag(t));
    if length(ta)>1
      [f1,e1]=log2(ta(1));
      t1=pow2(fix(f1*2^22)/2^22,e1);
      ta(2)=ta(2)+ta(1)-t1;
      ta(1)=t1;
      for di=2:length(ta)-1
        [fi,ei]=log2(ta(di));
        ti=pow2(fix(fi*2^(22-e1+ei))/2^(22-e1+ei),ei);
        ta(di+1)=ta(di+1)+ta(di)-ti;
        ta(di)=ti;
      end
      t=real(t)+j*s*ta;
    end
    [im,of,uf,in,warn,err] = seground(imag(t),code);
    if ~isempty(err),   error(['Error for number ',num2str(v(1,k)),': ', err]), end
    if ~isempty(warn),warning(['For number ',num2str(v(1,k)),': ',warn]), end
    % saving of, uf and inv values into 'ofs', 'ufs' and 'ins' index vectors
    ofs(k) = ofs(k)|of;
    ufs(k) = ufs(k)|uf;
    ins(k) = ins(k)|in;
    % put back complex data into 'z'
    % transpose operator is needed instead of ctranspose
    z(:,k) = (re+j*im);
  end
else
  % convert double arrays to colum-vector and cell arrays to column-cell
  % initialize statistic variables
  ofs = zeros(size(v));         % overflow index vector
  ufs = zeros(size(v));         % underflow index vector
  ins = zeros(size(v));         % 'unallowed NaN' index vector
  v = v(:);
  % initialize output
  z = zeros(SEGNUM,length(v));
  switch class(v)
    case 'cell'
      for k = 1:length(v)

        nyito = findstr(v{k},'[');
        csuko = findstr(v{k},']');
        hatv1 = findstr(v{k},'(*2^)');
        hatv2 = findstr(v{k},'@');
        hatv3 = findstr(v{k},'*2^');
        van_x = findstr(v{k},'x');
        van_o = findstr(v{k},'o');
        van_b = findstr(v{k},'b');
        van_d = findstr(v{k},'d');

        if (~isempty(nyito) & ~isempty(csuko))
          if (isempty(hatv1) & ~isempty(hatv2))
            err = 'Missing raising to the second power.';
            return
          end
        end

        if (isempty(nyito) & ~isempty(csuko))
          err = 'Missing ''['' character.';
          return
        end

        if (~isempty(nyito) & isempty(csuko))
          err = 'Missing '']'' character.';
          return
        end

        if ((isempty(nyito) & isempty(csuko) & isempty(hatv3) ))
          if any(ismember(v{k}(1),'to')), v{k}(1)='1'; end
          [segs,of,uf,inv,warn,err] = str2seg(v{k},code);
          if ~isempty(err), error(['At the number ''',v{k},''': ', err]), end
          if ~isempty(warn), warning(['At the number ''',v{k},''': ',warn]), end
          % saving of, uf and inv values into 'ofs', 'ufs' and 'ins' index vectors
          ofs(k) = of;
          ufs(k) = uf;
          ins(k) = inv;
          % put back complex data into 'z'
          % transpose operator (.') is needed instead of ctranspose (')
          z(:,k) = segs.';
        end

        if (~isempty(nyito) & ~isempty(csuko) & ~isempty(hatv1))
          try
            v2=cellstr(v);
            v3=char(v2);
            bin2numv3=bin2num(v3,'double');
            w=0;
            hatv1;
            szam='';
            for i=1:hatv1-1
              if (strcmp(v3(i),'1')) szam = [szam,'1']; end
              if (strcmp(v3(i),'0')) szam = [szam,'0']; end
            end
            for i = PRECISION+1:length(szam)
              if (strcmp(szam(i),'1')) w = 1; end
            end
            if (w == 1)
              warning(['Number of not zero bits at the end of mantissa in string is more than ',...
                'number of mantissa-bits of quantizer. ',...
                'It can cause false result.']);
            end
            y = qdata(bin2numv3,NAME);
            return;
          catch
            v2=cellstr(v);
            v3=char(v2);
            w=0;
            szam='';
            for i=1:hatv1-1
              if (strcmp(v3(i),'1')) szam = [szam,'1']; end
              if (strcmp(v3(i),'0')) szam = [szam,'0']; end
            end
            for i = PRECISION+1:length(szam)
              if (strcmp(szam(i),'1')) w = 1; end
            end
            if (w == 1)
              warning(['Number of not zero bits at the end of mantissa in string is more than ',...
                'number of mantissa-bits of quantizer. ',...
                'It can cause false result.']);
            end
            mantissza = v3(1:hatv1-1);
            exponens = [v3(nyito+1:csuko-2),' ',v3(csuko-1)];
            string=[mantissza,' b @ ',exponens];
            string(string=='(') = '';
            string(string==')') = '';
            y=qdata(string,NAME);
            return;
          end
        end
      end
    case 'double'
      for k = 1:length(v)
        % convert double to segments
        t = double2seg(v(k)).';
        % rounding and collecting stat info - real part
        [re,of,uf,in,warn,err] = seground(real(t),code);
        if ~isempty(err),   error(['Error for number ',num2str(v(k)),': ', err]), end
        if ~isempty(warn),warning(['For number ',num2str(v(k)),': ',warn]), end
        % saving of, uf and inv values into 'ofs', 'ufs' and 'ins' index vectors
        ofs(k) = ofs(k)|of;
        ufs(k) = ufs(k)|uf;
        ins(k) = ins(k)|in;
        % rounding and collecting stat info - imag part
        [im,of,uf,in,warn,err] = seground(imag(t),code);
        if ~isempty(err),   error(['Error for number ',num2str(v(k)),': ', err]), end
        if ~isempty(warn),warning(['For number ',num2str(v(k)),': ',warn]), end
        % saving of, uf and inv values into 'ofs', 'ufs' and 'ins' index vectors
        ofs(k) = ofs(k)|of;
        ufs(k) = ufs(k)|uf;
        ins(k) = ins(k)|in;
        % put back complex data into 'z'
        % transpose operator is needed instead of ctranspose
        z(:,k) = (re+j*im);
      end
    otherwise
      error('Internal error 2')
  end
end

% restore size of 'z' with leading segment dimensions
z = reshape(z,[SEGNUM,siz]);
% make object from 'z'
y.name = name;
y.data = z;
y = class(y,'qdata');
if strcmp(qget(name,'operation'),'PQN')
  %error('PQN is not yet ready (under debugging at this moment...)')
end

% store statistical info
if any(ofs(:)) | any(ufs(:)) | any(ins(:))
  % reshape stat info variables into size of 'data' without leading
  % segment dimensions
  if length(siz)>1
    ofs = reshape(ofs,siz);
    ufs = reshape(ufs,siz);
    ins = reshape(ins,siz);
  end
  st = dbstack;
  % clear the top of the stack containing info about current function
  st(1) = [];
  % prevent from storing empty 'dbstack' struct with complete fieldnames
  % (in case when command line event occured and the former operation cleared
  % this)
  if isempty(st), st = []; end
  % fill up struct
  stat.dbstack = st;
  stat.ofmask = ofs;
  stat.ufmask = ufs;
  stat.invmask = ins;
  stat.arg1 = v;
  stat.arg2 = [];
  stat.result = y;
  % send to 'quantizers'
  qdspdemo('quantizers','event',stat);
end
%