function [y,err] = quantizers(varargin)
%QUANTIZERS Contains quantizer properties. (*internal function*)
%   QUANTIZERS is the internal storage function of the quantizers. It implements
%   an 'object' that isn't simply removable from the memory. The quantizer manager
%   functions are front-ends for QUANTIZERS. Calling this function
%   with arguments directly by the user is dangerous.

%Arguments for developers:
%  [y,err]=quantizers(<q_number>)
%  [y,err]=quantizers(<q_name>)
%     get short codes of a quantizer (used by @qdata functions like plus(a,b) etc.)
%  [y,err]=quantizers('get',<q_number>)
%  [y,err]=quantizers('get',<q_name>)
%  [y,err]=quantizers('get','-all')       <- get the internal 'q' variable
%     get long codes of quantizer(s) (used by q.manager front-end functions
%                                     like qget, qset etc.)
%  [y,err]=quantizers('set',<q_number>,<newstruct>)
%  [y,err]=quantizers('set',<q_name>,<newstruct>)
%  [y,err]=quantizers('set','-new',<newstruct>)
%     replace a quantizer with a given new one or create a new quantizer
%  [y,err]=quantizers('rem',<q_number>)
%  [y,err]=quantizers('rem',<q_name>)
%  [y,err]=quantizers('rem','-all')
%     remove quantizer or all the quantizers
%  quantizers('event','clear')
%     clear the events cell vector
%  quantizers('event',<struct>)
%     insert a new event to the events cell vector that contains structs
%     field names: dbstack       - first output of dbstack, struct vector with
%                                  information about line and function name
%                  ofmask        - mask of overflows
%                  ufmask        - mask of underflows
%                  invmask       - mask of invalid operations
%                  arg1          - first argument (or only argument)
%                  arg2          - second argument (if exist)
%                  result        - result in object form, it contains the name
%                                  of the quantizer too
%  [y,err]=quantizers('event','get')
%     obtain the events cell vector

% Copyright (C) 2003, 2004 by JUHASZ Laszlo

% ***** ATTENTION *****
% uncomment the line below to prevent function from 'clear all'
mlock

persistent fr q c f events
% 'fr' indicates first run of function -> for greetings message only
% 'q' contains quantizer properties in long, structure format
% 'c' contains quantizer properties in short, character string format
% 'f' contains the fieldnames of 'q'
% 'events' contains the events struct vector
y = []; err = [];

if isempty(c)|~isfield(c,'BigFP')|~isfield(c,'BFP')|~isfield(c,'double')|~isfield(c,'single')|...
    ~isfield(c,'Q8bit')|~isfield(c,'Q16bit')|~isfield(c,'Q32bit')|~isfield(c,'Q64bit')
  if (length(varargin)~=3) | ...
      ~( isstr(varargin{1})&strcmp(varargin{1},'set')&...
      isstr(varargin{2})&strcmp(varargin{2},'-new')&...
      isstruct(varargin{3})&isfield(varargin{3},'name')&...
      ( strcmp(getfield(varargin{3},'name'),'BigFP')|...
			strcmp(getfield(varargin{3},'name'),'BFP')|...
      strcmp(getfield(varargin{3},'name'),'double')|...
      strcmp(getfield(varargin{3},'name'),'single')|...
      strcmp(getfield(varargin{3},'name'),'Q8bit')|...
      strcmp(getfield(varargin{3},'name'),'Q16bit')|...
      strcmp(getfield(varargin{3},'name'),'Q32bit')|...
      strcmp(getfield(varargin{3},'name'),'Q64bit')) )
    if ~isfield(c,'BigFP'), qmake('name','BigFP','precision',110), end
    if ~isfield(c,'BFP'), qmake('name','BFP','precision',110), end
    if ~isfield(c,'double'), qmake('name','double','precision',53), end
    if ~isfield(c,'single'), qmake('name','single','precision',24), end
    if ~isfield(c,'Q8bit'), qmake('name','Q8bit','type','fixed-point','Bits',8), end
    if ~isfield(c,'Q16bit'), qmake('name','Q16bit','type','fixed-point','Bits',16), end
    if ~isfield(c,'Q32bit'), qmake('name','Q32bit','type','fixed-point','Bits',32), end
    if ~isfield(c,'Q64bit'), qmake('name','Q64bit','type','fixed-point','Bits',64), end
  end
end

if isempty(varargin), return, end

% call for short codes, argument is name
if nargin==1
  if isstr(varargin{1})&strcmp(varargin{1},'BigFP')&~isquant('BigFP')
    qmake('name','BigFP','precision',110)
	elseif isstr(varargin{1})&strcmp(varargin{1},'BFP')&~isquant('BFP')
    qmake('name','BFP','precision',110)
  elseif isstr(varargin{1})&strcmp(varargin{1},'double')&~isquant('double')
    qmake('name','double')
  elseif isstr(varargin{1})&strcmp(varargin{1},'single')&~isquant('single')
    qmake('name','single')
  elseif isstr(varargin{1})&strcmp(varargin{1},'Q8bit')&~isquant('Q8bit')
    qmake('name','Q8bit')
  elseif isstr(varargin{1})&strcmp(varargin{1},'Q16bit')&~isquant('Q16bit')
    qmake('name','Q16bit')
  elseif isstr(varargin{1})&strcmp(varargin{1},'Q32bit')&~isquant('Q32bit')
    qmake('name','Q32bit')
  elseif isstr(varargin{1})&strcmp(varargin{1},'Q64bit')&~isquant('Q64bit')
    qmake('name','Q64bit')
  end
  if isempty(c)
    err = 'No quantizers available';
    return,
  end
  y = eval(['c.',varargin{1}]); return
end

in = celldisint(varargin);
if isempty(in), return, end
t = in{1};

if isempty(fr)
  s = '*';
  %fprintf('%s\n',s(ones(1,97)))
  %fprintf('%s\n',[s,'                                                            ',...
  %      '                                   ',s])
  %fprintf('%s\n',[s,'                     QUANTIZATION and DSP SIMULATION Toolbox  ',...
  %      s,' v0.13b                         ',s,])
  %fprintf('%s\n',[s,'                                                            ',...
  %      '                                   ',s])
  %fprintf('%s\n\n',s(ones(1,97)))
  % clear first-run indicator
  fr = 0;
  % initialize event struct
  events.dbstack    = [];
  events.ofmask     = [];
  events.ufmask     = [];
  events.invmask    = [];
  events.arg1       = [];
  events.arg2       = [];
  events.result     = [];
end

if length(in)<2, error('Internal error 2'), end
u = in{2};

switch t
  case 'get'
    % get property code
    if length(in)~=2, error('Internal error 3'), end
    if strcmp(u,'BigFP')
      y=q;
      for k=length(y):-1:1
        if ~strcmp(y(k).name,'BigFP'), y(k)=[]; end
      end
      return
    end
    if strcmp(u,'double')
      y=q;
      for k=length(y):-1:1
        if ~strcmp(y(k).name,'double'), y(k)=[]; end
      end
      return
    end
    if strcmp(u,'single')
      y=q;
      for k=length(y):-1:1
        if ~strcmp(y(k).name,'single'), y(k)=[]; end
      end
      return
    end
    if strcmp(u,'Q16bit')
      y=q;
      for k=length(y):-1:1
        if ~strcmp(y(k).name,'Q16bit'), y(k)=[]; end
      end
      return
    end
    if strcmp(u,'Q32bit')
      y=q;
      for k=length(y):-1:1
        if ~strcmp(y(k).name,'Q32bit'), y(k)=[]; end
      end
      return
    end
    if strcmp(u,'Q64bit')
      y=q;
      for k=length(y):-1:1
        if ~strcmp(y(k).name,'Q64bit'), y(k)=[]; end
      end
      return
    end
    if strcmp(u, '-all')      % all
      y = q;
      if isstruct(y)&~isfield(y,'operation'), y.operation=''; end
      for k=length(y):-1:1
        if strcmp(y(k).name,'BigFP')|strcmp(y(k).name,'BFP')|strcmp(y(k).name,'double')|strcmp(y(k).name,'single')|...
            strcmp(y(k).name,'Q8bit')|strcmp(y(k).name,'Q16bit')|strcmp(y(k).name,'Q32bit')|strcmp(y(k).name,'Q64bit')
          %Hide standard quantizers
          y(k)=[];
        end
      end %for k
      return
    else                          % only one
      if isempty(q)
        err = 'No quantizers in memory.';
        return,
      end
      switch class(u)
        case 'char'
          % check argument
          if size(u,1)~=1, error('Internal error 4'), end
          % look for quantizer with given name
          for k = 1:length(q)
            if strcmp(q(k).name,u), y = q(k); return, end
          end
          y=[]; err='';
          if strncmpi(u,'name',3)
            %for k = 1:length(q), y=[y;{q(k).name}]; end
            y=q;
          elseif strncmpi(u,'type',3)
            %for k = 1:length(q), y=[y;{q(k).type}]; end
            y=q;
          end
          if ~isempty(y), return, end
          err = ['Invalid quantizer name ''',u,'''.']; return
        case 'double'
          % check argument
          if prod(size(u))~=1 | ~all(isfinite(u)) | ~all(rem(u,1)==0)
            error('Internal error 5')
          elseif u<1 | u>length(q)
            err = ['Invalid quantizer number ',num2str(u),'.']; return
          else
            y = q(u);
            return
          end
        otherwise error('Internal error 6')
      end
    end
  case 'set'
    % replace a quantizer with a given new one or create a new quantizer
    if length(in)~=3, error('Internal error 7'), end
    v = in{3};
    switch class(u)
      case 'char'
        if strcmp(u,'-new')
          pos = length(q)+1; %position: after existing ones
          for k = 1:length(q)
            if strcmp(q(k).name,v.name), pos = k;
              warning(['Replacing existing quantizer ''',v.name,'''']),
              if (strcmp(v.name,'BigFP')&isfield(c,'BigFP')) | ...
                  (strcmp(v.name,'double')&isfield(c,'BFP')) | ...
                  (strcmp(v.name,'double')&isfield(c,'double')) | ...
                  (strcmp(v.name,'single')&isfield(c,'single')) | ...
                  (strcmp(v.name,'Q8bit')&isfield(c,'Q8bit')) | ...                  
                  (strcmp(v.name,'Q16bit')&isfield(c,'Q16bit')) | ...
                  (strcmp(v.name,'Q32bit')&isfield(c,'Q32bit')) | ...
                  (strcmp(v.name,'Q64bit')&isfield(c,'Q64bit'))
                error(['Cannot set properties of default quantizer ''',v.name,''''])
              end
              break,
            end
          end %if existing quantizer, it will be replaced
        else
          if (strcmp(v.name,'BigFP')&isfield(c,'BigFP')) | ...
              (strcmp(v.name,'BFP')&isfield(c,'BFP')) | ...
              (strcmp(v.name,'double')&isfield(c,'double')) | ...
              (strcmp(v.name,'single')&isfield(c,'single')) | ...
              (strcmp(v.name,'Q8bit')&isfield(c,'Q8bit')) | ...
              (strcmp(v.name,'Q16bit')&isfield(c,'Q16bit')) | ...
              (strcmp(v.name,'Q32bit')&isfield(c,'Q32bit')) | ...
              (strcmp(v.name,'Q64bit')&isfield(c,'Q64bit'))
            error(['Cannot set properties of quantizer ''',v.name,''''])
          end
          if isempty(q), err = 'No quantizers in memory.'; return, end
          pos = [];
          for k = 1:length(q)
            if strcmp(q(k).name,u), pos = k; break, end
          end
          if isempty(pos), err = ['Invalid quantizer name ''',u,'''.']; return, end
        end
      case 'double'
        if prod(size(u))~=1 | ~all(isfinite(u)) | ~all(rem(u,1)==0)
          error('Internal error 8')
        elseif u<1 | u>length(q)
          err = ['Invalid quantizer number ',num2str(u),'.']; return
        else pos = u;
        end
      otherwise error('Internal error 9')
    end
    % 'pos' is the serial number of column containing the quantizer to replace
    % checking the new value
    if ~isa(v,'struct') | prod(size(v))~=1, error('Internal error 10'), end
    if isempty(q)
      f = qtest('name','a');     % generating test struct for fieldnames
    end
    % checking fieldnames by trying to concatenate the reference and new structs
    try, [v,f]; catch, error('Internal error 11'), end
    if ~isempty(q)
      tq = q; tq(pos) = v;
      % checking duplicated names
      for k = 1:length(tq)-1
        for m = k+1:length(tq)
          if strcmp(tq(k).name, tq(m).name)
            if ~strcmp(v.rights,'writable');
              err = ['Quantizer with name ''',v.name,''' already exists.']; return
            end
          end
        end
      end
      % if no error occured -> changing the old value to new
      q = tq;
    else
      q = v;
    end
    if ~isfield(q,'operation'), q.operation=''; end
    c = codegen(q);
    return
  case 'rem'
    % remove quantizer
    if length(in)~=2, error('Internal error 12'), end
    %if isempty(q), err = 'No quantizers in memory.'; return, end
    if strcmp(u, '-all')      % all
      for k=length(q):-1:1
        if ~strcmp(getfield(q(k),'name'),'BigFP')&~strcmp(getfield(q(k),'name'),'BFP')& ...
						~strcmp(getfield(q(k),'name'),'double') &...
            ~strcmp(getfield(q(k),'name'),'single')&~strcmp(getfield(q(k),'name'),'Q16bit') & ...
            ~strcmp(getfield(q(k),'name'),'Q8bit')&~strcmp(getfield(q(k),'name'),'Q32bit') & ...
            ~strcmp(getfield(q(k),'name'),'Q64bit')
          q(k) = [];
        end
      end
      %fld1=getfield(c,'BigFP'); fld2=getfield(c,'double');
      %c = [];
      %c.BigFP=fld1;
      %c.double=fld2;
      return
    else                          % only one
      switch class(u)
        case 'char'
          % check argument
          if size(u,1)~=1, error('Internal error 13'), end
          % look for quantizer with given name
          if strcmp(u,'BigFP')|strcmp(u,'BFP')|strcmp(u,'double')|strcmp(u,'single')|...
              strcmp(u,'Q8bit')|strcmp(u,'Q16bit')|strcmp(u,'Q32bit')|strcmp(u,'Q64bit')
            error(['Cannot set properties of quantizer ''',u,''''])
          end
          for k = 1:length(q)
            if strcmp(q(k).name,u), q(k)=[]; return, end
          end
          %err = ['Invalid quantizer name ''',u,'''.'];
          return
        case 'double'
          % check argument
          if prod(size(u))~=1 | ~all(isfinite(u)) | ~all(rem(u,1)==0)
            error('Internal error 14')
          elseif u<1 | u>length(q)
            err = ['Invalid quantizer number ',num2str(u),'.']; return
          else q(u)=[]; return
          end
        otherwise error('Internal error 15')
      end
    end
  case 'event'
    % event operations
    if length(in)~=2, error('Internal error 16'), end
    switch class(u)
      case 'char'
        switch u
          case 'get'
            if length(events)<2
              err = 'No events.';
            else
              y = events(2:end);
            end
            return
          case 'clear'
            events(2:end) = [];
            return
          otherwise error('Internal error 17')
        end
      case 'struct'
        % inserting event
        events = [events,u];
      otherwise error('Internal error 18')
    end
  otherwise error('Internal error 16')
end
if ~isfield(y,'operation'), y.operation=''; end
