function out = qget(varargin)
%QGET  Get quantizer properties.
%   V = QGET(H,'PropertyName') returns the value of the specified
%   property for the quantizer with name or number H. If H is a
%   vector of quantizer numbers or cell array of quantizer numbers
%   and/or names, then QGET will return an M-by-1 cell array of values
%   where M is equal to length(H). If 'PropertyName' is replaced by a
%   1-by-N or N-by-1 cell array of strings containing property names,
%   then QGET will return an M-by-N cell array of values.
%
%   QGET(H) displays all property names and their current values for
%   the quantizer with name/number H.
%
%   QGET displays all property names and their current values for all
%   quantizers.
%
%   V = QGET(H) where H is a quantizer name/number, returns a structure
%   where each field name is the name of a property of H and each field
%   contains the value of that property.
%
%   V = QGET  returns a struct vector containing properties and values
%   of all quantizers.
%
%   V = QGET(H, 'Default')
%   V = QGET(H, 'Default<PropertyName>')
%   returns information about default property values (H must be a
%   quantizer name/number). 'Default<PropertyName>' returns the default
%   value for the specific property.
%
%   See also QSET, .

% Copyright (C) 2003 by JUHASZ Laszlo

%Test example: qget('Qdsp','CODE')

% input condition
for k = 1:length(varargin)
   x = varargin{k};
   if isa(x,'qdata'), x=x.name; end
   if isempty(x), error(['Empty ',class(x),' type input.']), end
   switch class(x)
   case 'double'
      if ~all(isfinite(x))
         error('Not finite numeric input.')
      end
      if ~all(rem(x,1)==0)
         error('Not integer numeric input.')
      end
   case 'char'
      if ndims(x)~=2 | prod(size(x))~=length(x) | size(x,1)~=1
         error('Not row string char input.')
      end
      [a,count,errmsg] = sscanf(x,'%g');
      if count~=1 | ~isempty(errmsg) | ~isfinite(a) | rem(a,1)
         %x = lower(x);
      else
         x = a;
      end
   case 'cell'
      if ndims(x)~=2 | prod(size(x))~=length(x)
         error('Not array cell input.')
      end
      for m = 1:length(x)
         z = lower(x{m});
         if isempty(z), error(['Empty ',class(z),' type input in cell.']), end
         switch class(z)
         case 'double'
            if ~all(isfinite(z))
               error('Not finite numeric input in cell.')
            end
            if ~all(rem(z,1)==0)
               error('Not integer numeric input in cell.')
            end
         case 'char'
            if ndims(z)~=2 | prod(size(z))~=length(z) | size(z,1)~=1
               error('Not row string char input in cell.')
            end
            [a,count,errmsg] = sscanf(z,'%g');
            if count~=1 | ~isempty(errmsg) | ~isfinite(a) | rem(a,1)
               z = lower(z);
            else
               z = a;
            end
         otherwise
            error(['Invalid type ''',class(z),''' in cell.'])
         end
         x{m} = z;
      end
   case 'struct'
     try, qstructlist(varargin{:}), return
     catch, error(['Invalid type ''',class(q),''' first argument.'])
     end
     case 'qdata'
       get(x); return
     otherwise
      error(['Invalid type ''',class(x),'''.'])
   end
   varargin{k} = x;
end
x = varargin;
%-----------------------------------------------------------------------------------------
switch length(x)
case 0
   % QGET
   % V = QGET
   [t,err] = quantizers('get','-all');
   if ~isempty(t), t=upperlower(t); end
   if nargout==0
      for k = 1:length(t), disp(t(k)), disp(' '),end
   else
      out = t;
   end
case 1
   % QGET(H)
   % V = QGET(H)
   x = x{1};
   if ~((isa(x,'double') & prod(size(x))==1) | ischar(x))
      error('Input must be scalar numeric or string when only one argument is given.')
   end
   [t,err] = quantizers('get',x); 
   if isempty(t)&(strcmp(x,'all')|strcmp(x,'-all')), [t,err] = quantizers('get','-all'); err=''; end
   error(err)
   if strcmp(x,'name')|strcmp(x,'type')
     y=[]; for k = 1:length(t), y=[y;{getfield(t(k),x)}]; end
     if nargout==0, disp(y), else out=y; end
     return
   elseif (strcmp(x,'all'))|strcmp(x,'-all')
     y=[];
     for k = 1:length(t)
       yk=[getfield(t(k),'name'),': ',getfield(t(k),'type')];
       if strncmp(getfield(t(k),'type'),'floating',5), yk=[yk,', p = ',num2str(getfield(t(k),'precision'))];
       else yk=[yk,'B = ',num2str(getfield(t(k),'bits'))];
       end
       y=[y;{yk}];
     end
     if nargout==0, disp(y), else out=y; end
     return
   end
   tul=upperlower(t);
   if nargout==0
     fn=fieldnames(tul); tshow=[];
     for ii=1:length(fn)
       if isempty(getfield(tul,fn{ii})), %tshow=rmfield(tshow,fn{ii}); 
       else
         if strcmp(tul.Type,'fixed-point')&strcmp(fn{ii},'Precision')
           tshow=setfield(tshow,'Bits',getfield(tul,fn{ii})+1);
         else
           tshow=setfield(tshow,fn{ii},getfield(tul,fn{ii}));
         end
       end
     end
     disp(tshow)
   else
      out = t;
   end
case 2
   t = x{1};
   u = x{2};
   switch class(u)
   case 'char'
      if strncmp(u,'default',7)
         % V = QGET(H, 'Default')
         % V = QGET(H, 'Default<PropertyName>')
         % check input
         switch class(t)
         case 'double'
            if prod(size(t))~=1, error('Not scalar numeric input.'), end
         case 'char'
         otherwise error('Invalid type input.')
         end
         if strcmp(u,'default')
            % V = QGET(H, 'Default')
            [q,err] = quantizers('get',t); error(err)
            f = fieldnames(q);
            tout = {};
            for k = 1:length(f)
               val = eval(['q.',f{k}]);
               def = eval(['def_',f{k},'(q)']);
               switch class(val)
               case 'char'
                  if strcmp(val,def), tout = {tout{:},f{k}}; end
               case 'double'
                  if val==def, tout = {tout{:},f{k}}; end
               otherwise error('Internal error')
               end
            end
            out = tout';
         else
            % V = QGET(H, 'Default<PropertyName>')
            name = u(8:end);
            [q,err] = quantizers('get',t); error(err)
            f = fieldnames(q);
            % determine whether name is ok, invalid, or ambiguous
            occ = [];
            for k = 1:length(f)
               if strncmp(name,f{k},length(name)), occ = [occ,k]; end
            end
            switch length(occ)
            case 0
               error(['Invalid property name ''',name,'''.'])
            case 1
               out = eval(['def_',f{k},'(q)']);
            otherwise error(['Ambiguous property name ''',name,'''.'])
            end
         end
      else
         % V = QGET(H,'PropertyName') with simple char second input
         % check input
         if ((isa(t,'cell') | isa(t,'double')) & length(t)==1) | isa(t,'char')
            % scalar first input
            if isa(t,'cell'), t = t{1}; end
            [q,err] = quantizers('get',t); error(err)
            if isstr(u)&strcmpi(u,'qcode'), out=getfield(codegen(q),t); return, end
            % determine whether name is ok, invalid, or ambiguous
            f = fieldnames(q);
            occ = [];
            for k = 1:length(f)
               if strncmpi(u,f{k},length(u)), occ = [occ,k]; end
            end
            switch length(occ)
            case 0
              if strncmpi(u,'Bits',length(u)), out=q.precision+1;
              else error(['Invalid property name ''',u,'''.'])
              end
            case 1
               out = eval(['q.',f{occ}]);
            otherwise error(['Ambiguous property name ''',u,'''.'])
            end
         else
            % vector first input
            if isa(t,'double'), t = num2cell(t); end
            % collect structs
            for k = 1:length(t)
               [temp,err] = quantizers('get',t{k}); error(err)
               q(k) = temp;
            end
            % determine whether name is ok, invalid, or ambiguous
            f = fieldnames(q);
            occ = [];
            for k = 1:length(f)
               if strncmp(u,f{k},length(u)), occ = [occ,k]; end
            end
            switch length(occ)
            case 0
               error(['Invalid property name ''',u,'''.'])
            case 1
               % collect out values to 'tout'
               tout = {};
               for k = 1:length(q)
                  tout = {tout{:},eval(['q(',num2str(k),').',f{occ}])};
               end
               out = tout';
            otherwise error(['Ambiguous property name ''',u,'''.'])
            end
         end
      end
   case 'cell'
      % V = QGET(H,'PropertyName') with cell second input
      % check input
      if ((isa(t,'cell') | isa(t,'double')) & length(t)==1) | isa(t,'char')
         % scalar first input
         if isa(t,'cell'), t = t{1}; end
         [q,err] = quantizers('get',t); error(err)
         % determine whether name is ok, invalid, or ambiguous
         f = fieldnames(q);
         for k = 1:length(u)
            occ = [];
            for m = 1:length(f)
               if strncmp(u{k},f{m},length(u{k})), occ = [occ,m]; end
            end
            switch length(occ)
            case 0
               error(['Invalid property name ''',u{k},'''.'])
            case 1
               % completing name
               u{k} = f{occ};
            otherwise error(['Ambiguous property name ''',u{k},'''.'])
            end
         end
         % all names are OK
         % collect values
         tout = {};
         for k = 1:length(u)
            tout = {tout{:},eval(['q.',u{k}])};
         end
         out = tout;
      else
         % vector first input
         if isa(t,'double'), t = num2cell(t); end
         % collect structs
         for k = 1:length(t)
            [temp,err] = quantizers('get',t{k}); error(err)
            q(k) = temp;
         end
         % determine whether name is ok, invalid, or ambiguous
         f = fieldnames(q);
         for k = 1:length(u)
            occ = [];
            for m = 1:length(f)
               if strncmp(u{k},f{m},length(u{k})), occ = [occ,m]; end
            end
            switch length(occ)
            case 0
               error(['Invalid property name ''',u{k},'''.'])
            case 1
               % completing name
               u{k} = f{occ};
            otherwise error(['Ambiguous property name ''',u{k},'''.'])
            end
         end
         % all names are OK
         tout = {};
         for k = 1:length(q)
            for m = 1:length(u)
               tout{k,m} = eval(['q(',num2str(k),').',u{m}]);
            end
         end
         out = tout;
      end
   otherwise
      error('Second argument has invalid type.')
   end
otherwise
   error('Too many input arguments.')
end
