function Name=qmake(varargin)
%QMAKE   Create new quantizer.
%   QMAKE NAME creates a new quantizer with given name.
%   QMAKE PROPNAME PROPVALUE ... specifies one or more properties of the quantizer.
%     (fixed-point, floating-point, rounding, etc.)
%   QMAKE NAME <structure of properties>
%   QMAKE NAME <string of binary representation of a number>
%   The unspecified properties will get their computed default values.
%   Default quantizers (always usable, cannot redefine): 
%          BigFP or BFP, 110-bit floating-point Q
%          double 53-bit floating-point Q, IEEE standard double
%          single 24-bit floating-point Q, IEEE standard single
%          Q64bit fixed-point, two's complement 64-bit
%          Q32bit fixed-point, two's complement 32-bit
%          Q16bit fixed-point, two's complement 16-bit
%          Q8bit fixed-point, two's complement 8-bit
%
%   Examples:
%     qmake('quantizer_a')
%     qmake('Name','quantizer_a','type','floating-point','precision',24)
%   The same with fixed-point property names and values:
%     qmake('na','quantizer_b','ty','fi','B',16)
%
%   See also: QCOPY, QSAVE, QGET, QSET

%   Copyright (C) 2003 by JUHASZ Laszlo

if (nargin==2)&isstruct(varargin{2})
  %process structure of properties
  tmp=varargin{2}; varargin(2)=[];
  fn=fieldnames(tmp);
  for ii=length(fn):-1:1
    varargin=[varargin(1),{fn{ii}},{getfield(tmp,fn{ii})},varargin(2:end)];
  end
elseif (nargin>=2)&isstr(varargin{2})
  propnames = {'name' 'notes' 'rights' 'type' 'coding' 'precision' 'leadingbit'...
    'fractionbits' 'nansignal' 'overflow' 'ofsignal' 'underflow' 'ufsignal'...
    'roundingdir' 'treathalf' 'expcoding' 'expbits' 'expbias' 'emax' 'emin' 'maximum' 'minimum'...
        'dither' 'qnoise'...
    'operation'};
  %~any(strmatch(varargin{1},propnames,'exact'))
  if any(ismember(varargin{2},'.^ +-[]()'))|ismember(varargin{2}(1),'0123456789')
    %parser
    str=varargin{2}; Q=[];
    %eliminate spaces
    ind=findstr(str,' '); if ~isempty(ind), str(ind)=''; end
  end
end

x = celldisint(varargin);
if isempty(x), error('Not enough input arguments.'), end

if length(x)==1
  % only name
  keyw={'Q8bit','Q16bit','Q32bit','Q64bit','double','single','BigFP','BFP'}; %
  inde=strmatch(x{1},keyw,'exact');
  ind=strmatch(x{1},keyw);
  if any(ind)%&~any(inde)
    error(['Potentially misleading quantizer name ''',x{1},''' (similar to keyword ''',keyw{ind},''')'])
  end
  x=[{'name'},x];
  [t,err] = qtest(x); ct=1;
elseif (length(x)==2)&isstr(x{1})&~strncmpi(x{1},'name',3)&isstr(x{2})&...
    all(ismember(x{2},'0123456789abcdefodht.+* ()[]2^@'))
  %extract quantizer from string x{2}
  s=x{2};
  ind=findstr(s,' '); if ~isempty(ind), s(ind)=''; end
  ind=findstr(s,'@'); 
  if ~isempty(ind), m=s(1:ind-1); e=s(ind+1:end); 
  else
    ind=findstr('*2^',s);
    if ~isempty(ind), m=s(1:ind-1); e=s(ind+1:end);
    else
      ind=findstr('2^',s);
      if ~isempty(ind), m=s(1:ind-1); e=s(ind+1:end); else m=s; e=[]; end
    end
  end
  if ~isempty(e)
    typ='floating-point';
    emax=1023; emin=-1022;
  else
    typ='fixed-point'; emax=[]; emin=[];
  end
  if strcmp(m(1),'+')|strcmp(m(1),'-'), sign=num2str([m(1),'1']); else sign=1; end
  mind=find(((s>='0')&(s<='9'))|(s=='.'));
  ms=m(mind);
  d=find(ms=='.'); if isempty(d), d=0; end
  fb=length(ms(d+1:end));
  if d>0, ms(d)=''; end
  if strcmp(typ,'floating-point'), p=length(ms); 
  else B=length(ms)+1; 
  end
  if strcmp(typ,'floating-point')
    Q={'type',typ,'p',p,'emin',emin,'emax',emax,'coding','sign-mag'};
  else
    Q={'type',typ,'B',B,'fractionbits',fb};
  end
  [t,err] = qtest([x{1},Q]); ct=0;
else
  if (length(x)==3)&(strcmp(x{3},'8bit')|strcmp(x{3},'16bit')|strcmp(x{3},'32bit'))
    typ='fixed-point'; x=[x(1:2),{'type',typ,'B',str2num(x{3}(1:2))}];
  else
    typ='floating-point';
  end
  % propnames - values
  for k=length(x)-1:-2:2,
    if strncmpi(x{k},'Bits',1), typ='fixed-point'; break, end
    if strncmpi(x{k},'Type',3), typ=x{k+1}; break, end
  end
  %if isstr(x{1})&strcmpi(x{k},'type'), x=[{'name'},x]; end
  if length(x)>=3
    [t,err] = qtest([x,{'type',typ}]); ct=2;
  else
    [t,err] = qtest(x); ct=3;
  end
end

error(err)
[y,err] = quantizers('set','-new',t);
error(err)
if nargout>0, 
  if iscell(x), Name=t.name; else Name=t.name; end
end
%