function out = qset(varargin)
%QSET      Set quantizer properties.
%   QSET(H,'PropertyName',PropertyValue) sets the value of the
%   specified property for the quantizer with name/number H.
%   H can be a vector of quantizer numbers or cell array of
%   quantizer names/numbers, in which case SET sets the
%   properties' values for all the quantizers.
%
%   QSET(H,a) where a is a structure whose field names are
%   quantizer property names, sets the properties named in each
%   field name with the values contained in the structure.
%
%   QSET(H,pn,pv) sets the named properties specified in the cell array
%   of strings pn to the corresponding values in the cell array pv for
%   all quantizers specified in H. The cell array pn must be 1-by-N or
%   N-by-1, but the cell array pv can be M-by-N where M is equal to
%   length(H) so that each quantizer will be updated with a different
%   set of values for the list of property names contained in pn.
%
%   QSET(H,'PropertyName1',PropertyValue1,'PropertyName2',PropertyValue2,...)
%   sets multiple property values with a single statement.  Note that it
%   is permissible to use property/value string pairs, structures, and
%   property/value cell array pairs in the same call to QSET.
%
%   A = QSET('PropertyName') 
%   QSET('PropertyName')
%   returns or displays the possible values for the specified
%   property. The returned array is a cell array of possible value
%   strings or an empty cell array if the property does not have a finite
%   set of possible string values.
%   
%   A = QSET
%   QSET
%   returns or displays all property names and their possible values for
%   quantizers. The return value is a structure whose field names are
%   the property names, and whose values are cell arrays of possible
%   property values or empty cell arrays.
%
%   One string have special meaning for PropertyValues:
%     'default' - use default value
%
%   See also QGET,

% Copyright (C) 2003 by JUHASZ Laszlo

if (length(varargin)==2)&isquant(varargin{1})
  varargin(1)=[]; 
end
if isempty(varargin)
   % A = QSET
   % QSET
   [t,err] = qtest('get=internal=values'); error(err)
   s = t{1};      % original struct
   v = t{2};      % 'values' cell
   f = fieldnames(s);
   for k = 1:length(f)
      eval(['s.',f{k},'=v{k};'])
   end
   if nargout==0
     s=upperlower(s); 
     disp(s)
   else
      out = s;
   end
   return
end

if length(varargin)==1 | ischar(varargin{1})
   % first and only argument is property name -> default value
   % A = QSET('PropertyName') 
   % QSET('PropertyName')
   q = varargin{1};
   % check 'q'
   if ndims(q)~=2 | prod(size(q))~=length(q) | size(q,1)~=1
      error('Not row string char input.')
   end
   [t,err] = qtest('get=internal=values'); error(err)
   s = t{1};      % original struct
   v = t{2};      % 'values' cell
   f = fieldnames(s);
   occ = [];
   for k = 1:length(f)
      if strncmp(q,f{k},length(q)), occ = [occ,k]; end
   end
   if length(occ)==1
      % only the perfect occurrence is filtered
      if nargout==0
         if isempty(v{occ})
            disp(' '),disp('     {}'),disp(' ')
         else
            disp(v{occ})
         end
      else
         out = v{occ};
      end
      return
   end
end

% check the first argument
% the group of quantizers must be exist in one variable (double,cell)
q = varargin{1};
varargin(1) = [];
switch class(q)
case 'double', q = num2cell(q);
case 'char', q = {q};
  if strcmp(q,'BigFP')|strcmp(q,'double')|strcmp(q,'single')|strcmp(q,'Q16bit')|strcmp(q,'Q32bit')
    error(['Cannot set properties of quantizer ''',u,''''])
  end
case 'cell'
case 'struct'
  try, if nargin==1, qlist, else error('Illegal call'), end, return
  catch, error(['Something wrong having ''',class(q),''' as first argument.'])
  end
otherwise error(['Invalid type ''',class(q),''' first argument.'])
end
% collect quantizers to modify
for k = 1:length(q)
  if strcmp(q{k},'Q64'), q{k}='Q64bit'; 
  elseif strcmp(q{k},'Q32'), q{k}='Q32bit'; 
  elseif strcmp(q{k},'Q16'), q{k}='Q16bit'; 
  elseif strcmp(q{k},'Q8'), q{k}='Q8bit'; 
  end
   [t,err] = quantizers('get',q{k}); error(err)
   % checking the already collected quantizers
   if k~=1
      for m = 1:length(s)
         actname = eval(['s(',num2str(m),').name']);
         if strcmp(actname,t.name)
            if ischar(q{k})
               error(['Quantizer with name ''',q{k},''' already in the list'])
            else
               error(['Quantizer with number ',num2str(q{k}),' already in the list'])
            end
         end
      end
   end
   s(k) = t;
end

if isempty(varargin), error('Not enough input arguments.'), end

% check 2nd & 3rd argument
if length(varargin)==2
   t = varargin{1};
   u = varargin{2};
   if (~isempty(t) & ~isempty(u) & isa(t,'cell') & isa(u,'cell') & ...
         prod(size(t))==length(t) & prod(size(u))~=length(u) & ...
         ndims(t)==2 & ndims(u)==2)
      % QSET(H,pn,pv)
      if size(u,1)~=length(s)
         error(sprintf(['The 3rd argument cell must have equal number of rows with\n',...
               'the number of quantizers specified in the 1st argument']))
      end
      if size(u,2)~=length(t)
         error(sprintf(['The 3rd argument cell must have equal number of columns with\n',...
               'the number of properties specified in the 2nd argument']))
      end
      % sizes are ok
      % convert to format that is usable
      z = {};
      for k = 1:size(u,1)
         for m = 1:length(t)
            z = {z{:},t{m},u{k,m}};
         end
      end
      varargin = z;
   end
end

% input condition
for k = 1:length(varargin)
   x = varargin{k};
   if ~isempty(x)
      switch class(x)
      case 'double'
         % vector is enabled
         if ~all(isfinite(x))
            error('Not finite numeric input.')
         end
         if ~all(rem(x,1)==0)
            error('Not integer numeric input.')
         end
         % convert to cell
         x = num2cell(x);
      case 'char'
         if ndims(x)~=2 | prod(size(x))~=length(x) | size(x,1)~=1
            error('Not row string char input.')
         end
         [a,count,errmsg] = sscanf(x,'%g');
         if count~=1 | ~isempty(errmsg) | ~isfinite(a) | rem(a,1)
            if ~strcmp(x,'PQN'), x = lower(x); end
         else
            x = a;
         end
      case 'cell'
         if ndims(x)~=2
            error('Not two dimensional cell input.')
         end
         for m = 1:prod(size((x)))
            z = x{m};
            if ~isempty(z)
               switch class(z)
               case 'double'
                  % vector is enabled
                  if ~all(isfinite(z))
                     error('Not finite numeric input.')
                  end
                  if ~all(rem(z,1)==0)
                     error('Not integer numeric input.')
                  end
                  % convert to cell
                  z = num2cell(z);
               case 'char'
                  if ndims(z)~=2 | prod(size(z))~=length(z) | size(z,1)~=1
                     error('Not row string char input in cell.')
                  end
                  [a,count,errmsg] = sscanf(z,'%g');
                  if count~=1 | ~isempty(errmsg) | ~isfinite(a) | rem(a,1)
                     z = lower(z);
                  else
                     z = a;
                  end
               otherwise
                  error(['Invalid type ''',class(z),''' in cell.'])
               end
            end
            x{m} = z;
         end
      case 'struct'
         f = fieldnames(x);
         st = {};
         % check values
         for m = 1:prod(size(x))
            for n = 1:length(f)
               z = eval(['x(',num2str(m),').',f{n}]);
               if ~isempty(z)
                  switch class(z)
                  case 'double'
                     % vector is enabled
                     if ~all(isfinite(z))
                        error('Not finite numeric input.')
                     end
                     if ~all(rem(z,1)==0)
                        error('Not integer numeric input.')
                     end
                     % convert to cell
                     z = num2cell(z);
                  case 'char'
                     if ndims(z)~=2 | prod(size(z))~=length(z) | size(z,1)~=1
                        error('Not row string char input in struct.')
                     end
                     [a,count,errmsg] = sscanf(z,'%g');
                     if count~=1 | ~isempty(errmsg) | ~isfinite(a) | rem(a,1)
                        z = lower(z);
                     else
                        z = a;
                     end
                  otherwise
                     error(['Invalid type ''',class(z),''' in struct.'])
                  end
               end
               st = {st{:},f{n},z};
            end
         end
         x = st;
      otherwise
         error(['Invalid type ''',class(x),'''.'])
      end
   end
   varargin{k} = x;
end
x = celldisint(varargin);
%-----------------
if rem(length(x),2)     % odd num of inputs
   error('The sum of property names and values must be even.')
end
if rem(length(x)/2/length(s),1)
   error('There must be enough property names and values for all the addressed quantizers.')
end
parperq = length(x)/length(s);

% saving s for names
olds = s;

for k = 1:length(s)
   actparams = x(parperq*(k-1)+1:parperq*k);
   ind=strmatch('expma',lower(actparams(1:2:end)));
   if ~isempty(ind), actparams{2*ind-1}='emax'; end
   ind=strmatch('expmi',lower(actparams(1:2:end)));
   if ~isempty(ind), actparams{2*ind-1}='emin'; end
   ind=strmatch('expbits',lower(actparams(1:2:end)));
   if ~isempty(ind), actparams(2*ind-1:2*ind)=[]; end
   [t,err] = qtest(actparams,s(k)); error(err)
   s(k) = t;
end

% inserting in the fake quantizers list
[t,err] = quantizers('get','-all'); error(err)
for k = 1:length(t)
   for m = 1:length(s)
     if strcmp(s(m).name,t(k).name)
       t(k) = s(m); 
       %if s(m).emax>1023, error('Maximum of exponent surpasses 11'), end
     end
   end
end

% checking duplicated names
for k = 1:length(t)-1
   for m = k+1:length(t)
      if strcmp(t(k).name,t(m).name)
        if ~strcmp(t(k).rights,'writable');
          error(['Duplicated name ''',t(k).name,'''.'])
        end
      end
   end
end

% names are OK
for k = 1:length(s)
   [t,err] = quantizers('set',olds(k).name,s(k)); error(err)
end
if nargout>=1, error('Cannot assign quantizer to an output argument'), out=olds(k).name; end
