function [xprime,ovfind,varargout]=roundq(x,varargin)
%ROUNDOFF Quantize (roundoff) number with given quantizer
%
%   [xprime,ovfind,dinfo]=roundq(x,runmod)
%      executes quantization (rounding) on real or complex, scalar or array x.
%      dinfo: field d: value of dither; field m: serial number of
%      sequential dither; field dq: double quantization extra error limits
%      roundq(x,runmod,'list') also lists the settings of the quantizer
%      implemented.
%      roundq(x,runmod,'complex') adds complex dither or complex PQN to
%      the number if set to these options.
%   Possibilities for parameters in the structure runmod:
%     runmod.type: type of quantizer: fixed-point or floating-point
%     runmod.rdir: direction of quantization: 'r', 'f', 'c' 'x', 'i', 't'
%         for round, floor, ceil, fix, toinf, trunc
%     runmod.treath: way of treating *.5 values for round: 'f', 'c', 'x', 'e', 'o', 'r'
%         for floor, ceil, fix, toinf, toeven, toodd, rand
%     runmod.coding = coding of mantissa, 's', 'o', 't'
%         for sign-magnitude, one's complement, two's complement
%   For floating-point only
%     runmod.p: precision: number of bits of the significand, sign not
%         included, but hidden bit is included
%	    runmod.lb: leading bit representation, 'h' or 'hidden' for hidden,
%         's' or 'shown' for shown
%     runmod.ufl: UnderFlow, 'g' for gradual, 'f' for flush
%     runmod.Emin: minimum value of exponent
%     runmod.Emax: maximum value of exponent
%   For fixed-point only
%     runmod.B: number of bits of the number, sign included
%     runmod.p: number of bits of the number, sign not included
%     runmod.fb: number of fractional bits
%     runmod.ovf: overflow handling, 'c', 'm', 't', 'i', 'n'
%         for clip, modular, triang, inf, none
%     runmod.xnmin: minimum representable (negative) number
%     runmod.xpmax: maximum representable number
%   Default values:
%     Floating-point:
%       IEEE double precision: p=53, lb='hidden', rdir='round',
%         treath='toeven', coding='s', ufl='gradual', Emin=-1022, Emax=1023
%       IEEE single precision: p=24, lb='hidden', rdir='round',
%         treath='toeven', coding='s', ufl='gradual', Emin=-126, Emax=127
%       Limits: p<=53, Emin>=-1022, Emax<=1023
%     Fixed-point:
%         B=16, p=15, fb=0, rdir='r', treath='toeven', coding='t', ovf='m',
%         xnmin=-2^(p-fb-1) for 'two''s', else xnmin=-2^(p-fb-1)+1,
%         xpmax=2^(B--fb-1)-1
%       Limits: B<=54, p<=53
%  	Output arguments:
%	    xprime = quantized variable
%	    ovfind = indices of elements of x which cause overflow, indices
%		       of negative overflow are multiplied by (-1), indices of
%		       imaginary parts multiplied furthermore by sqrt(-1)
%   Examples:
%     roundq(1+2^(-24),'s')-1  %returns 0
%     roundq(1+2^(-23),'s')-1  %returns 2^(-23)=1.1921e-007
%     roundq(realmin/2,qstruct('ufl','f'),'list')  %returns 0
%     roundq(realmin/2,qstruct('ufl','g'))  %returns  realmin/2=1.1125e-308
%     roundq(1.75,qstruct('B',4,'fb',0)) %2
%   Remark: type 'roundoff remark'

%	Copyright (c) I. Kollar, 1995-2006, all rights reserved
%	Last modified: 21-Oct-2006

global list_quant_details
if isempty(list_quant_details), list_quant_details=0; end

if nargin<2, varargin={}; end
if (nargin==1)&&ischar(x)&&strncmpi(x,'remark',3), type qf_remark.m, return, end %remark
if ((nargin==2)&&ischar(varargin{1})&&strcmp(varargin{1},'list')) || ...
    ((nargin==3)&&ischar(varargin{2})&&strcmp(varargin{2},'list'))
  varargin=[{''},varargin];
end
%private
if (nargin>=2)&&ischar(x)&&strcmpi(x,'private')
  if isnumeric(varargin{1}), varargin{1}=char(varargin{1}); end
  if nargout==0, feval(varargin{:});
  elseif nargout==1, xprime=feval(varargin{:});
  elseif nargout==2, [xprime,ovfind]=feval(varargin{:});
  elseif nargout==3, [xprime,ovfind,out3]=feval(varargin{:});
    varargout={out3};
  elseif nargout==4, [xprime,ovfind,out3,out4]=feval(varargin{:});
    varargout={out3,out4};
  else error('Too many output arguments')
  end
  return
end
%
if ~isnumeric(x)
  if isa(x,'qdata'), error('Cannot use roundoff on qdata objects. Use the function qdata instead.')
  else error('x is not numeric')
  end
end
varargout={zeros(size(x))};

if ~isempty(varargin)
  if (length(varargin)==1)||(length(varargin)==2)
    keyw={'double','single','8-bit','16-bit','32-bit','s','d'};
    if  ~ischar(varargin{1})
      if isstruct(varargin{1})
        Q=qstruct(varargin{1});
      end
    elseif ~isempty(varargin{1})&&any(strmatch(varargin{1},keyw))
      Q=qstruct(varargin{1});
    else
      Q=qstruct(varargin{1},'double');
    end
  else
    while isempty(varargin{1}), varargin(1)=[]; end
    while ischar(varargin{end})&&strcmp(varargin{end},'list'), varargin(end)=[]; end
    Q=qstruct(varargin{:});
  end

  if isfield(Q,'Dither')
    if strcmpi(Q.Dither,'on')
      if strncmp(Q.Type,'fixed-point',5)||strncmp(Q.Type,'floating-point',5)
        xorig=x;
        [x,d,m,restart] = add_dither(x,Q);
        if nargout>=3, dinfo.d=d; dinfo.m=m; varargout={dinfo}; end
        %varargout={d};
        if (restart == 0)
          Q.Other.NumberOfUsage(m) = Q.Other.NumberOfUsage(m)+1;
        elseif strcmpi(Q.Sequenceperm,'enabled')
          Q.Other.NumberOfUsage = zeros(1,length(Q.Other.Hmtimes));
          Q.Other.NumberOfUsage(m) = Q.Other.NumberOfUsage(m)+1;
        end
        %  dindsave = setfield(dindsave,inputname(2),Q.Other.NumberOfUsage);
        if list_quant_details
          fprintf('Original number: %g \n',xorig);
          Qx=qstruct('p',53,'ExpBias',0,'Emax',0,'Emin',0,'lb','s');
          xstr=num2bin(xorig,Qx); fprintf(xstr), indp=findstr(xstr,'.');
          fprintf('\n');
          Bd=Q.Number_of_DitherBits+Q.FractionBits+1; dbits=Q.Number_of_DitherBits;
          if strcmp(Q.ToD,'u2')||strcmp(Q.ToD,'t2'), Bd=Bd+1; dbits=dbits+1; end
          if Bd>54, error('Cannot make dither in roundoff for %.0f bits. Use qdsp rather.',Bd), end
          Qd=qstruct('B',Bd,'fb',Bd-1,'Coding','s');
          daddstr=num2bin(d,Qd);
          if isfield(Q,'Bits'), ind=Q.Bits-1; else ind=Q.Precision; end
          if ind<length(daddstr), daddstr=[daddstr(1:ind),'||',daddstr(ind+1:end)]; end
          indp2=findstr(daddstr,'.');
          indn=sort([findstr(daddstr,'0'),findstr(daddstr,'1')]);
          %indb=findstr(daddstr,'||');
          %if strcmp(daddstr(indb+1),'1'), daddstr(indb+1)='-';
          %elseif strcmp(daddstr(indb+1),'0'), daddstr(indb+1)='+';
          %end
          for ii=fliplr(5:4:length(indn))
            %if indn(ii)>indb
            daddstr=[daddstr(1:indn(ii)-1),' ',daddstr(indn(ii):end)];
            %end
          end
          daddstr=char(32*ones(1,indp-indp2));
          if isfield(Q,'Bits'), Bits=Q.Bits; else Bits=Q.Precision+1; end
          fprintf(['Dither-index: %.0f, this value is added to the number (s',...
            num2str(Bits-1-Q.FractionBits),'.',num2str(Q.FractionBits),'+',...
            num2str(dbits),'): %g, in binary form:\n%s\n'],m,d,daddstr);
          fprintf('Number with the dither added: %f \n',x);
          fprintf(num2bin(x,Qx))
          fprintf('\n');
        end
        %elseif strncmp(Q.Type,'floating-point',5)
        %  warning('Dither is not yet implemented for floating-point')
      else
        error(['Unknown quantization type ''',Q.Type,''''])
      end
    end %dither on
  end %dither

  if strcmp(Q.Operation,'PQN')
    if strncmp(Q.Type,'floating-point',5)&&(Q.Precision>=50), error('Cannot add PQN when Precision>=50'), end
    if strncmp(Q.Type,'fixed-point',5)&&(Q.Bits>=49), error('Cannot add PQN when Bits>=49'), end
    q=lsb(x,Q);
    xprime=x+q*(rand(size(x))-0.5);
  elseif isstruct(Q)&&isfield(Q,'Type')&&strncmp(Q.Type,'floating-point',5)
    [xprime,ovfind]=qfloat(x,Q);
  elseif isstruct(Q)&&isfield(Q,'Type')&&strncmp(Q.Type,'fixed-point',5)
    [xprime,ovfind]=qfix(x,Q);
    if strcmpi(Q.Dither,'on')&&list_quant_details
      xpstr=num2bin(xprime,Q);
      if strcmp(xpstr(1),'0'), xpstr(1)='+'; else xpstr(1)='-'; end
      fprintf('Quantized result: %f, binary form: %s\n',xprime,xpstr);
    end
    qname=inputname(2);
    if isempty(qname)
      if strcmp(Q.Dither,'on'), warning('Quantizer is not referred to as a variable'), end %#ok<WNTAG>
    else assignin('caller',qname,Q);
    end
  end
  if (nargout>=3)&&~(isfield(Q,'Dither')&&strcmp(Q.Dither,'on'))
    %double quantization?
    dq=[0,0];
    if abs(xprime-x)==0.5*lsb(x,Q)
      if xprime-x>0, dq=[0,0.5*lsb(x,Q)];
      else dq=[-0.5*lsb(x,Q),0];
      end
    end
    varargout={dq};
  end
else
  Q=qstruct('double');
  [xprime,ovfind]=qfloat(x,Q);
  if strcmpi(Q.Dither,'on')&&list_quant_details
    fprintf('Quantized result: %f \n',x);
    fprintf(num2bin(xprime,Q))
  end
  if nargout>=3, dq=[0,0]; varargout={dq}; end
  return
end
%
%End of roundoff