function [y,dcomp]=compress(x,Delta,mode,muorA)
%COMPRESS  Compressor for floating point quantization
%
%        [y,dcomp]=compress(x,Delta,mode)
%
%        Delta: parameter of the compressor (see book "Quantization Noise")
%           If string, 'd*' means IEEE double-precision floating-point: Delta=2^(-1022), 
%           's*' means IEEE single precision fp: Delta=2^(-126)
%           Default: Delta = 1
%        mode: '' or 'p*' or 'f*' for piecewise linear floating-point compressor,
%           'u' (upper) for upper approximate compressor (y=x for -2*Delta <= x <= 2*Delta)
%           'l' (lower) for lower approximate compressor (y=x for -1.5*Delta <= x <= 1.5*Delta)
%           's' (sinh) or 'a' (asinh) for asinh approximation
%           'm' (mu-law) or 'A' (A-law)
%                For these cases V=128*Delta, and a last parameter (mu or A) may also be given. 
%                Defaults: mu = 255, A = 87.6
%                The output can be rescaled to match the other compressors 
%                by y*(1+log_2(V))/V = y/16 
%           Default: 'piecewise'
%        dcomp: derivative of the compressor at the given points
%           (not yet implemented for all)
%
%        Example (delete the unnecessary):
%          x=[0:0.01:128]; cla, hold off
%          plot(x,compress(x,1,'asinh'),'k'), hold on
%          plot(x,compress(x,1,'u'),'r')
%          plot(x,compress(x,1,'lower'),'g')
%          plot(x,compress(x,1,'mu-law')/16,'c')
%          plot(x,compress(x,1,'A-law')/16,'m')
%          plot(x,compress(x,1,'p'),'b')
%          hold off, zoom on, shg
%
%        See also EXPAND

%	Copyright (c) I. Kollar, 2005, all rights reserved
%	Last modified: 19-Dec-2005

if nargin<2, Delta=[]; end, if isempty(Delta), Delta=1; end
if isstr(Delta)
  if strncmp(Delta,'double',1), Delta=2^(-1022);
  elseif strncmp(Delta,'single',1), Delta=2^(-126);
  else error(['Delta is not allowed as ''',Delta,''''])
  end
end
if ~isnumeric(Delta)|(length(Delta)~=1), error('Delta is not a scalar'), end
if (Delta<=0)|~isfinite(Delta), error('Delta is not positive'), end
%
if nargin<3, mode=''; end, if isempty(mode), mode='piecewise'; end
if any(isnan(mode)), error('mode is not a string'), end
%
if nargin<4, muorA=[]; end
%
if strncmp(mode,'piecewise',1)|strncmp(mode,'floating',1)
  %piecewise linear
  k=x; ind=find(abs(x)>2*Delta);
  k(ind)=floor(log2(abs(x(ind))/Delta));
  y=pow2(x,-k)+sign(x).*k*Delta;
  ind2=find(abs(x)<=2*Delta);
  if ~isempty(ind2), y(ind2)=x(ind2); end
  %
  if nargout>1
    [xmant,xexp]=log2(abs(x/Delta));
    dcomp=2.^(-xexp+1);
    %jumps:
    indm=find(rem(y,1)==0); if ~isempty(indm), dcomp(indm)=1.5*dcomp(indm); end
    %center:
    ind=find(xexp<=1); if ~isempty(ind), dcomp(ind)=ones(size(ind)); end
  end
elseif strncmp(mode,'asinh',1)|strncmp(mode,'sinh',1)
  %asinh compressor
  y=Delta*asinh(x/Delta)/log(2);
  if nargout>1
    dcomp=1./sqrt(1+x.^2)/log(2);
  end
elseif strncmp(mode,'upper',1)
  %upper approximate compressor for floating point quantization
  y=Delta*sign(x).*(log(abs(x+realmin)/Delta)/log(2)+1);
  ind=find(abs(x)<2*Delta);
  y(ind)=x(ind);
  if nargout>1
    dcomp=1./abs((x+realmin)/Delta)/log(2);
    if ~isempty(ind), dcomp(ind)=ones(size(ind)); end
  end
elseif strncmp(mode,'lower',1)
  %lower approximating compressor for floating point quantization
  y=Delta*sign(x).*(log(abs(4/3*x+realmin)/Delta)/log(2)+0.5);
  ind=find(abs(x)<=1.5*Delta);
  y(ind)=x(ind);
  %
  if nargout>1
    dcomp=4/3./abs(4/3*(x+realmin)/Delta)/log(2);
    dcomp(ind)=ones(size(ind));
  end
elseif strncmp(mode,'mu-law',1)
  %mu-law 
  mu=255; V=128*Delta;
  if any(abs(x)>V), warning('For an element of x, x < -V  or  V < x'), end
  if ~isempty(muorA)
    if isnumeric(muorA)&isfinite(muorA)&(length(muorA)==1)&(muorA>0), mu=muorA; 
    else error(sprintf('mu = %.0g is not allowed',muorA))
    end
  end
  y=V*log(1+mu*abs(x)/V)./log(1+mu).*sign(x);
  %
  if nargout>1
    error('Derivative is not yet ready')
  end
elseif strncmp(mode,'A-law',1)
  %A-law
  A=87.6; V=128*Delta;
  if any(abs(x)>V), warning('For an element of x, x < -V  or  V < x'), end
  if ~isempty(muorA)
    if isnumeric(muorA)&isfinite(muorA)&(length(muorA)==1)&(muorA>1), A=muorA; 
    else error(sprintf('A = %.0g is not allowed',muorA))
    end
  end
  y=V*(1+log(A*(abs(x)+realmin)/V))./(1+log(A)).*sign(x);
  ind=abs(x)<V/A; if ~isempty(ind), y(ind)=A/(1+log(A))*x(ind); end
  %
  if nargout>1
    error('Derivative is not yet ready')
  end
else
  error(['mode is not allowed as ''',mode,''''])
end