function [cf,cf1,cf2,cf3]=cfchi2(u,sigma,k,mu)
%CFCHI2 Characteristic function of a chi-squared random variable.
%
%	[cf,cf1,cf2,cf3]=cfchi2(u,sigma,k,mu)
%
%	Definitions:
%	f(x) = 1/sigma * ...
%	        ((x/sigma)^((k-2)/2)*exp(-(x/sigma)/(2)) / (2^(k/2)*gamma(k/2))
%		x>=0, sigma>0
%	F(u) = (1/(1-j*2*sigma*u))^(k/2)
%
%	Output arguments:
%       cf = values of the characteristic function at u
%       cf1 = first derivative of the characteristic function at u
%       cf2 = second derivative of the characteristic function at u
%       cf3 = third derivative of the characteristic function at u
%
%	Input arguments:
%	u = values of the input argument (may be a vector or an array)
%	sigma = scaling factor of the standard deviation
%	k = degrees of freedom
%	mu = additional mean value (default value: zero)
%
%	See also: CFCALC, CFGAMMA, CFEXPON

%       Copyright (c) I. Kollar, 1994
%       All rights reserved.
%       Last modified: 01-Dec-1994

if nargin<4, mu=[]; end, if isempty(mu), mu=0; end
if (length(mu)>1)&(length(u)==1), u=u*ones(size(mu));
elseif (length(mu)>1)&any(size(mu)~=size(u))
  error('The size of mu differs from that of u')
end
if nargin<3, error('k is not given'), end
if length(k)>1, error('k is not a scalar'), end
if k<1, error('k is less than 1'), end
if rem(k,1)~=0, error('k is not integer'), end
if nargin<2, sigma=[]; end, if isempty(sigma), sigma=1; end
if length(sigma)>1, error('sigma is not a scalar'), end
if imag(sigma)~=0, error('sigma is not real'), end
if sigma<=0, error('sigma is not positive'), end
if any(any(imag(u)~=0)), error('u is not real'), end
%
cf=(1./(1-j*2*sigma*u)).^(k/2);
cf1=[]; cf2=[]; cf3=[];
if nargout>=2
  term=1./(1-j*2*sigma*u);
  cf1=(k/2)*j*2*sigma*cf.*term;
end
if nargout>=3
  cf2=(k/2+1)*j*2*sigma*cf1.*term;
end
if nargout>=4
  cf3=(k/2+2)*j*2*sigma*cf2.*term;
end
%
%Combine with mu if it differs from zero
if mu~=0
  [cf,cf1,cf2,cf3]=cfmu(u,mu,cf,cf1,cf2,cf3);
end %mu~=0
%
%End of cfchi2
