function [cf,cf1,cf2,cf3]=cfgaussian(u,sigma,mu)
%CFGAUSSIAN Characteristic function and derivatives of a Gaussian variable.
%
%	[cf,cf1,cf2,cf3]=cfgaussian(u,sigma,mu)
%
%	Definitions:
%	f(x) = 1/(sqrt(2*pi)*sigma)*exp(-(x)^2/(2*sigma^2))
%		sigma>0
%	F(u) = exp(-sigma^2*u^2/2)
%
%	Output arguments:
%       cf = values of the characteristic function at u
%       cf1 = first derivative of the characteristic function at u
%       cf2 = second derivative of the characteristic function at u
%       cf3 = third derivative of the characteristic function at u
%
%	Input arguments:
%	u = values of the input argument (may be a vector or an array)
%	sigma = standard deviation (default value: 1)
%	mu = mean value (default value: zero)

%	See also:

%       Copyright (c) I. Kollar, 1994
%       All rights reserved.
%       Last modified: 01-Dec-1994

if nargin<3, mu=[]; end, if isempty(mu), mu=0; end
if length(mu)>1, error('mu is not a scalar'), end
if imag(mu)~=0, error('mu is not real'), end
if nargin<2, sigma=[]; end, if isempty(sigma), sigma=1; end
if length(sigma)>1, error('sigma is not a scalar'), end
if imag(sigma)~=0, error('sigma is not real'), end
if sigma<=0, error('sigma is not positive'), end
if any(any(imag(u)~=0)), error('u is not real'), end
%
cf=exp(-u.^2*sigma^2/2);
cf1=[]; cf2=[]; cf3=[];
if nargout>=2
  cf1=(-sigma^2*u).*cf;
end
if nargout>=3
  cf2=(-sigma^2+sigma^4*u.^2).*cf;
end
if nargout>=4
  cf3=(3*sigma^4*u-sigma^6*u.^3).*cf;
end
%
%Combine with mu if it differs from zero
if mu~=0
  [cf,cf1,cf2,cf3]=cfmu(u,mu,cf,cf1,cf2,cf3);
end %mu~=0
%
%End of cfgaussian
