function [cf,cf1,cf2,cf3]=cflaplace(u,alpha,mu)
%CFLAPLACE Characteristic function and derivatives of a Laplacian random variable.
%
%	[cf,cf1,cf2,cf3]=cflaplace(u,alpha,mu)
%
%	Definitions:
%	f(x) = alpha/2*exp(-alpha*abs(x)), alpha>0
%	F(u) = alpha^2/(u^2+alpha^2)
%
%	Output arguments:
%       cf = values of the characteristic function at u
%       cf1 = first derivative of the characteristic function at u
%       cf2 = second derivative of the characteristic function at u
%       cf3 = third derivative of the characteristic function at u
%
%	Input arguments:
%	u = values of the input argument (may be a vector or an array)
%	alpha = parameter of the distribution (default=1)
%	mu = mean value (default value: zero)
%
%	See also: CFCALC

%       Copyright (c) I. Kollar, 1994
%       All rights reserved.
%       Last modified: 01-Dec-1994

if nargin<3, mu=[]; end, if isempty(mu), mu=0; end
if length(mu)>1, error('mu is not a scalar'), end
if imag(mu)~=0, error('mu is not real'), end
if nargin<2, alpha=[]; end, if isempty(alpha), alpha=1; end
if length(alpha)>1, error('alpha is not a scalar'), end
if imag(alpha)~=0, error('alpha is not real'), end
if alpha<=0, error('alpha is not positive'), end
if any(any(imag(u)~=0)), error('u is not real'), end
%
if isfinite(alpha), cf=alpha^2./(u.^2+alpha^2);
elseif isnan(alpha), cf=NaN*ones(size(u));
else cf=ones(size(u));
end
cf1=[]; cf2=[]; cf3=[];
if nargout>=2
  cf1=-2*alpha^2*u./(u.^2+alpha^2).^2;
end
if nargout>=3
  cf2=-2*alpha^2*(-3*u.^2+alpha^2)./(u.^2+alpha^2).^3;
end
if nargout>=4
  cf1=24*alpha^2*u.*(-u.^2+alpha^2)./(u.^2+alpha^2).^4;
end
%
%Combine with mu if it differs from zero
if mu~=0
  [cf,cf1,cf2,cf3]=cfmu(u,mu,cf,cf1,cf2,cf3);
end %mu~=0
%
%End of cflapl
