function [cf,cf1,cf2,cf3]=cfsine(u,A,mu)
%CFSINE Characteristic function and derivatives of a sinusoidal random variable.
%
%	[cf,cf1,cf2,cf3]=cfsine(u,A,mu)
%
%	Definitions:
%	f(x) = 1/(A*pi*sqrt(1-(x/A)^2)) , abs(x)<A, A>0
%	F(u) = J_0(A*u)
%
%	Output arguments:
%       cf = values of the characteristic function at u
%       cf1 = first derivative of the characteristic function at u
%       cf2 = second derivative of the characteristic function at u
%       cf3 = third derivative of the characteristic function at u
%
%	Input arguments:
%	u = values of the input argument (may be a vector)
%	A = amplitude (default value: 1)
%	mu = mean value (default value: zero)
%
%	See also: CFCALC

%       Copyright (c) I. Kollar, 1994
%       All rights reserved.
%       Last modified: 01-Dec-1994

if nargin<3, mu=[]; end, if isempty(mu), mu=0; end
if length(mu)>1, error('mu is not a scalar'), end
if imag(mu)~=0, error('mu is not real'), end
if nargin<2, A=[]; end, if isempty(A), A=1; end
if length(A)>1, error('A is not a scalar'), end
if imag(A)~=0, error('A is not real'), end
if A<=0, error('A is not positive'), end
if any(any(imag(u)~=0)), error('u is not real'), end
%
cf=bessel(0,u*A);
cf1=[]; cf2=[]; cf3=[];
if nargout>=2
  b1u=bessel(1,u*A);
  cf1=A*(-b1u);
  %cf1=A*(-bessel(1,u*A));
end
if nargout>=3
  ind=find(abs(u)<eps); u(ind)=eps*ones(size(ind));
  b2u=bessel(2,u*A);
  cf2=A*(-A)*(-b2u+b1u./(u*A));
  %cf2=A*(-A)*(-bessel(2,u*A)+bessel(1,u*A)./(u*A));
end
if nargout>=4
  cf3=(-A^2)*(-A)*(-bessel(3,u*A)+2*b2u./(u*A)-(-b2u+b1u./(u*A))./(u*A));
end
%
%Combine with mu if it differs from zero
if mu~=0
  [cf,cf1,cf2,cf3]=cfmu(u,mu,cf,cf1,cf2,cf3);
end %mu~=0
%
%End of cfsine
