function X=rounddft_chirp(x,Q,runmod)
%ROUNDDFT_CHIRP Quantized DFT algorithm based on the chirp function
%
%   X=rounddft_chirp(x,Q,runmod)
%
%   Input arguments:
%     x = input sequence
%     Q = definition of the quantizer (if not given, IEEE double)
%     runmod = run modifiers, see roundfft
%   Output arguments:
%     X = vector, or structure for block-float, fields for bf: data, bfe
%   Example:
%     roundifft(roundfft([1:13]))-[1:13]

%	Copyright (c) I. Kollar, 2006, all rights reserved
%   Last modified: 11-Nov-2006, IK

if nargin<3, runmod=[]; end
if nargin<2, Q='double'; end, if isempty(Q), Q='double'; end
if isstruct(x), bfe=x.bfe; x=x.data;
else bfe=0;
end
N=length(x);
%
Qcos=[]; Qinput=[]; direction='';
if isstr(runmod)
  if strncmpi(runmod,'inverse',1), runmod=struct('direction','inverse');
  elseif strncmpi(runmod,'block-float',2), runmod=struct('exponent','block-float');
  else error('Unknown runmod')
  end
elseif ~isempty(runmod)&~isstruct(runmod)
  error('runmod is not a structure')
end
if isstruct(runmod)
  fn=fieldnames(runmod);
  for ii=1:length(fn)
    if strcmp(fn{ii},'Qinput')
      Qinput=runmod.Qinput;
    elseif strcmp(fn{ii},'Qsin')
      Qcos=runmod.Qsin;
    elseif strcmp(fn{ii},'Qcos')
      Qcos=runmod.Qcos;
    elseif strcmp(fn{ii},'direction')
      direction=runmod.direction;
    elseif strcmp(fn{ii},'algorithm')
    elseif strcmp(fn{ii},'exponent')
    elseif strcmp(fn{ii},'sign_of_life')
    else error(['Unrecognised runmod field ''',fn{ii},''''])
    end
  end %for ii
end
if isempty(direction), direction='forward'; end

if isstr(Q), Q=qstruct(Q); end
if isfield(runmod,'exponent'), exponent=runmod.exponent;
  if strncmpi(Q.Type,'floating-point',2)
    error('Cannot implement block-float for floating-point quantizer')
  end
else exponent='';
end

if isempty(Qcos)
  if strcmp(Q.Type,'fixed-point')
    Qcos=qstructset(Q,'FractionBits',Q.Bits-2,...
      'Maximum',2-2^(-(Q.Bits-2)),'Minimum',-2); %amplitude of sine is 1, range is up to 2
    if ~strcmp(Q.Coding,'t'), Qcos=qstructset(Qcos,'Minimum',-Qcos.Maximum); end
  else
    Qcos=Q;
  end
end
if isempty(Qinput), Qinput=Q; end
%
if strncmpi(exponent,'block-float',1)
  %scale input to maximum range
  while all(real(x)<=Qinput.Maximum/2)&all(real(x)>=Qinput.Minimum/2)&...
      all(imag(x)<=Qinput.Maximum/2)&all(imag(x)>=Qinput.Minimum/2)&...
      (any(real(x))|any(imag(x)))
    x=x*2; bfe=bfe-1;
  end
  while any(real(x)>Qinput.Maximum)|any(real(x)<Qinput.Minimum)|...
      any(imag(x)>Qinput.Maximum)|any(imag(x)<Qinput.Minimum)
    x=x/2; bfe=bfe+1;
  end
end
[x,ovfind]=roundq(x,Qinput);
if strncmpi(exponent,'block-float',1)
  %scale input to maximum range
  while all(real(x)<=Q.Maximum/2)&all(real(x)>=Q.Minimum/2)&...
      all(imag(x)<=Q.Maximum/2)&all(imag(x)>=Q.Minimum/2)&...
      (any(real(x))|any(imag(x)))
    x=x*2; bfe=bfe-1;
  end
  while any(real(x)>Q.Maximum)|any(real(x)<Q.Minimum)|...
      any(imag(x)>Q.Maximum)|any(imag(x)<Q.Minimum)
    x=x/2; bfe=bfe+1;
  end
  x=roundq(x,Q);
elseif ~isempty(ovfind)
  error('Overflow error during input scaling')
end
%
Nm=pow2(ceil(log2(N))+1); sx=size(x);
persistent Wnk WNK WNKi Qcos_saved N_saved exponent_saved direction_saved
%N_saved=[]; %Force recalculation
%WNK: IDFT of Wnk, WNKi: DFT of Wnk 
if isequal(N,N_saved)&isequal(length(Wnk),Nm)&isequal(Qcos,Qcos_saved)&...
    isequal(exponent,exponent_saved) %& ~isequal(direction,direction_saved)
  if (~strncmpi(direction,'inverse',1)&strncmpi(direction_saved,'inverse',1)) | ...
      (strncmpi(direction,'inverse',1)&~strncmpi(direction_saved,'inverse',1))
    Wnk=conj(Wnk);
    if ~isempty(WNKi)
      tmp=WNK; WNK=WNKi; WNKi=tmp;
    else
      WNKi=WNK;
      if strcmp(Q.Type,'fixed-point')&~strncmpi(exponent,'block-float',1)
        %fixed-point but not block float: use precalculated result
        [WNK,ovfind]=roundq(fft(conj(Wnk)),Q);
        if ~isempty(ovfind), error('Overflow error during calculation of WNK'), end
      else
        WNK=roundfft(conj(Wnk),Q,setfield(runmod,'direction','forward'));
      end
    end
    direction_saved=direction;
  end
else %something has been changed
  nk=rem(([0:floor((Nm-1)/2),floor(Nm/2):-1:1]'.^2)/2,N);
  if ~strncmpi(direction,'inverse',1), [Wnk,ovfind]=roundq(exp(-j*2*pi*nk/N),Qcos);
  else [Wnk,ovfind]=roundq(exp(j*2*pi*nk/N),Qcos);
  end
  if ~isempty(ovfind), error('Overflow during calculation of Wnk'), end
  if strcmp(Q.Type,'fixed-point')&~strncmpi(exponent,'block-float',1)
    [WNK,ovfind]=roundq(fft(conj(Wnk)),Q);
    if ~isempty(ovfind), error('Overflow during calculation of WNK'), end
  else
    WNK=roundfft(conj(Wnk),Q,setfield(runmod,'direction','forward'));
  end
  WNKi=[];
  N_saved=N; Qcos_saved=Qcos; exponent_saved=exponent; direction_saved=direction;
end
%
x(Nm)=0;
[xw,ovf]=roundq(x(:).*Wnk,Q); %chirp 1
if strncmpi(exponent,'block-float',1)&any(ovf)
  x=roundq(x/2,Q); bfe=bfe+1;
  [xw,ovf]=roundq(x(:).*Wnk,Q);
elseif any(ovf)
  error('Overflow error after multiplication with W(n^2)')
end
%
Xw=roundfft(xw,Q,setfield(runmod,'direction','forward'));
if strncmpi(exponent,'block-float',1)
  bfe=bfe+Xw.bfe+WNK.bfe;
  X=roundifft(Xw.data.*WNK.data,Q,runmod); %chirp 2
  bfe=bfe+X.bfe; X=X.data;
else
  X=roundifft(Xw.*WNK,Q,runmod);
end
[X,ovf]=roundq(X(1:N).*Wnk(1:N),Q); %chirp 3
if strncmpi(exponent,'block-float',1)&any(ovf)
  X=roundq(X/2,Q); bfe=bfe+1;
  [X,ovf]=roundq(X(1:N).*Wnk(1:N),Q);
elseif any(ovf)
  error('Overflow error during multiplication with W(k^2)')
end
%
X=reshape(X,sx(1),sx(2));
%
if strncmpi(direction,'inverse',1)
  if strncmpi(exponent,'block-float',1)
    bfe=bfe-floor(log2(N));
    [X,ovf]=roundq(X*pow2(floor(log2(N)))/N,Q,'complex');
  else
    X=roundq(1/N*X,Q,'complex'); 
  end
end
if strncmpi(exponent,'block-float',1)
  X=struct('data',X,'bfe',bfe);
end
%End of rounddft_chirp
