function [xprime,ovfind]=qfix(x,varargin)
%QFIX Fixed-point quantizer
%
%   xprime=qfix(x,runmod) executes fixed-point rounding (quantization) 
%      on real or complex, scalar or array x.
%      qfix(x,runmod,'list') also lists the settings of the quantizer
%      implemented.
%   Possibilities for parameters in the structure runmod:
%     runmod.B: number of bits of the number, sign included
%     runmod.p: number of bits of the number, sign not included
%     runmod.fb: number of fractional bits
%     runmod.rdir: direction of quantization: 'r', 'f', 'c' 'x', 'i', 't'
%         for round, floor, ceil, fix, toinf, trunc
%     runmod.treath: way of treating *.5 values for round: 'f', 'c', 'x', 'e', 'o', 'r'
%         for floor, ceil, fix, toinf, toeven, toodd, rand
%     runmod.coding = coding of mantissa, 's', 'o', 't', 'u'
%         for sign-magnitude, one's complement, two's complement, unsigned 
%     runmod.ovf: overflow handling, 'c', 'm', 't', 'i', 'n'
%         for clip, modular, triang, inf, none
%     runmod.xnmin: minimum representable (negative) number
%     runmod.xpmax: maximum representable number
%   Default values:
%     B=16, p=15, fb=0, rdir='r', treath='toeven', coding='t', ovf='m', 
%         xnmin=-2^(p-fb-1) for 'two''s', else xnmin=-2^(p-fb-1)+1,
%         xpmax=2^(B--fb-1)-1 
%     Limits: B<=54, p<=53
%  	Output arguments:
%	    xprime = quantized variable
%	    ovfind = indices of elements of x which cause overflow, indices
%		       of negative overflow are multiplied by (-1), indices of
%		       imaginary parts furthermore by sqrt(-1)
%   Examples:
%     qfix(1.75,qstruct('B',4,'fb',0)) %2
%   Remark: type 'qfix remark'

%	Copyright (c) I. Kollar, 1995-2006, all rights reserved
%	Last modified: 5-Feb-2006

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Old-form call:
%function [xprime,ovfind]=qfix(x,p,fb,rdir,treath,coding,ovf,xnmin,xpmax)
%QFIX Fixed-point arithmetic quantizer
%
%	[xprime,ovfind]=qfix(x,B,fb,rdir,treath,coding,ovf,xnmin,xpmax)
%
%	Input arguments:
%	x = input variable (scalar or array, may be complex)
%	B = bit number, including sign
%	fb = number of fraction bits
%	rdir = direction of quantization: 'r', 'f', 'c' 'x', 'i', 't'
%              for round, floor, ceil, fix, toinf, trunc
%	treath = way of treating *.5 values for round: 'f', 'c', 'x', 'e', 'o', 'r' 
%              for floor, ceil, fix, toinf, toeven, toodd, rand
%	coding = number representation: 'o', 't', 's'
%              for one's, two's, sign-mag
%	ovf = overflow handling, 'c', 'm', 't', 'i', 'n'
%              for clip, modular, triang, inf, none
%	xnmin = minimum representable (negative) number 
%	xpmax = maximum representable number
%
%	Output arguments:
%	xprime = quantized variable
%	ovfind = indices of elements of x which cause overflow, indices
%		of negative overflow are multiplied by (-1), indices of
%		imaginary parts furthermore by sqrt(-1)
%
%	Default values: fb=0, rdir='r', treath='e', coding='t',
%		ovf='m', xpmax=2^(p-fb-1)-1
%	xnmin=-2^(p-fb-1) for 'two''s', else xnmin=-2^(p-fb-1)+1

if (nargin==1)&isstr(x)&strncmpi(x,'remark',3), type qf_remark.m, return, end %remark
if nargin>1
  %try, 
    if isstruct(varargin{1}), Q=varargin{1};
    else Q=qstruct(varargin{1});
    end
  %catch, Q=[];
  %end
  if isstruct(Q)&isfield(Q,'Type')&strncmp(Q.Type,'floating-point',5)
    [xprime,ovfind]=qfloat(x,Q);
    return
  end
else
  Q=qstruct('double');
end

makeqlist=''; %list quantizer parameters
if nargin>=2
  if isstr(varargin{end})
    if strcmp(varargin{end},'list')
      makeqlist='list'; varargin(end)=[];
    end
  end
end
%
p=[]; B=[]; fb=''; rdir=''; treath=''; coding=''; ovf=''; xnmin=[]; xpmax=[]; typ='';
if ~isempty(Q) 
  fn=fieldnames(Q);
  for ii=1:length(fn)
    if strcmp(fn{ii},'Type'), typ=Q.Type;
    else
      %eval([fn{ii},'=getfield(Q,fn{ii});'])
      if strcmp(fn{ii},'Name'), name=Q.Name;
      elseif strcmp(fn{ii},'Precision'), p=Q.Precision;
      elseif strcmp(fn{ii},'Bits'), B=Q.Bits;
      elseif strcmp(fn{ii},'FractionBits'), fb=Q.FractionBits;
      elseif strcmp(fn{ii},'RoundingDir'), rdir=Q.RoundingDir;
      elseif strcmp(fn{ii},'TreatHalf'), treath=Q.TreatHalf;
      elseif strcmp(fn{ii},'Coding'), coding=Q.Coding;
      elseif strcmp(fn{ii},'OverFlow'), ovf=Q.OverFlow;
      elseif strcmp(fn{ii},'Minimum'), xnmin=Q.Minimum;
      elseif strcmp(fn{ii},'Maximum'), xpmax=Q.Maximum;
      elseif strcmp(fn{ii},'Emin'), Emin=Q.ExpMinimum;
      elseif strcmp(fn{ii},'Emax'), Emax=Q.ExpMaximum;
      elseif strcmp(fn{ii},'Ebias'), Ebias=Q.ExpBias;
      elseif strcmp(fn{ii},'Operation'), op=Q.Operation;          
      elseif strcmpi(fn{ii},'Dither'), dith=Q.Dither; 
      elseif strcmpi(fn{ii},'ToD'), tod=Q.ToD;
      elseif strcmpi(fn{ii},'Number_of_DitherBits'), L=Q.Number_of_DitherBits;
      elseif strcmpi(fn{ii},'Amplitude_of_Sinusoid_Dither'), A=Q.Amplitude_of_Sinusoid_Dither;  
      elseif strcmpi(fn{ii},'Ratio_of_Frequencies')
          RoFn=Q.Ratio_of_Frequencies_numerator; 
          RoFd=Q.Ratio_of_Frequencies_denumerator;
      elseif strcmpi(fn{ii},'Phase'), ph=Q.Phase;
      elseif strcmpi(fn{ii},'Variance_of_Gaussian_Dither'), gvar=Q.Variance_of_Gaussian_Dither;   
      elseif strcmpi(fn{ii},'DFunction'), dfun=Q.Dfunction; 
      elseif strcmpi(fn{ii},'Sequenceperm'), seqp=Q.Sequenceperm;
      elseif strcmpi(fn{ii},'Other'), other=Q.Other;
      else error(['Unknown field ''',fn{ii},''''])
      end
    end
  end
  if ~exist('p')|isempty(p), p=Q.Bits-~strncmp(Q.Coding,'u',1); end
else %numeric B or old call
  %	[xprime,ovfind]=qfix(x,B,fb,rdir,treath,coding,ovf,xnmin,xpmax)
  if ~isempty(varargin), B = varargin{1}; end
  if length(varargin)>=2, fb=varargin{2}; end
  if length(varargin)>=3, rdir=varargin{3}; end
  if length(varargin)>=4, treath=varargin{4}; end
  if length(varargin)>=5, coding=varargin{5}; end
  if length(varargin)>=6, ovf=varargin{6}; end
  if length(varargin)>=7, xnmin=varargin{7}; end
  if length(varargin)>=8, xpmax=varargin{8}; end
  if ~exist('p')|isempty(p), p=B-~strncmp(coding,'u',1); end
end
%
if  ~isempty(typ)&~strncmpi(typ,'fixed-point',3), error(['type ''',typ,''' is not fixed-point']), end
if length(p)>1, error('p is not scalar'), end
if p<=1, error('bit number is not positive'), end
if rem(p,1)~=0, error('p is not integer'), end
if ~isempty(B)&~isempty(p)&~isequal(B,p+~strncmp(coding,'u',1))
  error('Both p and B are given, inconsistently with coding')
end
if isempty(p), p=15; end
if isempty(B), B=p+1; end
if ~isequal(rem(B,1),0), error(sprintf('B = %.5g is not integer',B)), end
if B>54, error(sprintf('B = %.f is larger than 54',B)), end
%
if isnan(fb), fb=[]; end, if isempty(fb), fb=0; end %decimal point at right
if fb<0, error('fb is negative'), end, if rem(fb,1)~=0, error('fb is not integer'), end
if fb>B-~strncmp(coding,'u',1), error('fb is larger than number of binary bits'), end
if isnan(rdir), rdir=''; end, if isempty(rdir), rdir='r'; end %rounding
if isnan(treath), treath=''; end, if isempty(treath), treath='e'; end %round to even
if isnan(coding), coding=''; end, if isempty(coding), coding='t'; end %two's complement
if strncmp(coding,'two''s complement',1)|strncmp(coding,'one''s complement',1)|...
    strncmp(coding,'sign-magnitude',1)|strncmp(coding,'unsigned',1)
else error(['coding=''',coding,''' is not allowed'])
end
if isnan(ovf), ovf=''; end
if isempty(ovf), ovf='m'; end %modular
if isnan(xnmin), xnmin=[]; end
if strncmp(coding,'t',1), xnminl=-pow2(B-fb-1); else xnminl=-pow2(pow2(B-1)-1,-fb); end
if isempty(xnmin), xnmin=xnminl; end
if xnmin<xnminl, error('xnmin is too small for the given bit number'), end
if isnan(xpmax), xpmax=[]; end
xpmaxl=pow2(pow2(B-1)-1,-fb);
if isempty(xpmax), xpmax=xpmaxl; end
if xpmax>xpmaxl, error('xpmax is too large for the given bit number'), end
if xpmax<xnmin, error('xpmax < xnmin'), end
%
if all(imag(x(:))==0) %real
  xr=1; xsc=pow2(x,fb); xcompl=0;
else %complex
  xr=0; xsc=pow2([real(x),imag(x)],fb); xcompl=1;
end
%xsc: to be rounded to an integer number
if strncmp(rdir,'t',1) %truncate
  if strncmp(coding,'s',1), rdir='x'; %fix for sign-magnitude
  else rdir='f'; %floor for the other cases
  end
end
pmax=-log2(eps)+1; %53 for ieee
if (B<pmax+1)|(abs(xsc)<2^(pmax-1)) %only if p<53 for ieee; otherwise there is no fraction in xsc
  if strncmp(rdir,'r',1) %round
    if strncmp(treath,'e',1)|strncmp(treath,'toeven',3) %shortcut for convergent rounding
      p2m1=pow2(pmax-1);
      xprime=(xsc+sign(xsc)*p2m1)-sign(xsc)*p2m1; %quantize
      1;
    else %not convergent or not ieee
      xprime=round(xsc);
      hind=find(abs(rem(xsc,1))==0.5);
      if ~isempty(hind) %half-values present
        if strncmp(treath,'x',1)
          xprime(hind)=fix(xsc(hind));
        elseif strncmp(treath,'c',1)
          xprime(hind)=ceil(xsc(hind));
        elseif strncmp(treath,'f',1)
          xprime(hind)=floor(xsc(hind));
        elseif strncmp(treath,'i',1)|strncmp(treath,'toinf',3) %toinf, done
        elseif strncmp(treath,'e',1)|strncmp(treath,'o',1)|strncmp(treath,'toeven',3)|strncmp(treath,'toodd',3) 
          %toeven,toodd
          uind=find(abs(rem(xsc(hind),2))>1); uind=hind(uind);
          lind=find(abs(rem(xsc(hind),2))<1); lind=hind(lind);
          if strncmp(treath,'e',1) %toeven
            if ~isempty(uind), xprime(uind)=fix(xsc(uind))+sign(xsc(uind)); end
            if ~isempty(lind), xprime(lind)=fix(xsc(lind)); end
          else %toodd
            if ~isempty(lind), xprime(lind)=fix(xsc(lind)+sign(xsc(lind))); end
            if ~isempty(uind), xprime(uind)=fix(xsc(uind)); end
          end
        elseif strncmp(treath,'r',1)
          xprime(hind)=round(xsc(hind)+rand(size(hind))-0.5);
        else
          error(['treath=''',treath,''' is not allowed'])
        end
      end %half-values
    end %not convergent or not ieee
  elseif strncmp(rdir,'i',1) %toinf
    xprime=ceil(xsc);
    nind=find(x<0); if ~isempty(nind), xprime(nind)=floor(xsc(nind)); end
  elseif strcmp(rdir,'x')|strncmp(rdir,'fix',2), xprime=fix(xsc);
  elseif strncmp(rdir,'c',1), xprime=ceil(xsc);
  elseif strcmp(rdir,'f')|strncmp(rdir,'floor',2), xprime=floor(xsc);
  else error(['rdir ''',rdir,''' is not allowed'])
  end
else %B==54
  xprime=xsc;
end
%
largeind=find(xprime>pow2(xpmax,fb));
smallind=find(xprime<pow2(xnmin,fb));
%
if nargout>1 %ovfind requested
  infnani=find(isnan(xprime)); %MEG KELL NEZNI KELL-E?
  ovfind=sort([largeind(:);-smallind(:)]);
  doi=find(diff(ovfind)==0);
  if ~isempty(doi), ovfind(doi)=[]; end %eliminate duplicate indices
end
if strncmp(ovf,'m',1) %modular overflow
  cyc=pow2(xpmax-xnmin,fb)+1; offs=pow2(xnmin,fb);
  xmin=min(min(xprime));
  if ~isfinite(xmin), xmin=min([0,min(xprime(isfinite(xprime)))]); end
  if B<pmax
    xprime=rem(xprime-offs-cyc*floor(xmin/cyc),cyc)+offs;
  else
    if xprime>pow2(xpmax,fb)
      xprime=xprime-fix(xprime/(pow2(xpmax,fb)-pow2(xnmin,fb)))*(pow2(xpmax,fb)-pow2(xnmin,fb));
    end
    if xprime<pow2(xnmin,fb)
      xprime=xprime+fix(-xprime/(pow2(xpmax,fb)-pow2(xnmin,fb)))*(pow2(xpmax,fb)-pow2(xnmin,fb));
    end
    %while xprime>pow2(xpmax,fb)
    %  xprime2=xprime-(pow2(xpmax,fb)-pow2(xnmin,fb));
    %  if xprime2<xprime, xprime=xprime2; else break, end
    %end
    %while xprime<pow2(xnmin,fb)
    %  xprime2=xprime+pow2(xpmax,fb)-pow2(xnmin,fb);
    %  if xprime2>xprime, xprime=xprime2; else break, end
    %end
  end
elseif strncmp(ovf,'c',1) %clip at min-max
  xprime(largeind)=ones(size(largeind))*pow2(xpmax,fb);
  xprime(smallind)=ones(size(smallind))*pow2(xnmin,fb);
elseif strncmp(ovf,'i',1) %infty for min-max
  xprime(largeind)=inf*ones(size(largeind));
  xprime(smallind)=-inf*ones(size(smallind));
elseif strncmp(ovf,'t',1) %triangular overflow
  cyc=pow2(B+1); xmin=min(min(xprime));
  if ~isfinite(xmin), xmin=min([0,min(xprime(isfinite(xprime)))]); end
  xprime=rem(xprime+cyc/2-cyc*floor(xmin/cyc),cyc)-cyc/2;
  ind=find(xprime>cyc/4); xprime(ind)=cyc/2-xprime(ind);
  ind=find(xprime<-cyc/4); xprime(ind)=-cyc/2-xprime(ind);
  ind=find(xprime>pow2(xpmax,fb)); xprime(ind)=pow2(xpmax,fb)*ones(size(ind));
  ind=find(xprime<pow2(xnmin,fb)); xprime(ind)=pow2(xnmin,fb)*ones(size(ind));
elseif strncmp(ovf,'n',1) %nothing to do
else error(['ovf=''',ovf,''' is not allowed'])
end
xprime=pow2(xprime,-fb);
%
if xr==0
  if size(xprime,2)>2
    xprime=reshape(xprime(1:size(x,2))+j*xprime(size(x,2)+[1:size(x,2)]),size(x));
  else
    xprime=reshape(xprime(:,1)+j*xprime(:,2),size(x));
  end
  indimag=find(ovfind>prod(size(x)));
  if ~isempty(indimag)
    ovfind(indimag)=j*(ovfind(indimag)-prod(size(x)));
  end
end
%
if strcmp(makeqlist,'list')
  qstructget(varargin{:})
  %Show quantized number
  if (length(xprime)==1)&~xcompl
    fprintf('  '), num2bin(xprime,struct('B',B,'fb',fb,'coding',coding),'d');
  end
end
%
%End of qfix