function [X,ovfind]=rounddft(x,Q,runmod)
%ROUNDDFT Quantized DFT algorithm
%
%   X=rounddft(x,Q,runmod)
%
%   Input arguments:
%     x = input sequence (if the length is not 2's power, call roundfft_non2p
%     Q = definition of the quantizer (if not given, IEEE double)
%     runmod = run modifiers
%        Qinput: (quantizer, default: Q)
%        Qacc: (accumulator quantizer, default: Q)
%        Qcos: (quantizer for trigonometric constants, default: Q)
%        direction: 'forward' (default), 'inverse'
%        exponent: 'block-float', 'const' (default for fixed-point quantizer)
%             'downscale': downscale after each addition/subtraction by 2
%                     for 'grouped' only
%             for floating-point, 'block-float' is not allowed 
%        sign_of_life: for nonzero, display . characters as proceeding
%        algorithm: for 'grouped', evaluate the sum in a tree, groups of 2
%   Output arguments:
%     X = vector, or structure for block-float 
%          fields for block-float: data, bfe
%
%   Example:
%     roundidft(rounddft([1:8]))-[1:8]

%	Copyright (c) I. Kollar, 2006, all rights reserved
%	Last modified: 23-Oct-2006

if nargin<3, runmod=[]; end, ovfind=[];
if nargin<2, Q='double'; end, if isempty(Q), Q='double'; end
if isstruct(x), bfe=x.bfe; x=x.data;
else bfe=0;
end
N=length(x);

Qcos=[]; Qinput=[]; Qacc=[]; sign_of_life=0; alg='';
if isstr(runmod)
  if strncmpi(runmod,'inverse',1), runmod=struct('direction','inverse');
  elseif strncmpi(runmod,'block-float',2), runmod=struct('exponent','block-float');
  elseif strncmpi(runmod,'downscaling',2), runmod=struct('exponent','downscaling');
  end
elseif isstruct(runmod)
  fn=fieldnames(runmod);
  for ii=1:length(fn)
    if strcmp(fn{ii},'Qinput')
      Qinput=runmod.Qinput;
    elseif strcmp(fn{ii},'Qacc')
      Qacc=runmod.Qacc;
    elseif strcmp(fn{ii},'Qsin')
      Qcos=runmod.Qsin;
    elseif strcmp(fn{ii},'Qcos')
      Qcos=runmod.Qcos;
    elseif strcmp(fn{ii},'direction')
    elseif strcmp(fn{ii},'exponent')
    elseif strcmp(fn{ii},'sign_of_life')
      sign_of_life=runmod.sign_of_life;
    elseif strcmp(fn{ii},'algorithm')
      alg=runmod.algorithm;
    else error(['Unrecognized runmod field ''',fn{ii},''''])
    end
  end %for ii
end
if ~isfield(runmod,'direction'), runmod.direction='f'; end

if isfield(runmod,'exponent')
  exponent=runmod.exponent;
  if strncmpi(Q.Type,'floating-point',2)&strncmpi(exponent,'block-float',1)
    error('Cannot implement block-float for floating-point quantizer')
  elseif strncmpi(Q.Type,'floating-point',2)&strncmpi(exponent,'downscaling',1)
    error('Cannot implement downscaling for floating-point quantizer')
  elseif ~strncmpi(alg,'grouped',2)&strncmpi(exponent,'downscaling',1)
    error('Can implement downscaling for grouped algorithm only')
  end
else exponent='';
end
if isstr(Q), Q=qstruct(Q); end
if isempty(Qcos)
  if strcmp(Q.Type,'fixed-point')
    Qcos=qstructset(Q,'FractionBits',Q.Bits-2,...
      'Maximum',2-2^(-(Q.Bits-2)),'Minimum',-2); %amplitude of sine is 1, range is up to 2
    if ~strcmp(Q.Coding,'t'), Qcos=qstructset(Qcos,'Minimum',-Qcos.Maximum); end
  else
    Qcos=Q;
  end
end
if isempty(Qinput), Qinput=Q; end
if isempty(Qacc), Qacc=Q; end
%
if strncmpi(exponent,'block-float',1)
  %scale input to maximum range
  while all(real(x)<=Qinput.Maximum/2)&all(real(x)>=Qinput.Minimum/2)&...
      all(imag(x)<=Qinput.Maximum/2)&all(imag(x)>=Qinput.Minimum/2)&...
      (any(real(x))|any(imag(x)))
    x=x*2; bfe=bfe-1;
  end
  while any(real(x)>Qinput.Maximum)|any(real(x)<Qinput.Minimum)|...
      any(imag(x)>Qinput.Maximum)|any(imag(x)<Qinput.Minimum)
    x=x/2; bfe=bfe+1;
  end
end
[x,ovf]=roundq(x,Qinput);
if strncmpi(exponent,'block-float',1)
  %scale input to maximum range
  while all(real(x)<=Q.Maximum/2)&all(real(x)>=Q.Minimum/2)&...
      all(imag(x)<=Q.Maximum/2)&all(imag(x)>=Q.Minimum/2)&...
      (any(real(x))|any(imag(x)))
    x=roundq(x*2,Q); bfe=bfe-1;
  end
  while any(real(x)>Q.Maximum)|any(real(x)<Q.Minimum)|...
      any(imag(x)>Q.Maximum)|any(imag(x)<Q.Minimum)
    x=roundq(x/2,Q); bfe=bfe+1;
  end
else
  ovfind=[ovfind;ovf(:)];
  if ~isempty(ovf), error('Overflow during input scaling'), end
end
%
X=ones(size(x))+j; bfemod=0; bfenew=bfe;
for k=0:length(X)-1
  if sign_of_life
    fprintf('.'), if rem(k+1,50)==0, disp(' '), end
  end
  X(k+1)=0;
  if ~strncmpi(alg,'grouped',2)
    for n=0:length(x)-1
      ei=rem(k*n,N);
      if ~strncmpi(runmod.direction,'inverse',1)
        W=exp(-j*2*pi*ei/N);
        if rem(ei,N/4)==0, Wq=W; else [Wq,ovf]=roundq(W,Qcos); end
        term=x(n+1)*Wq;
      else
        W=exp(j*2*pi*ei/N);
        if rem(ei,N/4)==0, Wq=W; else [Wq,ovf]=roundq(W,Qcos); end
        term=x(n+1)*W;
      end
      if any(ovf), error('Overflow error'), end
      [term,ovf]=roundq(term/2^bfemod,Qacc,'complex');
      if any(ovf), error('Overflow error'), end
      [Xkp1,ovfind]=roundq(X(k+1)+term,Qacc,'complex');
      if strncmpi(exponent,'block-float',1)&any(ovfind)
        while any(ovfind)
          bfe=bfe+1; bfemod=bfemod+1;
          for ki=0:k
            [X(ki+1),ovfind]=roundq(X(ki+1)/2,Q,'complex');
          end
          term=roundq(term/2,Qacc,'complex');
          [Xkp1,ovfind]=roundq(X(k+1)+term,Qacc,'complex');
        end %while ovfind
      elseif any(ovfind)
        error('Overflow error')
      end
      X(k+1)=Xkp1;
    end %for n
    %X(k+1) is in Xkp1: Qacc
    [X(k+1),ovfind]=roundq(Xkp1,Q,'complex');
    if strncmpi(exponent,'block-float',1)&any(ovfind)
      while any(ovfind)
        bfe=bfe+1; bfemod=bfemod+1;
        for ki=0:k-1
          X(ki+1)=roundq(X(ki+1)/2,Q);
        end
        Xkp1=Xkp1/2;
        [X(k+1),ovfind]=roundq(Xkp1,Q,'complex');
      end
    elseif any(ovfind)
      error('Overflow error')
    end
  else %grouped algorithm
    bfek=bfe;
    kn=zeros(size(x));
    kn(1:N)=rem([0:N-1]*k,N);
    if ~strncmpi(runmod.direction,'inverse',1)
      [W,ovf]=roundq(exp(-j*2*pi*kn/N),Qcos);
      Xn=x.*W;
    else
      [W,ovf]=roundq(exp(j*2*pi*kn/N),Qcos);
      Xn=x.*W;
    end
    if any(ovf), error('Overflow error'), end
    Nm=pow2(ceil(log2(N)));
    Xn=[Xn,zeros(1,Nm-N)];
    for L=pow2([log2(Nm):-1:0])
      if strncmpi(exponent,'downscale',2)
        if L==Nm
          [Xn1L,ovfind]=roundq(Xn(1:L)/2,Q);
        else
          [Xn1L,ovfind]=roundq(Xn(1:L)+Xn(L+[1:L])/2,Q);
        end
      else
        if L==Nm
          [Xn1L,ovfind]=roundq(Xn(1:L),Q);
        else
          [Xn1L,ovfind]=roundq(Xn(1:L)+Xn(L+[1:L]),Q);
        end
      end
      if strncmpi(exponent,'block-float',1)&any(ovfind)
        while any(ovfind)
          bfek=bfek+1;
          if L==Nm, 
            Xn(1:L)=roundq(Xn(1:L)/2,Q);
            [Xn1L,ovfind]=roundq(Xn(1:L),Q);
          else
            Xn(1:max(end,2*L))=roundq(Xn(1:max(end,2*L))/2,Q);
            [Xn1L,ovfind]=roundq(Xn(1:L)+Xn(L+[1:L]),Q);
          end
        end
      elseif any(ovfind)
        error('Overflow error')
      end
      Xn(1:L)=Xn1L;
    end %for L
    if bfek>=bfenew
      for kb=1:k 
        [Xkb,ovf]=roundq(X(kb)*pow2(bfenew-bfek),Q); 
        if any(ovf), error('Overflow error'), end
        X(kb)=Xkb;
      end
      bfenew=bfek;
      [Xkp1,ovf]=roundq(Xn(1),Q);
    else
      [Xkp1,ovf]=roundq(Xn(1)/pow2(bfenew-bfek),Q);
    end
    if any(ovf), error('Overflow error'), end
    X(k+1)=Xkp1;
    if isequal(k,length(X)-1), bfe=bfenew; end
  end %grouped
  if sign_of_life&isequal(k,length(X)-1), disp(' '), end
end %for k
%
if strncmpi(runmod.direction,'inverse',1)
  if strncmpi(exponent,'block-float',1)
    bfe=bfe-log2(N);
  elseif strncmpi(exponent,'downscale',2)
  else
    X=roundq(1/N*X,Q,'complex'); 
  end
end
%
if strncmpi(exponent,'block-float',1)
  X=struct('data',X,'bfe',bfe);
end
%End of rounddft
