function [X,ovfind]=roundfft(x,Q,runmod)
%ROUNDFFT Quantized FFT algorithm
%
%   X=roundfft(x,Q,runmod)
%
%   Input arguments:
%     x = input sequence (if the length is not 2's power, rounddft_chirp is
%          called (except for qdata objects)
%        if x is qdata, then rounding is qdata-based
%        X=roundfft(x,runmod)
%     Q = definition of the quantizer (if not given, IEEE double)
%        e.g. Q=qstruct('B',32,'fb',31);
%     runmod = run modifiers
%        Qinput: (quantizer, default: Q)
%        Qcos: (quantizer for trigonometric constants, default: Q)
%        Qmul: (quantizer after multiplication, default: Q)
%        direction: 'forward' (default), 'inverse'
%        exponent: 'block-float', 'const' (default for fixed-point)
%             'downscale': downscale after each addition/subtraction by 2 
%             for floating-point, block-float or downscale are not allowed
%        algorithm: 'DIT' or 'DIF', default: 'DIT'
%   Output arguments:
%     X = vector, or structure for block-float 
%          fields for bf: data, bfe
%
%   Example:
%     roundifft(roundfft([1:8]))-[1:8]

%	Copyright (c) I. Kollar, 2006-8, all rights reserved
%	Last modified: 6-Aug-2008

global show_indices %test listing
if isempty(show_indices), show_indices=0; end %0/1 show indices as called

ovfind=[];
if isa(x,'qdata'), error(nargchk(1,2,nargin)), end
if (isa(x,'qdata')&(nargin<2))|(nargin<3), runmod=[]; end
if ~isa(x,'qdata')
  if nargin<2, Q=''; end, if isempty(Q), Q='double'; end
elseif nargin>1, runmod=Q; 
end
if isstruct(x), bfe=x.bfe; x=x.data;
else bfe=0;
end
N=length(x); 
if ~isequal(N,pow2(round(log2(N))))
  if isa(x,'qdata')
    error('Chirp dft is not yet ready for qdata')
  else
    if isequal(bfe,0), X=rounddft_chirp(x,Q,runmod);
    else X=rounddft_chirp(struct('data',x,'bfe',bfe),Q,runmod);
    end 
  end
  return
end
M=log2(N);

Qcos=[]; Qinput=[]; Qmul=''; sign_of_life=0; alg='';
if isstr(runmod)
  if strncmpi(runmod,'inverse',1), runmod=struct('direction','inverse');
  elseif strncmpi(runmod,'block-float',2), runmod=struct('exponent','block-float');
  elseif strncmpi(runmod,'downscale',2), runmod=struct('exponent','downscale');
  end
elseif isstruct(runmod)
  fn=fieldnames(runmod);
  for ii=1:length(fn)
    if strcmp(fn{ii},'Qinput')
      Qinput=runmod.Qinput;
    elseif strcmp(fn{ii},'Qcos')|strcmp(fn{ii},'Qsin')
      Qcos=getfield(runmod,fn{ii});
    elseif strcmp(fn{ii},'Qmul')
      Qmul=runmod.Qmul;
    elseif strcmp(fn{ii},'direction')
    elseif strcmp(fn{ii},'exponent')
    elseif strcmp(fn{ii},'sign_of_life')
      sign_of_life=runmod.sign_of_life;
    elseif strcmp(fn{ii},'algorithm')
      alg=runmod.algorithm;
    else error(['Unrecognised runmod field ''',fn{ii},''''])
    end
  end %for ii
end
if ~isfield(runmod,'direction'), runmod.direction='f'; end
if isempty(alg)|strcmpi(alg,'dit'), alg='DIT';
elseif strcmpi(alg,'dif'), alg='DIF';
else error(['Algorithm is invalid as ''',alg,''''])
end
if isfield(runmod,'exponent')
  exponent=runmod.exponent;
  if strncmpi(Q.Type,'floating-point',2)
    error('Cannot implement block-float for floating-point quantizer')
  elseif strncmpi(Q.Type,'downscale',2)
    error('Cannot implement downscaling for floating-point quantizer')
  end
else exponent='';
end

if isa(x,'double')
  if isstr(Q), Q=qstruct(Q); end
  if isempty(Qinput), Qinput=Q; end, if isstr(Qinput), Qinput=qstruct(Qinput); end 
  if isempty(Qmul), Qmul=Q; end, if isstr(Qmul), Qmul=qstruct(Qmul); end
  if isempty(Qcos)
    if strcmp(Q.Type,'fixed-point')
      Qcos=qstructset(Q,'FractionBits',Q.Bits-2,...
        'Maximum',2-2^(-(Q.Bits-2)),'Minimum',-2); %amplitude of sine is 1, range is up to 2
      if ~strcmp(Q.Coding,'t'), Qcos=qstructset(Qcos,'Minimum',-Qcos.Maximum); end
    else
      Qcos=Q;
    end
  end
  if isstr(Qcos), Qcos=qstruct(Qcos); end
else %qdata
  Q=x.name;
  if isempty(Qcos)
    if strcmp(qget(Q,'Type'),'fixed-point')
      Qcos=Q; qset(Qcos,'FractionBits',Q.Bits-2,...
        'Maximum',2-2^(-(Q.Bits-2)),'Minimum',-2); %amplitude of sine is 1, range is up to 2
      if ~strcmp(Q.Coding,'t'), Qcos=qset(Qcos,'Minimum',-Qcos.Maximum); end
    else
      Qcos=Q;
    end
  end
  qpi=qpi_str(x); qpi=qdata(qpi,Q);
end
%
if strncmpi(exponent,'block-float',1)
  %scale input to maximum range
  if strncmp(Qinput.Type,'fixed-point',2)
    while all(real(x)<=Qinput.Maximum/2)&all(real(x)>=Qinput.Minimum/2)&...
        all(imag(x)<=Qinput.Maximum/2)&all(imag(x)>=Qinput.Minimum/2)&...
        (any(real(x))|any(imag(x)))
      x=x*2; bfe=bfe-1;
    end
    while any(real(x)>Qinput.Maximum)|any(real(x)<Qinput.Minimum)|...
        any(imag(x)>Qinput.Maximum)|any(imag(x)<Qinput.Minimum)
      x=x/2; bfe=bfe+1;
    end
  end
end
[x,ovfindi]=roundq(x,Qinput);
if strncmpi(exponent,'block-float',1)
  %scale input to maximum range
  if strncmp(Q.Type,'fixed-point',2)
    while all(real(x)<=Q.Maximum/2)&all(real(x)>=Q.Minimum/2)&...
        all(imag(x)<=Q.Maximum/2)&all(imag(x)>=Q.Minimum/2)&...
        (any(real(x))|any(imag(x)))
      x=roundq(x*2,Q); bfe=bfe-1;
    end
    while any(real(x)>Q.Maximum)|any(real(x)<Q.Minimum)|...
        any(imag(x)>Q.Maximum)|any(imag(x)<Q.Minimum)
      x=roundq(x/2,Q); bfe=bfe+1;
    end
  end
else
  ovfind=[ovfind;ovfindi(:)];
end %block-float
%
b_prev=0;
for b=1:M %number of stages
  ind=[0:N-1];
  wind=0; butterfly2=0;
  for ii=0:N-1 %samples
    if strcmpi(alg,'DIT')
      jj=ii+N/2^b;
      if ismember(ii,ind)
        %ii,jj
        if butterfly2, wind=wind+1; end
        kn=bitreverse(wind,N/2);
        if strncmpi(runmod.direction,'inverse',1), kn=-kn; end
        if show_indices
          if (b==1)&(ii==0), fprintf([alg]), end, if b_prev==b-1, b_prev=b; disp(' '), end
          fprintf('Stage: %.0f, x(%.0f), x(%.0f), exp: %.0f\n',b,ii,jj,rem(kn,N))
        end
        ovfind=1; %do first cycle
        while any(ovfind)
          xii=x(ii+1); xjj=x(jj+1); 
          if isa(x,'qdata')
            Wkn=cos(qpi*2/N*kn)-j*sin(qpi*2/N*kn); 
            Wknq=qdata(Wkn,Qcos);
            term2=real(xjj)*real(Wknq)-imag(xjj)*imag(Wknq)+...
              j*(real(xjj)*imag(Wknq)+imag(xjj)*real(Wknq));
            ovfindm=[];
          else %numeric
            Wkn=exp(-j*2*pi/N*kn);
            if rem(kn,N/4)~=0, [Wknq,ovfindw]=roundq(Wkn,Qcos); ovfind=[ovfind;ovfindw(:)];
            else Wknq=Wkn;
            end
            [term2,ovfindm]=cmulq(xjj,Wknq,Qmul,Q);
          end
          xiinew=xii+term2; xjjnew=xii-term2;
          if strncmp(exponent,'downscaling',2)
            xiinew=xiinew/2; xjjnew=xjjnew/2;
          end
          if isa(x,'numeric')&~strncmp(Q.Type,'fixed-point',2)
            [xiinew,ovfindi]=roundq(xiinew,Q);
            [xjjnew,ovfindj]=roundq(xjjnew,Q);
          else
            ovfindi=[]; ovfindj=[];
            if ~isa(x,'qdata')&strncmp(Q.Type,'fixed-point',2)
              if any([real(xiinew),imag(xiinew)]>Q.Maximum)|any([real(xiinew),imag(xiinew)]<Q.Minimum)
                [xiinew,ovfindi]=roundq(xiinew,Q);
              end
              if any([real(xjjnew),imag(xjjnew)]>Q.Maximum)|any([real(xjjnew),imag(xjjnew)]<Q.Minimum)
                [xjjnew,ovfindj]=roundq(xjjnew,Q);
              end
            end
          end
          ovfind=[ovfindm(:);ovfindi(:);ovfindj(:)];
          ovfind=[real(ovfind);imag(ovfind)];
          if any(ovfind)
            if strncmpi(exponent,'block-float',2), x=roundq(x/2,Q); bfe=bfe+1;
            else
              error('Overflow error')
              %ovfind=[ovfind;ovfind*[ii+1;jj+1]];
              ovfind=[];
            end
          else %no overflow
            %test only
            if strncmpi(exponent,'block-float',1)
              if any([real(term2),imag(term2)]>Q.Maximum)|any([real(term2),imag(term2)]<Q.Minimum)|...
                  any([real(xiinew),imag(xiinew)]>Q.Maximum)|any([real(xiinew),imag(xiinew)]<Q.Minimum)|...
                  any([real(xjjnew),imag(xjjnew)]>Q.Maximum)|any([real(xjjnew),imag(xjjnew)]<Q.Minimum)
                error('block-float error: overflow not detected')
              end
            end
          end
        end %while any(ovfind)
        x(ii+1)=xiinew; x(jj+1)=xjjnew;
        %
        indjj=find(jj==ind);
        ind(indjj)=-1;
        butterfly2=0;
      else %not a member, second element of butterfly
        butterfly2=1;
      end
    elseif strcmp(alg,'DIF') %DIF
      jj=ii+N/2^b;
      if ismember(ii,ind)
        %ii,jj
        kn=wind;
        if strncmpi(runmod.direction,'inverse',1), kn=-kn; end
        if show_indices
          if (b==1)&(ii==0), fprintf([alg]), end, if b_prev==b-1, b_prev=b; disp(' '), end
          fprintf('Stage: %.0f, samples: %.0f/%.0f, kn = %.0f\n',b,ii,jj,rem(kn,N))
        end
        ovfind=1; %do first cycle
        while any(ovfind)
          xii=x(ii+1); xjj=x(jj+1); 
          xiinew=xii+xjj; xjjnew=xii-xjj;
          if strncmp(exponent,'downscaling',2)
            xiinew=xiinew/2; xjjnew=xjjnew/2;
          end
          if isa(x,'qdata')
            Wkn=cos(qpi*2/N*kn)-j*sin(qpi*2/N*kn); Wkn=qdata(Wkn,Qcos);
            xjjnew=real(xjjnew)*real(Wkn)-imag(xjjnew)*imag(Wkn)+...
              j*(real(xjjnew)*imag(Wkn)+imag(xjjnew)*real(Wkn));
            ovfindm=[];
          else
            if ~strncmp(Q.Type,'fixed-point',2)
              [xiinew,ovfindi]=roundq(xiinew,Q);
              [xjjnew,ovfindj]=roundq(xjjnew,Q);
            else ovfindi=[]; ovfindj=[];
            end
            Wkn=exp(-j*2*pi/N*kn);
            if rem(kn,N/4)~=0
              [Wknq,ovfindw]=roundq(Wkn,Qcos); ovfind=[ovfind;ovfindw(:)];
            else Wknq=Wkn;
            end
            [xjjnew,ovfindm]=cmulq(xjjnew,Wknq,Qmul,Q);
          end
          ovfind=[ovfindm(:);ovfindi(:);ovfindj(:)];
          ovfind=[real(ovfind);imag(ovfind)];
          if any(ovfind)
            if strncmpi(exponent,'block-float',2), x=roundq(x/2,Q); bfe=bfe+1;
            else
              error('Overflow error')
              %ovfind=[ovfind;ovfind*[ii+1;jj+1]];
              ovfind=[];
            end
          end
        end %while any(ovfind)
        x(ii+1)=xiinew; x(jj+1)=xjjnew;
        %
        indjj=find(jj==ind);
        ind(indjj)=-1;
        butterfly2=0;
        if ~butterfly2, wind=rem(wind+2^(b-1),N/2); end
      else %not a member, second element of butterfly
        butterfly2=1;
      end
    else error('alg not allowed')
    end %if alg
  end %for ii
end %for b
%
if isa(x,'qdata'), X=x; %x(N)=x(N)+j;
else X=zeros(size(x))+j; %assign X
end
for ii=0:N-1, X(bitreverse(ii,N)+1)=x(ii+1); end
if strncmpi(runmod.direction,'inverse',1)
  if strncmpi(exponent,'block-float',1)
    bfe=bfe-log2(N);
  elseif strncmpi(exponent,'downscale',2)
  else
    if isa(x,'qdata'),
      X=roundq(1/N*X);
    else
      X=roundq(1/N*X,Q,'complex');
    end
  end
end
if strncmpi(exponent,'block-float',1)
  X=struct('data',X,'bfe',bfe);
end
%End of roundfft

function ind=bitreverse(in,N)
if (in<0)|(in>N), error('in is out of range'), end
if rem(in,1), error('in is not integer'), end
M=log2(N);
ind=sum((2.^[M-1:-1:0]).*(rem(in,2.^[1:M])>=2.^[0:M-1]));
