function value=ulp(x,p)
%ULP Unit in the last place for floating-point quantization
%       The gap between the two finite floating-point numbers nearest x, even if x is one of them
%         for abs(x)<=realmax+ulp(realmax)/2, the gap contains x
%         otherwise its width is ulp(realmax)
%
%       value=ulp(x,p)
%
%   In the place of p, structure of run modifiers can also be given, as
%       struct('p',p,'UnderFlow','flush to zero')
%
%   Examples: 
%       ulp(1)  %returns eps/2 = 2^(-53) = 1.1102e-016
%       ulp(2)  %returns eps = 2^(-52) = 2.2204e-016
%       ulp(1.55,24)  %returns 1.1921e-007 = 2^(-23)
%       ulp(qfloat(1.55,24),24)  %returns 1.1921e-007 = 2^(-23)
%       log2(ulp(qfloat(2.55,24),24))  %returns -22
%       ulp(realmin/1.1,struct('underflow','flush to zero'))  %returns 2^(-1022)/2 = 1.1125e-308
%       ulp(realmin)  %returns 2^(-1022-52) = 4.9407e-324

%  See J.-M. Muller, "On the definition of ulp(x)," Research Report No. 2005-09,
%    Ecole Normale  Superieure de Lyon, France, 
%    http://www.ens-lyon.fr/LIP/Pub/Rapports/RR/RR2005/RR2005-09.pdf
    
%	Copyright (c) I. Kollar, 2005, all rights reserved
%	Last modified: 12-Dec-2005

if ~isnumeric(x), error('x is not numeric'), end
if length(x)==0, value=[]; return, end
value=zeros(size(x));
%
lb='h'; Emin=-1022; Emax=1023; bias=[];
if ~exist('p'), p=''; end, if isempty(p), p=53; end
if ~exist('ufl'), ufl=''; end, if isempty(ufl), ufl='g'; end
if nargin>1
  runmod=p;
  if isstruct(runmod)
    p=53;
    fn=fieldnames(runmod);
    for ii=1:length(fn)
      if strncmpi(fn{ii},'type',3), typ=getfield(runmod,fn{ii});
      elseif strcmp(fn{ii},'p'), p=getfield(runmod,'p');
      elseif strcmp(fn{ii},'Emin'), Emin=getfield(runmod,'Emin'); Emingiven=1;
      elseif strcmp(fn{ii},'Emax'), Emax=getfield(runmod,fn{ii});
      elseif strcmp(fn{ii},'bias'), bias=getfield(runmod,'Ebias');
      elseif strcmpi(fn{ii},'lb')|strncmpi(fn{ii},'leadingbit',1), 
        lb=getfield(runmod,fn{ii}); lb=lower(lb(1));
      elseif strcmpi(fn{ii},'underflow')|strncmpi(fn{ii},'ufl',2), 
        ufl=getfield(runmod,fn{ii}); ufl=lower(ufl(1));
      elseif strncmpi(fn{ii},'Operation',2)
        op=getfield(runmod,fn{ii});
      else error(['Unknown field ''',fn{ii},''' of runmod'])
      end
    end
  elseif isnumeric(runmod)
    p=runmod;
  end
end

if isstr(p)
  if strncmp(p,'single',1)
    lb='h'; p=24; Emin=-126; Emax=127; ufl='g';
  else 
    error(['p = ''',p,''' is not allowed'])
  end
else
  if isempty('Emin'), Emin=-1022; Emax=1023; end
end
%
for ii=1:prod(size(x))
  xi=x(ii);
  if isnan(xi), value(ii)=NaN;
  else
    if ~isfinite(xi), xi=sign(xi)*pow2(1-2^(-p+1),Emax+1); end %adjust infinity
    xp=qfloat(xi,qstruct('p',p,'ufl',ufl,'Emin',Emin,'Emax',Emax));
    %if ~isequal(xi,xp), error(sprintf('Element %.0f of x has not precision p',ii)), end
    if isequal(xi,0), ei=Emin;
    else [fi,ei]=log2(xi);
    end
    value(ii)=pow2(max(ei-p,Emin-p+1));
    if strcmp(ufl,'f')
      if abs(xi)<=pow2(1,Emin)
        value(ii)=pow2(1,Emin-1); %flush to zero
      end
    elseif strcmp(ufl,'g')
    else error('ufl is not allowed')
    end
    if isequal(abs(fi),0.5)&(ei>Emin+1), value(ii)=value(ii)/2; end
  end
end
%
%End of ulp